@@ -15,6 +15,7 @@
 #include "osmocom/bsc/gsm_data.h"
 #include "osmocom/bsc/bts_trx.h"
 #include "osmocom/bsc/bts_sm.h"
+#include "osmocom/bsc/abis_om2000.h"
 
 enum bts_counter_id {
 	BTS_CTR_CHREQ_TOTAL,
@@ -26,7 +27,15 @@ enum bts_counter_id {
 	BTS_CTR_CHREQ_ATTEMPTED_OTHER,
 	BTS_CTR_CHREQ_ATTEMPTED_UNKNOWN,
 	BTS_CTR_CHREQ_SUCCESSFUL,
+	BTS_CTR_CHREQ_SUCCESSFUL_EMERG,
+	BTS_CTR_CHREQ_SUCCESSFUL_CALL,
+	BTS_CTR_CHREQ_SUCCESSFUL_LOCATION_UPD,
+	BTS_CTR_CHREQ_SUCCESSFUL_PAG,
+	BTS_CTR_CHREQ_SUCCESSFUL_PDCH,
+	BTS_CTR_CHREQ_SUCCESSFUL_OTHER,
+	BTS_CTR_CHREQ_SUCCESSFUL_UNKNOWN,
 	BTS_CTR_CHREQ_NO_CHANNEL,
+	BTS_CTR_CHREQ_MAX_DELAY_EXCEEDED,
 	BTS_CTR_CHAN_RF_FAIL,
 	BTS_CTR_CHAN_RF_FAIL_TCH,
 	BTS_CTR_CHAN_RF_FAIL_SDCCH,
@@ -62,6 +71,7 @@ enum bts_counter_id {
 	BTS_CTR_LCHAN_BORKEN_EV_RF_CHAN_REL_ACK,
 	BTS_CTR_LCHAN_BORKEN_EV_VTY,
 	BTS_CTR_LCHAN_BORKEN_EV_TEARDOWN,
+	BTS_CTR_LCHAN_BORKEN_EV_TS_ERROR,
 	BTS_CTR_LCHAN_BORKEN_FROM_WAIT_RR_CHAN_MODE_MODIFY_ACK,
 	BTS_CTR_LCHAN_BORKEN_FROM_WAIT_RSL_CHAN_MODE_MODIFY_ACK,
 	BTS_CTR_TS_BORKEN_FROM_NOT_INITIALIZED,
@@ -121,6 +131,13 @@ enum bts_counter_id {
 	BTS_CTR_INTRA_BSC_HO_TIMEOUT,
 	BTS_CTR_INTRA_BSC_HO_FAILED,
 	BTS_CTR_INTRA_BSC_HO_ERROR,
+	BTS_CTR_INCOMING_INTRA_BSC_HO_ATTEMPTED,
+	BTS_CTR_INCOMING_INTRA_BSC_HO_COMPLETED,
+	BTS_CTR_INCOMING_INTRA_BSC_HO_STOPPED,
+	BTS_CTR_INCOMING_INTRA_BSC_HO_NO_CHANNEL,
+	BTS_CTR_INCOMING_INTRA_BSC_HO_TIMEOUT,
+	BTS_CTR_INCOMING_INTRA_BSC_HO_FAILED,
+	BTS_CTR_INCOMING_INTRA_BSC_HO_ERROR,
 	BTS_CTR_INTER_BSC_HO_OUT_ATTEMPTED,
 	BTS_CTR_INTER_BSC_HO_OUT_COMPLETED,
 	BTS_CTR_INTER_BSC_HO_OUT_STOPPED,
@@ -134,12 +151,49 @@ enum bts_counter_id {
 	BTS_CTR_INTER_BSC_HO_IN_FAILED,
 	BTS_CTR_INTER_BSC_HO_IN_TIMEOUT,
 	BTS_CTR_INTER_BSC_HO_IN_ERROR,
+	BTS_CTR_SRVCC_ATTEMPTED,
+	BTS_CTR_SRVCC_COMPLETED,
+	BTS_CTR_SRVCC_STOPPED,
+	BTS_CTR_SRVCC_NO_CHANNEL,
+	BTS_CTR_SRVCC_TIMEOUT,
+	BTS_CTR_SRVCC_FAILED,
+	BTS_CTR_SRVCC_ERROR,
+	BTS_CTR_ALL_ALLOCATED_SDCCH,
+	BTS_CTR_ALL_ALLOCATED_STATIC_SDCCH,
+	BTS_CTR_ALL_ALLOCATED_TCH,
+	BTS_CTR_ALL_ALLOCATED_STATIC_TCH,
+	BTS_CTR_CM_SERV_REJ,
+	BTS_CTR_CM_SERV_REJ_IMSI_UNKNOWN_IN_HLR,
+	BTS_CTR_CM_SERV_REJ_ILLEGAL_MS,
+	BTS_CTR_CM_SERV_REJ_IMSI_UNKNOWN_IN_VLR,
+	BTS_CTR_CM_SERV_REJ_IMEI_NOT_ACCEPTED,
+	BTS_CTR_CM_SERV_REJ_ILLEGAL_ME,
+	BTS_CTR_CM_SERV_REJ_PLMN_NOT_ALLOWED,
+	BTS_CTR_CM_SERV_REJ_LOC_NOT_ALLOWED,
+	BTS_CTR_CM_SERV_REJ_ROAMING_NOT_ALLOWED,
+	BTS_CTR_CM_SERV_REJ_NETWORK_FAILURE,
+	BTS_CTR_CM_SERV_REJ_SYNCH_FAILURE,
+	BTS_CTR_CM_SERV_REJ_CONGESTION,
+	BTS_CTR_CM_SERV_REJ_SRV_OPT_NOT_SUPPORTED,
+	BTS_CTR_CM_SERV_REJ_RQD_SRV_OPT_NOT_SUPPORTED,
+	BTS_CTR_CM_SERV_REJ_SRV_OPT_TMP_OUT_OF_ORDER,
+	BTS_CTR_CM_SERV_REJ_CALL_CAN_NOT_BE_IDENTIFIED,
+	BTS_CTR_CM_SERV_REJ_INCORRECT_MESSAGE,
+	BTS_CTR_CM_SERV_REJ_INVALID_MANDANTORY_INF,
+	BTS_CTR_CM_SERV_REJ_MSG_TYPE_NOT_IMPLEMENTED,
+	BTS_CTR_CM_SERV_REJ_MSG_TYPE_NOT_COMPATIBLE,
+	BTS_CTR_CM_SERV_REJ_INF_ELEME_NOT_IMPLEMENTED,
+	BTS_CTR_CM_SERV_REJ_CONDTIONAL_IE_ERROR,
+	BTS_CTR_CM_SERV_REJ_MSG_NOT_COMPATIBLE,
+	BTS_CTR_CM_SERV_REJ_PROTOCOL_ERROR,
+	BTS_CTR_CM_SERV_REJ_RETRY_IN_NEW_CELL,
 };
 
 extern const struct rate_ctr_desc bts_ctr_description[];
 extern const struct rate_ctr_group_desc bts_ctrg_desc;
 
 enum {
+	BTS_STAT_UPTIME_SECONDS,
 	BTS_STAT_CHAN_LOAD_AVERAGE,
 	BTS_STAT_CHAN_CCCH_SDCCH4_USED,
 	BTS_STAT_CHAN_CCCH_SDCCH4_TOTAL,
@@ -155,8 +209,8 @@ enum {
 	BTS_STAT_CHAN_CCCH_SDCCH4_CBCH_TOTAL,
 	BTS_STAT_CHAN_SDCCH8_CBCH_USED,
 	BTS_STAT_CHAN_SDCCH8_CBCH_TOTAL,
-	BTS_STAT_CHAN_TCH_F_TCH_H_PDCH_USED,
-	BTS_STAT_CHAN_TCH_F_TCH_H_PDCH_TOTAL,
+	BTS_STAT_CHAN_OSMO_DYN_USED,
+	BTS_STAT_CHAN_OSMO_DYN_TOTAL,
 	BTS_STAT_T3122,
 	BTS_STAT_RACH_BUSY,
 	BTS_STAT_RACH_ACCESS,
@@ -164,6 +218,8 @@ enum {
 	BTS_STAT_RSL_CONNECTED,
 	BTS_STAT_LCHAN_BORKEN,
 	BTS_STAT_TS_BORKEN,
+	BTS_STAT_NUM_TRX_RSL_CONNECTED,
+	BTS_STAT_NUM_TRX_TOTAL,
 };
 
 extern const struct osmo_stat_item_desc bts_stat_desc[];
@@ -225,6 +281,8 @@ struct gsm_bts_model {
 	int (*power_ctrl_enc_rsl_params)(struct msgb *msg, const struct gsm_power_ctrl_params *cp);
 	/* (Optional) function for sending default MS/BS Power Control paramaters */
 	int (*power_ctrl_send_def_params)(const struct gsm_bts_trx *trx);
+	/* (Optional) function for toggling BCCH carrier power reduction operation */
+	int (*power_ctrl_set_c0_power_red)(const struct gsm_bts *bts, const uint8_t red);
 
 	void (*config_write_bts)(struct vty *vty, struct gsm_bts *bts);
 	void (*config_write_trx)(struct vty *vty, struct gsm_bts_trx *trx);
@@ -280,6 +338,8 @@ struct gsm_bts {
 
 	/* Connected PCU version (if any) */
 	char pcu_version[MAX_VERSION_LENGTH];
+	/* PCU sign_link, over OML line: */
+	struct e1inp_sign_link *osmo_link;
 
 	/* maximum Tx power that the MS is permitted to use in this cell */
 	int ms_max_power;
@@ -378,6 +438,7 @@ struct gsm_bts {
 				uint16_t limit;
 				uint16_t active;
 			} om2k_version[16];
+			enum om2k_sync_src sync_src;
 		} rbs2000;
 		struct {
 			uint8_t bts_type;
@@ -434,6 +495,14 @@ struct gsm_bts {
 	 * rather than starting from TRX0 and go upwards? */
 	int chan_alloc_reverse;
 
+	/* When true, interference measurements from the BTS are used in the channel allocator to favor lchans with less
+	 * interference reported in RSL Resource Indication. */
+	bool chan_alloc_avoid_interf;
+
+	/* When true (default), TCH can be allocated to serve
+	 * non-voicecall-related signalling services when SDCCHs are exhausted */
+	bool chan_alloc_allow_tch_for_signalling;
+
 	enum neigh_list_manual_mode neigh_list_manual_mode;
 	/* parameters from which we build SYSTEM INFORMATION */
 	struct {
@@ -448,6 +517,7 @@ struct gsm_bts {
 		struct bitvec si5_neigh_list;
 		struct osmo_earfcn_si2q si2quater_neigh_list;
 		size_t uarfcn_length; /* index for uarfcn and scramble lists */
+		size_t cell_chan_num; /* number of channels in Cell Allocation */
 		struct {
 			/* bitmask large enough for all possible ARFCN's */
 			uint8_t neigh_list[1024/8];
@@ -494,10 +564,13 @@ struct gsm_bts {
 
 	struct handover_cfg *ho;
 
-	/* A list of struct gsm_bts_ref, indicating neighbors of this BTS.
-	 * When the si_common neigh_list is in automatic mode, it is populated from this list as well as
-	 * gsm_network->neighbor_bss_cells. */
-	struct llist_head local_neighbors;
+	/* Local and remote neighbor configuration: a list of neighbors as written in the VTY config, not resolved to
+	 * actual cells. Entries may point at non-existing BTS numbers, or yet unconfigured ARFCN+BSIC. The point of
+	 * this list is to keep the config as the user entered it: a) to write it back exactly as entered, and b) to
+	 * allow adding neighbor cells that will only be configured further down in the config file.
+	 * An actual neighbor cell object (local or remote-BSS) is resolved "at runtime" whenever a neighbor is being
+	 * looked up. */
+	struct llist_head neighbors;
 
 	/* BTS-specific overrides for timer values from struct gsm_network. */
 	uint8_t T3122;	/* ASSIGNMENT REJECT wait indication */
@@ -517,13 +590,42 @@ struct gsm_bts {
 	struct llist_head oml_fail_rep;
 	struct llist_head chan_rqd_queue;
 
-	/* osmocom specific FACCH/SACCH repetition mode flags set by VTY to
-	 * enable/disable certain ACCH repeation features individually */
-	struct abis_rsl_osmo_rep_acch_cap repeated_acch_policy;
+	/* ACCH Repetition capabilities */
+	struct abis_rsl_osmo_rep_acch_cap rep_acch_cap;
+
+	/* ACCH Temporary overpower capabilities */
+	struct abis_rsl_osmo_temp_ovp_acch_cap top_acch_cap;
+	/* Channel mode(s) for which to allow TOP */
+	enum {
+		TOP_ACCH_CHAN_MODE_ANY = 0,	/* Any kind of channel mode */
+		TOP_ACCH_CHAN_MODE_SPEECH_V3,	/* Speech channels using AMR codec */
+	} top_acch_chan_mode;
 
 	/* MS/BS Power Control parameters */
 	struct gsm_power_ctrl_params ms_power_ctrl;
 	struct gsm_power_ctrl_params bs_power_ctrl;
+
+	/* Maximum BCCH carrier power reduction */
+	uint8_t c0_max_power_red_db;
+
+	/* Interference Measurement Parameters, as read from VTY */
+	struct gsm_interf_meas_params interf_meas_params_cfg;
+	/* Interference Measurement Parameters, as last sent via OML */
+	struct gsm_interf_meas_params interf_meas_params_used;
+
+	/* We will ignore CHAN RQD with access delay greater than rach_max_delay */
+	uint8_t rach_max_delay;
+
+	/* Is Fast return to LTE allowed during Chan Release in this BTS? */
+	bool srvcc_fast_return_allowed;
+
+	/* At what point in the channel allocation sequence to dispatch the Immediate Assignment (Abis optimization) */
+	enum imm_ass_time imm_ass_time;
+
+	struct osmo_time_cc all_allocated_sdcch;
+	struct osmo_time_cc all_allocated_static_sdcch;
+	struct osmo_time_cc all_allocated_tch;
+	struct osmo_time_cc all_allocated_static_tch;
 };
 
 #define GSM_BTS_SI2Q(bts, i)   (struct gsm48_system_information_type_2quater *)((bts)->si_buf[SYSINFO_TYPE_2quater][i])
@@ -543,7 +645,7 @@ static inline int is_ipaccess_bts(const struct gsm_bts *bts)
 	return 0;
 }
 
-static inline int is_sysmobts_v2(const struct gsm_bts *bts)
+static inline int is_osmobts(const struct gsm_bts *bts)
 {
 	switch (bts->type) {
 	case GSM_BTS_TYPE_OSMOBTS:
@@ -620,6 +722,8 @@ char *gsm_bts_name(const struct gsm_bts *bts);
 
 bool gsm_bts_matches_lai(const struct gsm_bts *bts, const struct osmo_location_area_id *lai);
 bool gsm_bts_matches_cell_id(const struct gsm_bts *bts, const struct gsm0808_cell_id *cell_id);
+void gsm_bts_cell_id(struct gsm0808_cell_id *cell_id, const struct gsm_bts *bts);
+void gsm_bts_cell_id_list(struct gsm0808_cell_id_list2 *cell_id_list, const struct gsm_bts *bts);
 
 int gsm_bts_local_neighbor_add(struct gsm_bts *bts, struct gsm_bts *neighbor);
 int gsm_bts_local_neighbor_del(struct gsm_bts *bts, const struct gsm_bts *neighbor);
@@ -633,6 +737,9 @@ struct gsm_bts_trx *gsm_bts_trx_num(const struct gsm_bts *bts, int num);
 
 int bts_gprs_mode_is_compat(struct gsm_bts *bts, enum bts_gprs_mode mode);
 
+#define BTS_STORE_UPTIME_INTERVAL 10 /* in seconds */
+void bts_store_uptime(struct gsm_bts *bts);
+
 unsigned long long bts_uptime(const struct gsm_bts *bts);
 
 char *get_model_oml_status(const struct gsm_bts *bts);
@@ -655,10 +762,12 @@ void gsm_bts_set_radio_link_timeout(struct gsm_bts *bts, int value);
 
 void gsm_bts_all_ts_dispatch(struct gsm_bts *bts, uint32_t ts_ev, void *data);
 
-int bts_count_free_ts(struct gsm_bts *bts, enum gsm_phys_chan_config pchan);
-
 int gsm_bts_set_system_infos(struct gsm_bts *bts);
 
+int gsm_bts_set_c0_power_red(struct gsm_bts *bts, const uint8_t red);
+
+void gsm_bts_stats_reset(struct gsm_bts *bts);
+
 int gsm_bts_model_register(struct gsm_bts_model *model);
 struct gsm_bts_model *bts_model_find(enum gsm_bts_type type);
 
