@@ -18,6 +18,7 @@
 #include <osmocom/gsm/gsm48.h>
 #include <osmocom/core/fsm.h>
 #include <osmocom/core/tdef.h>
+#include <osmocom/core/time_cc.h>
 
 #include <osmocom/crypt/auth.h>
 
@@ -31,7 +32,6 @@
 #include <osmocom/abis/e1_input.h>
 #include <osmocom/bsc/meas_rep.h>
 #include <osmocom/bsc/acc.h>
-#include <osmocom/bsc/neighbor_ident.h>
 #include <osmocom/bsc/osmux.h>
 
 #define GSM_T3122_DEFAULT 10
@@ -106,16 +106,53 @@ enum channel_rate {
 	CH_RATE_FULL,
 };
 
+enum channel_rate chan_t_to_chan_rate(enum gsm_chan_t chan_t);
+
+enum lchan_csd_mode {
+	LCHAN_CSD_M_NT,
+	LCHAN_CSD_M_T_1200_75,
+	LCHAN_CSD_M_T_600,
+	LCHAN_CSD_M_T_1200,
+	LCHAN_CSD_M_T_2400,
+	LCHAN_CSD_M_T_9600,
+	LCHAN_CSD_M_T_14400,
+	LCHAN_CSD_M_T_29000,
+	LCHAN_CSD_M_T_32000,
+};
+
 struct channel_mode_and_rate {
 	enum gsm48_chan_mode chan_mode;
 	enum channel_rate chan_rate;
 	uint16_t s15_s0;
+	/* only used for GSM48_CMODE_DATA_* */
+	enum lchan_csd_mode csd_mode;
+};
+
+enum assign_for {
+	ASSIGN_FOR_NONE,
+	ASSIGN_FOR_BSSMAP_REQ,
+	ASSIGN_FOR_CONGESTION_RESOLUTION,
+	ASSIGN_FOR_VTY,
+};
+
+extern const struct value_string assign_for_names[];
+static inline const char *assign_for_name(enum assign_for assign_for)
+{ return get_value_string(assign_for_names, assign_for); }
+
+/* If .present is false, use the default value defined elsewhere. If true, use .val below.
+ * (A practical benefit of this is that the default initialization sets .present to false, so that even if a .val == 0
+ * is a valid value, a struct containing this as member does not need to explicitly set .val = INVALID_VAL_CONSTANT.) */
+struct optional_val {
+	bool present;
+	int val;
 };
 
 /* Information retrieved during an Assignment Request from the MSC. This is storage of the Assignment instructions
  * parsed from the Assignment Request message, to pass on until the gscon and assignment FSMs have decided whether an
  * Assignment is actually going to be carried out. Should remain unchanged after initial decoding. */
 struct assignment_request {
+	enum assign_for assign_for;
+
 	bool aoip;
 
 	uint16_t msc_assigned_cic;
@@ -127,15 +164,37 @@ struct assignment_request {
 
 	/* Rate/codec setting in preference order (need at least 1 !) */
 	int n_ch_mode_rate;
-	struct channel_mode_and_rate ch_mode_rate[3];
+	struct channel_mode_and_rate ch_mode_rate_list[3];
+
+	/* An assignment request usually requests to assign any available lchan, to match above requirements. This may
+	 * also choose to just keep the current lchan and merely modify it as appropriate. In these cases, keep
+	 * target_lchan == NULL.
+	 * In some situations, an assignment to a specific target lchan is requested (congestion resolution, VAMOS
+	 * multiplexing, user request via VTY). In these situations, select a target lchan beforehand and point
+	 * target_lchan to it. */
+	struct gsm_lchan *target_lchan;
+
+	/* The TSC Set to use if 'use' is true, otherwise automatically determine the TSC Set value to use. Valid range
+	 * is 1 to 4, as described in 3GPP TS 45.002. */
+	struct optional_val tsc_set;
+	/* The TSC to use if 'use' is true, otherwise automatically determine the TSC value to use. Valid range is 0 to
+	 * 7, as described in 3GPP TS 45.002. */
+	struct optional_val tsc;
 };
 
 /* State of an ongoing Assignment, while the assignment_fsm is still busy. This serves as state separation to keep the
  * currently used lchan and gscon unmodified until the outcome of an Assignment is known. If the Assignment fails, this
  * state is simply discarded, and the gscon carries on with the original lchan remaining unchanged. */
 struct assignment_fsm_data {
+	/* The request as made by the caller, see GSCON_EV_ASSIGNMENT_START or reassignment_request_to_lchan() /
+	 * reassignment_request_to_chan_type().
+	 * conn->assignment.req is treated immutable: remains unchanged throughout the Assignment. The mutable fields
+	 * are below: choices and automatic adjustments are stored in conn->assignment.*, not conn->assignment.req.
+	 */
 	struct assignment_request req;
+
 	bool requires_voice_stream;
+	struct channel_mode_and_rate selected_ch_mode_rate;
 
 	struct osmo_fsm_inst *fi;
 	struct gsm_lchan *new_lchan;
@@ -174,10 +233,16 @@ extern const struct value_string handover_scope_names[];
 inline static const char *handover_scope_name(enum handover_scope val)
 { return get_value_string(handover_scope_names, val); }
 
+/* Cell ARFCN + BSIC. */
+struct cell_ab {
+	uint16_t arfcn;
+	uint8_t bsic;
+};
+
 struct handover_out_req {
 	enum hodec_id from_hodec_id;
 	struct gsm_lchan *old_lchan;
-	struct neighbor_ident_key target_nik;
+	struct cell_ab target_cell_ab;
 	enum gsm_chan_t new_lchan_type; /*< leave GSM_LCHAN_NONE to use same as old_lchan */
 };
 
@@ -185,6 +250,8 @@ struct handover_in_req {
 	struct gsm0808_channel_type ct;
 	struct gsm0808_speech_codec_list scl;
 	struct gsm0808_encrypt_info ei;
+	bool kc128_present;
+	uint8_t kc128[16];
 	struct gsm_classmark classmark;
 	/* chosen_encr_alg reflects the encoded value as in RSL_ENC_ALG_A5(a5_numer):
 	 * chosen_encr_alg == 1 means A5/0 i.e. no encryption, chosen_encr_alg == 4 means A5/3.
@@ -197,6 +264,8 @@ struct handover_in_req {
 	uint16_t msc_assigned_cic;
 	char msc_assigned_rtp_addr[INET6_ADDRSTRLEN];
 	uint16_t msc_assigned_rtp_port;
+	bool last_eutran_plmn_valid;
+	struct osmo_plmn_id last_eutran_plmn;
 };
 
 struct handover {
@@ -205,7 +274,11 @@ struct handover {
 	enum hodec_id from_hodec_id;
 	enum handover_scope scope;
 	enum gsm_chan_t new_lchan_type;
-	struct neighbor_ident_key target_cell;
+	struct cell_ab target_cell_ab;
+
+	/* For inter-BSC handover, this may reflect more than one Cell ID. Must also be set for intra-BSC handover,
+	 * because it is used as key for penalty timers (e.g. in handover decision 2). */
+	struct gsm0808_cell_id_list2 target_cell_ids;
 
 	uint8_t ho_ref;
 	struct gsm_bts *new_bts;
@@ -248,7 +321,7 @@ struct gsm_subscriber_connection {
 
 	struct {
 		int failures;
-		struct penalty_timers *penalty_timers;
+		struct llist_head penalty_timers;
 	} hodec2;
 
 	/* "Codec List (MSC Preferred)" as received by the BSSAP Assignment Request. 3GPP 48.008
@@ -325,6 +398,12 @@ struct gsm_subscriber_connection {
 
 	struct gsm48_classmark3 cm3;
 	bool cm3_valid;
+
+	struct {
+		bool allowed; /* Is fast return to LTE allowed once the conn is released? */
+		bool last_eutran_plmn_valid; /* Is information stored in field below available? */
+		struct osmo_plmn_id last_eutran_plmn;
+	} fast_return;
 };
 
 
@@ -349,6 +428,11 @@ enum gsm_chreq_reason_t {
 	GSM_CHREQ_REASON_PDCH,
 };
 
+static inline bool gsm_chreq_reason_is_voicecall(enum gsm_chreq_reason_t reason)
+{
+	return reason == GSM_CHREQ_REASON_EMERG || reason == GSM_CHREQ_REASON_CALL;
+}
+
 /* lchans 0..3 are SDCCH in combined channel configuration,
    use 4 as magic number for BCCH hack - see osmo-bts-../oml.c:opstart_compl() */
 #define CCCH_LCHAN 4
@@ -471,18 +555,6 @@ struct amr_multirate_conf {
 };
 /* /BTS ONLY */
 
-enum lchan_csd_mode {
-	LCHAN_CSD_M_NT,
-	LCHAN_CSD_M_T_1200_75,
-	LCHAN_CSD_M_T_600,
-	LCHAN_CSD_M_T_1200,
-	LCHAN_CSD_M_T_2400,
-	LCHAN_CSD_M_T_9600,
-	LCHAN_CSD_M_T_14400,
-	LCHAN_CSD_M_T_29000,
-	LCHAN_CSD_M_T_32000,
-};
-
 /* State of the SAPIs in the lchan */
 enum lchan_sapi_state {
 	LCHAN_SAPI_S_NONE,
@@ -498,6 +570,8 @@ struct gsm_encr {
 	uint8_t alg_id;
 	uint8_t key_len;
 	uint8_t key[MAX_A5_KEY_LEN];
+	bool kc128_present;
+	uint8_t kc128[16];
 };
 
 #define LOGPLCHAN(lchan, ss, level, fmt, args...) \
@@ -508,63 +582,47 @@ struct gsm_encr {
 	     bsc_subscr_name(lchan && lchan->conn ? lchan->conn->bsub : NULL), \
 	     ## args)
 
-/* Iterate lchans that have an FSM allocated based based on explicit pchan kind
- * (GSM_PCHAN_* constant).
- * Remark: PDCH related lchans are not handled in BSC but in PCU, so trying to
- * 	  iterate through GSM_PCHAN_PDCH is considered a void loop.
- */
-#define ts_as_pchan_for_each_lchan(lchan, ts, as_pchan) \
-	for (lchan = (ts)->lchan; \
-	     ((lchan - (ts)->lchan) < ARRAY_SIZE((ts)->lchan)) \
-	     && lchan->fi \
-	     && lchan->nr < pchan_subslots(as_pchan); \
-	     lchan++)
-
-/* Iterate lchans that have an FSM allocated based on current PCHAN
- * mode set in \ref ts.
+/* Iterate at most N lchans of the given timeslot.
  * usage:
  * struct gsm_lchan *lchan;
  * struct gsm_bts_trx_ts *ts = get_some_timeslot();
- * ts_for_each_lchan(lchan, ts) {
- * 	LOGPLCHAN(DMAIN, LOGL_DEBUG, "hello world\n");
+ * ts_for_n_lchans(lchan, ts, 3) {
+ *         LOG_LCHAN(lchan, LOGL_DEBUG, "hello world\n");
  * }
  */
-#define ts_for_each_lchan(lchan, ts) ts_as_pchan_for_each_lchan(lchan, ts, (ts)->pchan_is)
+#define ts_for_n_lchans(lchan, ts, N) \
+	for (lchan = (ts)->lchan; \
+	     ((lchan - (ts)->lchan) < ARRAY_SIZE((ts)->lchan)) \
+	     && lchan->fi \
+	     && ((lchan - (ts)->lchan) < (N)); \
+	     lchan++)
 
-/* Iterate over all possible lchans available that have an FSM allocated based
- * on PCHAN \ref ts (dynamic) configuration.
- * Iterate all lchan instances set up by this \ref ts type, including those
- * lchans currently disabled or in process of being enabled (e.g. due to dynamic
- * timeslot in switchover). Compare ts_for_each_lchan(), which iterates only the
- * enabled lchans.
- * For example, it is useful in case dynamic timeslot \ref ts is in process of
- * switching from configuration PDCH (no lchans) to TCH_F (1 lchan), where
- * pchan_is is still set to PDCH but \ref ts may contain already an \ref lchan
- * of type TCH_F which initiated the request to switch the \ts configuration.
- */
-#define ts_for_each_potential_lchan(lchan, ts) ts_as_pchan_for_each_lchan(lchan, ts, (ts)->pchan_on_init)
-
-enum lchan_activate_mode {
-	FOR_NONE,
-	FOR_MS_CHANNEL_REQUEST,
-	FOR_ASSIGNMENT,
-	FOR_HANDOVER,
-	FOR_VTY,
+enum lchan_activate_for {
+	ACTIVATE_FOR_NONE,
+	ACTIVATE_FOR_MS_CHANNEL_REQUEST,
+	ACTIVATE_FOR_ASSIGNMENT,
+	ACTIVATE_FOR_HANDOVER,
+	ACTIVATE_FOR_VTY,
+	ACTIVATE_FOR_MODE_MODIFY_RTP,
 };
 
 extern const struct value_string lchan_activate_mode_names[];
-static inline const char *lchan_activate_mode_name(enum lchan_activate_mode activ_for)
+static inline const char *lchan_activate_mode_name(enum lchan_activate_for activ_for)
 { return get_value_string(lchan_activate_mode_names, activ_for); }
 
+enum imm_ass_time {
+	IMM_ASS_TIME_POST_CHAN_ACK = 0,
+	IMM_ASS_TIME_PRE_CHAN_ACK,
+	IMM_ASS_TIME_PRE_TS_ACK,
+};
+
 struct lchan_activate_info {
-	enum lchan_activate_mode activ_for;
+	enum lchan_activate_for activ_for;
+	/* If activ_for == ACTIVATE_FOR_MS_CHANNEL_REQUEST, the original CHREQ reason. */
+	enum gsm_chreq_reason_t chreq_reason;
 	struct gsm_subscriber_connection *for_conn;
-	/* This always is for a specific lchan, so its lchan->type indicates full or half rate.
-	 * When a dyn TS was selected, the lchan->type has been set to the desired rate. */
-	enum gsm48_chan_mode chan_mode;
+	struct channel_mode_and_rate ch_mode_rate;
 	struct gsm_encr encr;
-	/* AMR config */
-	uint16_t s15_s0;
 	bool requires_voice_stream;
 	bool wait_before_switching_rtp; /*< true = requires LCHAN_EV_READY_TO_SWITCH_RTP */
 	uint16_t msc_assigned_cic;
@@ -573,6 +631,70 @@ struct lchan_activate_info {
 	struct gsm_lchan *re_use_mgw_endpoint_from_lchan;
 	bool ta_known;
 	uint8_t ta;
+
+	/* The TSC Set to use if 'use' is true, otherwise automatically determine the TSC Set value to use. Valid range
+	 * is 1 to 4, as described in 3GPP TS 45.002. */
+	struct optional_val tsc_set;
+	/* The TSC to use if 'use' is true, otherwise automatically determine the TSC value to use. Valid range is 0 to
+	 * 7, as described in 3GPP TS 45.002. */
+	struct optional_val tsc;
+
+	bool vamos;
+
+	/* A copy of bts->imm_ass_time at the time where Channel Activation was requested. A change in the VTY
+	 * configuration has immediate effect on the value, so make sure we don't get mixed up when it gets changed
+	 * while a channel activation is in progress. */
+	enum imm_ass_time imm_ass_time;
+};
+
+enum lchan_modify_for {
+	MODIFY_FOR_NONE,
+	MODIFY_FOR_ASSIGNMENT,
+	MODIFY_FOR_VTY,
+};
+
+extern const struct value_string lchan_modify_for_names[];
+static inline const char *lchan_modify_for_name(enum lchan_modify_for modify_for)
+{ return get_value_string(lchan_modify_for_names, modify_for); }
+
+struct lchan_modify_info {
+	enum lchan_modify_for modify_for;
+	struct channel_mode_and_rate ch_mode_rate;
+	bool requires_voice_stream;
+	uint16_t msc_assigned_cic;
+
+	/* The TSC Set to use if 'use' is true, otherwise automatically determine the TSC Set value to use. Valid range
+	 * is 1 to 4, as described in 3GPP TS 45.002. */
+	struct optional_val tsc_set;
+	/* The TSC to use if 'use' is true, otherwise automatically determine the TSC value to use. Valid range is 0 to
+	 * 7, as described in 3GPP TS 45.002. */
+	struct optional_val tsc;
+
+	bool vamos;
+};
+
+#define INTERF_DBM_UNKNOWN 0
+#define INTERF_BAND_UNKNOWN 0xff
+
+/* Measurement pre-processing state */
+struct gsm_power_ctrl_meas_proc_state {
+	/* Number of measurements processed */
+	unsigned int meas_num;
+	/* Algorithm specific data */
+	union {
+		struct {
+			/* Scaled up 100 times average value */
+			int Avg100;
+		} ewma;
+	};
+};
+
+struct lchan_power_ctrl_state {
+	/* Measurement pre-processing state (for dynamic mode) */
+	struct gsm_power_ctrl_meas_proc_state rxlev_meas_proc;
+	struct gsm_power_ctrl_meas_proc_state rxqual_meas_proc;
+	/* Number of SACCH blocks to skip (for dynamic mode) */
+	int skip_block_num;
 };
 
 struct gsm_lchan {
@@ -589,7 +711,15 @@ struct gsm_lchan {
 	struct osmo_mgcpc_ep_ci *mgw_endpoint_ci_bts;
 
 	struct {
+		/* The request as made by the caller, see lchan_activate().
+		 * lchan->activate.info is treated immutable: remains unchanged throughout the Activation.
+		 * The mutable versions are below: some values need automatic adjustments, in which case they are copied
+		 * from immutable lchan->activate.info.* to lchan->activate.*, for example lchan->activate.ch_mode_rate
+		 * is initially a copy of lchan->activate.info.ch_mode_rate, and is possibly adjusted afterwards. */
 		struct lchan_activate_info info;
+
+		struct channel_mode_and_rate ch_mode_rate;
+		struct gsm48_multi_rate_conf mr_conf_filtered;
 		bool activ_ack; /*< true as soon as RSL Chan Activ Ack is received */
 		bool immediate_assignment_sent;
 		/*! This flag ensures that when an lchan activation has succeeded, and we have already
@@ -597,14 +727,37 @@ struct gsm_lchan {
 		 * occur later, e.g. during release, that we don't send a NACK out of context. */
 		bool concluded;
 		enum gsm0808_cause gsm0808_error_cause;
+		/* Actually used TSC Set. */
+		int tsc_set;
+		/* Actually used TSC. */
+		uint8_t tsc;
 	} activate;
 
+	struct {
+		/* The request as made by the caller, see lchan_mode_modify().
+		 * lchan->modify.info is treated immutable: remains unchanged throughout the Mode Modify.
+		 * The mutable versions are below: some values need automatic adjustments, in which case they are copied
+		 * from immutable lchan->modify.info.* to lchan->modify.*, for example lchan->modify.ch_mode_rate
+		 * is initially a copy of lchan->modify.info.ch_mode_rate, and is possibly adjusted afterwards. */
+		struct lchan_modify_info info;
+
+		struct channel_mode_and_rate ch_mode_rate;
+		struct gsm48_multi_rate_conf mr_conf_filtered;
+		/* Actually used TSC Set. */
+		int tsc_set;
+		/* Actually used TSC. */
+		uint8_t tsc;
+		bool concluded;
+	} modify;
+
 	struct {
 		/* If an event to release the lchan comes in while still waiting for responses, just mark this
 		 * flag, so that the lchan will gracefully release at the next sensible junction. */
 		bool requested;
 		bool do_rr_release;
 		enum gsm48_rr_cause rr_cause;
+		bool last_eutran_plmn_valid;
+		struct osmo_plmn_id last_eutran_plmn;
 
 		/* There is an RSL error cause of value 0, so we need a separate flag. */
 		bool in_error;
@@ -614,30 +767,16 @@ struct gsm_lchan {
 		/* If a release event is being handled, ignore other ricocheting release events until that
 		 * release handling has concluded. */
 		bool in_release_handler;
-
-		/* is this release at the end of a CSFB call? */
-		bool is_csfb;
 	} release;
 
 	/* The logical channel type */
 	enum gsm_chan_t type;
-	/* RSL channel mode */
-	enum rsl_cmod_spd rsl_cmode;
-	/* If TCH, traffic channel mode */
-	enum gsm48_chan_mode tch_mode;
-	enum lchan_csd_mode csd_mode;
 	/* Power levels for MS and BTS */
-	uint8_t bs_power;
+	uint8_t bs_power_db;
 	uint8_t ms_power;
 	/* Encryption information */
 	struct gsm_encr encr;
 
-	/* AMR bits */
-	uint8_t mr_ms_lv[7];
-	uint8_t mr_bts_lv[7];
-	/* AMR bits were based on these rate bits: */
-	uint16_t s15_s0;
-
 	/* Established data link layer services */
 	uint8_t sapis[8];
 
@@ -678,12 +817,35 @@ struct gsm_lchan {
 
 	struct gsm_subscriber_connection *conn;
 
-	/* Depending on the preferences that where submitted together with
-	 * the assignment and the current channel load, the BSC has to select
-	 * one of the offered codec/rates. The final selection by the BSC is
-	 * stored here and is used when sending the assignment complete or
-	 * when performing a handover procedure. */
-	struct channel_mode_and_rate ch_mode_rate;
+	/* After the Channel Activation ACK or RSL Mode Modify ACK is received, this reflects the actually used
+	 * channel_mode_and_rate. */
+	struct channel_mode_and_rate current_ch_mode_rate;
+	struct gsm48_multi_rate_conf current_mr_conf;
+
+	/* Circuit-Switched TSC Set in use, or -1 if no specific TSC Set was requested. The valid range is 1-4 as
+	 * described in the spec 3GPP TS 45.002. */
+	int tsc_set;
+	/* Training Sequence Code in use. The valid range is 0-7 as described in the spec 3GPP TS 45.002. */
+	uint8_t tsc;
+
+	struct {
+		/* Whether this lchan represents a secondary "shadow" lchan to multiplex a second MS onto a primary
+		 * "normal" lchan */
+		bool is_secondary;
+
+		/* Whether this lchan is activated/modified into a mode that allows VAMOS multiplexing at this moment */
+		bool enabled;
+	} vamos;
+
+	/* dBm value of interference level as reported in the most recent Resource Indication, if any for this lchan. Or
+	 * INTERF_DBM_UNKNOWN if this lchan was not included in the most recent Resource Indication.
+	 * The range is typically -115 to -85 dBm, here stored 1:1 as a signed integer, to ease comparison. */
+	int16_t interf_dbm;
+	/* Actual reported interference band index, or INTERF_BAND_UNKNOWN if this lchan was not included in the most
+	 * recent Resource Indication. */
+	uint8_t interf_band;
+	/* MS power control state */
+	struct lchan_power_ctrl_state ms_power_ctrl;
 };
 
 /* One Timeslot in a TRX */
@@ -699,7 +861,7 @@ struct gsm_bts_trx_ts {
 	 * vty after OML activation. Gets written on vty 'write file'. */
 	enum gsm_phys_chan_config pchan_from_config;
 	/* When the timeslot OML is established, pchan_from_config is copied here. This is the pchan
-	 * currently in effect; for dynamic ts, this is the dyn kind (GSM_PCHAN_TCH_F_TCH_H_PDCH or
+	 * currently in effect; for dynamic ts, this is the dyn kind (GSM_PCHAN_OSMO_DYN or
 	 * GSM_PCHAN_TCH_F_PDCH) and does not show the pchan type currently active. */
 	enum gsm_phys_chan_config pchan_on_init;
 	/* This is the *actual* pchan type currently active. For dynamic timeslots, this reflects either
@@ -744,6 +906,15 @@ struct gsm_bts_trx_ts {
 		} rbs2000;
 	};
 
+	/* Maximum BCCH carrier power reduction */
+	uint8_t c0_max_power_red_db;
+
+	/* Maximum number of lchans that could become usable, for example by switching a dynamic timeslot's type or by
+	 * enabling VAMOS secondary lchans. This does include the maximum count of possible VAMOS secondary lchans. */
+	uint8_t max_lchans_possible;
+	/* Currently usable lchans, according to the current pchan mode (for dynamic timeslots, this may change).
+	 * Does not include count of secondary VAMOS lchans. */
+	uint8_t max_primary_lchans;
 	struct gsm_lchan lchan[TS_MAX_LCHAN];
 };
 
@@ -846,12 +1017,6 @@ struct load_counter {
 	unsigned int used;
 };
 
-/* Useful to track N-N relations between BTS, for example neighbors. */
-struct gsm_bts_ref {
-	struct llist_head entry;
-	struct gsm_bts *bts;
-};
-
 /* A single Page of a SMSCB message */
 struct bts_smscb_page {
 	/* SMSCB message we're part of */
@@ -937,7 +1102,7 @@ const char *gsm_lchant_name(enum gsm_chan_t c);
 const char *gsm_chreq_name(enum gsm_chreq_reason_t c);
 char *gsm_ts_name(const struct gsm_bts_trx_ts *ts);
 char *gsm_ts_and_pchan_name(const struct gsm_bts_trx_ts *ts);
-char *gsm_lchan_name_compute(const struct gsm_lchan *lchan);
+void lchan_update_name(struct gsm_lchan *lchan);
 
 static inline char *gsm_lchan_name(const struct gsm_lchan *lchan)
 {
@@ -956,170 +1121,33 @@ void *
 gsm_objclass2obj(struct gsm_bts *bts, uint8_t obj_class,
 	     const struct abis_om_obj_inst *obj_inst);
 
-uint8_t gsm_pchan2chan_nr(enum gsm_phys_chan_config pchan,
-			  uint8_t ts_nr, uint8_t lchan_nr);
-uint8_t gsm_lchan2chan_nr(const struct gsm_lchan *lchan);
-uint8_t gsm_lchan_as_pchan2chan_nr(const struct gsm_lchan *lchan,
-				   enum gsm_phys_chan_config as_pchan);
+int gsm_pchan2chan_nr(enum gsm_phys_chan_config pchan,
+		      uint8_t ts_nr, uint8_t lchan_nr, bool vamos_is_secondary);
+int gsm_lchan2chan_nr(const struct gsm_lchan *lchan, bool allow_osmo_cbits);
+int gsm_lchan_and_pchan2chan_nr(const struct gsm_lchan *lchan, enum gsm_phys_chan_config pchan, bool allow_osmo_cbits);
 
-void gsm48_lchan2chan_desc(struct gsm48_chan_desc *cd,
-			   const struct gsm_lchan *lchan);
-void gsm48_lchan2chan_desc_as_configured(struct gsm48_chan_desc *cd, const struct gsm_lchan *lchan);
+int gsm48_lchan2chan_desc(struct gsm48_chan_desc *cd,
+			  const struct gsm_lchan *lchan,
+			  uint8_t tsc, bool allow_osmo_cbits);
+int gsm48_lchan_and_pchan2chan_desc(struct gsm48_chan_desc *cd,
+				    const struct gsm_lchan *lchan,
+				    enum gsm_phys_chan_config pchan,
+				    uint8_t tsc, bool allow_osmo_cbits);
 
 uint8_t gsm_ts_tsc(const struct gsm_bts_trx_ts *ts);
 
-enum gsm_phys_chan_config ts_pchan(struct gsm_bts_trx_ts *ts);
 uint8_t pchan_subslots(enum gsm_phys_chan_config pchan);
+uint8_t pchan_subslots_vamos(enum gsm_phys_chan_config pchan);
 bool ts_is_tch(struct gsm_bts_trx_ts *ts);
 
+struct gsm_lchan *gsm_lchan_vamos_to_primary(const struct gsm_lchan *lchan_vamos);
+struct gsm_lchan *gsm_lchan_primary_to_vamos(const struct gsm_lchan *lchan_primary);
 
 struct gsm_bts *conn_get_bts(struct gsm_subscriber_connection *conn);
 
 void conn_update_ms_power_class(struct gsm_subscriber_connection *conn, uint8_t power_class);
 void lchan_update_ms_power_ctrl_level(struct gsm_lchan *lchan, int ms_power_dbm);
 
-enum {
-	BSC_CTR_ASSIGNMENT_ATTEMPTED,
-	BSC_CTR_ASSIGNMENT_COMPLETED,
-	BSC_CTR_ASSIGNMENT_STOPPED,
-	BSC_CTR_ASSIGNMENT_NO_CHANNEL,
-	BSC_CTR_ASSIGNMENT_TIMEOUT,
-	BSC_CTR_ASSIGNMENT_FAILED,
-	BSC_CTR_ASSIGNMENT_ERROR,
-	BSC_CTR_HANDOVER_ATTEMPTED,
-	BSC_CTR_HANDOVER_COMPLETED,
-	BSC_CTR_HANDOVER_STOPPED,
-	BSC_CTR_HANDOVER_NO_CHANNEL,
-	BSC_CTR_HANDOVER_TIMEOUT,
-	BSC_CTR_HANDOVER_FAILED,
-	BSC_CTR_HANDOVER_ERROR,
-	BSC_CTR_INTRA_CELL_HO_ATTEMPTED,
-	BSC_CTR_INTRA_CELL_HO_COMPLETED,
-	BSC_CTR_INTRA_CELL_HO_STOPPED,
-	BSC_CTR_INTRA_CELL_HO_NO_CHANNEL,
-	BSC_CTR_INTRA_CELL_HO_TIMEOUT,
-	BSC_CTR_INTRA_CELL_HO_FAILED,
-	BSC_CTR_INTRA_CELL_HO_ERROR,
-	BSC_CTR_INTRA_BSC_HO_ATTEMPTED,
-	BSC_CTR_INTRA_BSC_HO_COMPLETED,
-	BSC_CTR_INTRA_BSC_HO_STOPPED,
-	BSC_CTR_INTRA_BSC_HO_NO_CHANNEL,
-	BSC_CTR_INTRA_BSC_HO_TIMEOUT,
-	BSC_CTR_INTRA_BSC_HO_FAILED,
-	BSC_CTR_INTRA_BSC_HO_ERROR,
-	BSC_CTR_INTER_BSC_HO_OUT_ATTEMPTED,
-	BSC_CTR_INTER_BSC_HO_OUT_COMPLETED,
-	BSC_CTR_INTER_BSC_HO_OUT_STOPPED,
-	BSC_CTR_INTER_BSC_HO_OUT_TIMEOUT,
-	BSC_CTR_INTER_BSC_HO_OUT_FAILED,
-	BSC_CTR_INTER_BSC_HO_OUT_ERROR,
-	BSC_CTR_INTER_BSC_HO_IN_ATTEMPTED,
-	BSC_CTR_INTER_BSC_HO_IN_COMPLETED,
-	BSC_CTR_INTER_BSC_HO_IN_STOPPED,
-	BSC_CTR_INTER_BSC_HO_IN_NO_CHANNEL,
-	BSC_CTR_INTER_BSC_HO_IN_FAILED,
-	BSC_CTR_INTER_BSC_HO_IN_TIMEOUT,
-	BSC_CTR_INTER_BSC_HO_IN_ERROR,
-	BSC_CTR_PAGING_ATTEMPTED,
-	BSC_CTR_PAGING_DETACHED,
-	BSC_CTR_PAGING_RESPONDED,
-	BSC_CTR_PAGING_NO_ACTIVE_PAGING,
-	BSC_CTR_UNKNOWN_UNIT_ID,
-	BSC_CTR_MSCPOOL_SUBSCR_NO_MSC,
-	BSC_CTR_MSCPOOL_EMERG_FORWARDED,
-	BSC_CTR_MSCPOOL_EMERG_LOST,
-};
-
-static const struct rate_ctr_desc bsc_ctr_description[] = {
-	[BSC_CTR_ASSIGNMENT_ATTEMPTED] =	{"assignment:attempted", "Assignment attempts"},
-	[BSC_CTR_ASSIGNMENT_COMPLETED] =	{"assignment:completed", "Assignment completed"},
-	[BSC_CTR_ASSIGNMENT_STOPPED] = 		{"assignment:stopped", "Connection ended during Assignment"},
-	[BSC_CTR_ASSIGNMENT_NO_CHANNEL] = 	{"assignment:no_channel", "Failure to allocate lchan for Assignment"},
-	[BSC_CTR_ASSIGNMENT_TIMEOUT] = 		{"assignment:timeout", "Assignment timed out"},
-	[BSC_CTR_ASSIGNMENT_FAILED] = 		{"assignment:failed", "Received Assignment Failure message"},
-	[BSC_CTR_ASSIGNMENT_ERROR] = 		{"assignment:error", "Assignment failed for other reason"},
-
-	[BSC_CTR_HANDOVER_ATTEMPTED] = 		{"handover:attempted", "Intra-BSC handover attempts"},
-	[BSC_CTR_HANDOVER_COMPLETED] = 		{"handover:completed", "Intra-BSC handover completed"},
-	[BSC_CTR_HANDOVER_STOPPED] = 		{"handover:stopped", "Connection ended during HO"},
-	[BSC_CTR_HANDOVER_NO_CHANNEL] = 	{"handover:no_channel", "Failure to allocate lchan for HO"},
-	[BSC_CTR_HANDOVER_TIMEOUT] = 		{"handover:timeout", "Handover timed out"},
-	[BSC_CTR_HANDOVER_FAILED] = 		{"handover:failed", "Received Handover Fail messages"},
-	[BSC_CTR_HANDOVER_ERROR] = 		{"handover:error", "Re-assignment failed for other reason"},
-
-	[BSC_CTR_INTRA_CELL_HO_ATTEMPTED] = 	{"intra_cell_ho:attempted", "Intra-Cell handover attempts"},
-	[BSC_CTR_INTRA_CELL_HO_COMPLETED] = 	{"intra_cell_ho:completed", "Intra-Cell handover completed"},
-	[BSC_CTR_INTRA_CELL_HO_STOPPED] = 	{"intra_cell_ho:stopped", "Connection ended during HO"},
-	[BSC_CTR_INTRA_CELL_HO_NO_CHANNEL] = 	{"intra_cell_ho:no_channel", "Failure to allocate lchan for HO"},
-	[BSC_CTR_INTRA_CELL_HO_TIMEOUT] = 	{"intra_cell_ho:timeout", "Handover timed out"},
-	[BSC_CTR_INTRA_CELL_HO_FAILED] = 	{"intra_cell_ho:failed", "Received Handover Fail messages"},
-	[BSC_CTR_INTRA_CELL_HO_ERROR] = 	{"intra_cell_ho:error", "Re-assignment failed for other reason"},
-
-	[BSC_CTR_INTRA_BSC_HO_ATTEMPTED] = 	{"intra_bsc_ho:attempted", "Intra-BSC handover attempts"},
-	[BSC_CTR_INTRA_BSC_HO_COMPLETED] = 	{"intra_bsc_ho:completed", "Intra-BSC handover completed"},
-	[BSC_CTR_INTRA_BSC_HO_STOPPED] = 	{"intra_bsc_ho:stopped", "Connection ended during HO"},
-	[BSC_CTR_INTRA_BSC_HO_NO_CHANNEL] = 	{"intra_bsc_ho:no_channel", "Failure to allocate lchan for HO"},
-	[BSC_CTR_INTRA_BSC_HO_TIMEOUT] = 	{"intra_bsc_ho:timeout", "Handover timed out"},
-	[BSC_CTR_INTRA_BSC_HO_FAILED] = 	{"intra_bsc_ho:failed", "Received Handover Fail messages"},
-	[BSC_CTR_INTRA_BSC_HO_ERROR] = 		{"intra_bsc_ho:error", "Re-assignment failed for other reason"},
-
-	[BSC_CTR_INTER_BSC_HO_OUT_ATTEMPTED] =	{"interbsc_ho_out:attempted",
-						 "Attempts to handover to remote BSS"},
-	[BSC_CTR_INTER_BSC_HO_OUT_COMPLETED] =	{"interbsc_ho_out:completed",
-						 "Handover to remote BSS completed"},
-	[BSC_CTR_INTER_BSC_HO_OUT_STOPPED] =	{"interbsc_ho_out:stopped", "Connection ended during HO"},
-	[BSC_CTR_INTER_BSC_HO_OUT_TIMEOUT] =	{"interbsc_ho_out:timeout", "Handover timed out"},
-	[BSC_CTR_INTER_BSC_HO_OUT_FAILED] =	{"interbsc_ho_out:failed", "Received Handover Fail message"},
-	[BSC_CTR_INTER_BSC_HO_OUT_ERROR] =	{"interbsc_ho_out:error",
-						 "Handover to remote BSS failed for other reason"},
-
-	[BSC_CTR_INTER_BSC_HO_IN_ATTEMPTED] =	{"interbsc_ho_in:attempted",
-						 "Attempts to handover from remote BSS"},
-	[BSC_CTR_INTER_BSC_HO_IN_COMPLETED] =	{"interbsc_ho_in:completed",
-						 "Handover from remote BSS completed"},
-	[BSC_CTR_INTER_BSC_HO_IN_STOPPED] =	{"interbsc_ho_in:stopped", "Connection ended during HO"},
-	[BSC_CTR_INTER_BSC_HO_IN_NO_CHANNEL] =	{"interbsc_ho_in:no_channel",
-						 "Failure to allocate lchan for HO"},
-	[BSC_CTR_INTER_BSC_HO_IN_TIMEOUT] =	{"interbsc_ho_in:timeout", "Handover from remote BSS timed out"},
-	[BSC_CTR_INTER_BSC_HO_IN_FAILED] =	{"interbsc_ho_in:failed", "Received Handover Fail message"},
-	[BSC_CTR_INTER_BSC_HO_IN_ERROR] =	{"interbsc_ho_in:error",
-						 "Handover from remote BSS failed for other reason"},
-
-	[BSC_CTR_PAGING_ATTEMPTED] = 		{"paging:attempted", "Paging attempts for a subscriber"},
-	[BSC_CTR_PAGING_DETACHED] = 		{"paging:detached", "Paging request send failures because no responsible BTS was found"},
-	[BSC_CTR_PAGING_RESPONDED] = 		{"paging:responded", "Paging attempts with successful response"},
-	[BSC_CTR_PAGING_NO_ACTIVE_PAGING] =	{"paging:no_active_paging", "Paging response without an active paging request (arrived after paging expiration?)"},
-
-	[BSC_CTR_UNKNOWN_UNIT_ID] = 		{"abis:unknown_unit_id", "Connection attempts from unknown IPA CCM Unit ID"},
-
-	[BSC_CTR_MSCPOOL_SUBSCR_NO_MSC] =	{"mscpool:subscr:no_msc",
-						 "Complete Layer 3 requests lost because no connected MSC is found available"},
-	[BSC_CTR_MSCPOOL_EMERG_FORWARDED] =	{"mscpool:emerg:forwarded",
-						 "Emergency call requests forwarded to an MSC (see also per-MSC counters"},
-	[BSC_CTR_MSCPOOL_EMERG_LOST] =		{"mscpool:emerg:lost",
-						 "Emergency call requests lost because no MSC was found available"},
-};
-
-
-
-static const struct rate_ctr_group_desc bsc_ctrg_desc = {
-	"bsc",
-	"base station controller",
-	OSMO_STATS_CLASS_GLOBAL,
-	ARRAY_SIZE(bsc_ctr_description),
-	bsc_ctr_description,
-};
-
-/* Constants for the BSC stats */
-enum {
-	BSC_STAT_NUM_BTS_TOTAL,
-};
-
-/* BTS counter index if a BTS could not be found
- * Currently we are limited to bts 0 - 255 in the VTY, but that might change in
- * the future so use 2**16 */
-#define BTS_STAT_IDX_UNKNOWN (UINT16_MAX + 1)
-
 struct gsm_tz {
 	int override; /* if 0, use system's time zone instead. */
 	int hr; /* hour */
@@ -1128,12 +1156,6 @@ struct gsm_tz {
 };
 
 struct gsm_network {
-	/* TODO MSCSPLIT the gsm_network struct is basically a kitchen sink for
-	 * global settings and variables, "madly" mixing BSC and MSC stuff. Split
-	 * this in e.g. struct osmo_bsc and struct osmo_msc, with the things
-	 * these have in common, like country and network code, put in yet
-	 * separate structs and placed as members in osmo_bsc and osmo_msc. */
-
 	struct osmo_plmn_id plmn;
 
 	/* bit-mask of permitted encryption algorithms. LSB=A5/0, MSB=A5/7 */
@@ -1185,7 +1207,7 @@ struct gsm_network {
 	/* control interface */
 	struct ctrl_handle *ctrl;
 
-	/* Allow or disallow TCH/F on dynamic TCH/F_TCH/H_PDCH; OS#1778 */
+	/* Allow or disallow TCH/F on dynamic TCH/F_TCH/H_SDCCH8_PDCH; OS#1778 */
 	bool dyn_ts_allow_tch_f;
 
 	/* all active subscriber connections. */
@@ -1209,14 +1231,22 @@ struct gsm_network {
 	/* Timer for periodic channel load measurements to maintain each BTS's T3122. */
 	struct osmo_timer_list t3122_chan_load_timer;
 
+	/* Timer to write each BTS's uptime counter state to the stats system. */
+	struct osmo_timer_list bts_store_uptime_timer;
+
 	struct {
+		/* Single MGCP client configuration under msc node (also required for
+		 * MGCP proxy when sccp-lite is used) */
 		struct mgcp_client_conf *conf;
-		struct mgcp_client *client;
+
+		/* MGW pool, also includes the single MGCP client as fallback if no
+		 * pool is configured. */
+		struct mgcp_client_pool *mgw_pool;
+
+		/* Timer definitions, the same for all MGW pool members */
 		struct osmo_tdef *tdefs;
 	} mgw;
 
-	/* Remote BSS Cell Identifier Lists */
-	struct neighbor_ident_list *neighbor_bss_cells;
 	/* Remote BSS resolution sevice (CTRL iface) */
 	struct {
 		char *addr;
@@ -1231,6 +1261,11 @@ struct gsm_network {
 	struct osmo_nri_ranges *null_nri_ranges;
 
 	struct smlc_config *smlc;
+
+	struct osmo_time_cc all_allocated_sdcch;
+	struct osmo_time_cc all_allocated_static_sdcch;
+	struct osmo_time_cc all_allocated_tch;
+	struct osmo_time_cc all_allocated_static_tch;
 };
 
 struct gsm_audio_support {
@@ -1291,6 +1326,9 @@ enum gsm_power_ctrl_meas_avg_algo {
 
 /* MS/BS Power related measurement parameters */
 struct gsm_power_ctrl_meas_params {
+	/* Are these measurement paremeters to be taken into account by loop? */
+	bool enabled;
+
 	/* Thresholds (see 3GPP TS 45.008, section A.3.2.1) */
 	uint8_t lower_thresh; /* lower (decreasing) direction */
 	uint8_t upper_thresh; /* upper (increasing) direction */
@@ -1329,6 +1367,9 @@ enum gsm_power_ctrl_mode {
 	GSM_PWR_CTRL_MODE_STATIC,
 	/* Send MS/BS Power [Parameters] IEs (dynamic mode) */
 	GSM_PWR_CTRL_MODE_DYN_BTS,
+	/* Do not send MS/BS Power IEs and use BSC Power Loop */
+	GSM_PWR_CTRL_MODE_DYN_BSC,
+
 };
 
 /* MS/BS Power Control Parameters */
@@ -1352,8 +1393,29 @@ struct gsm_power_ctrl_params {
 	/* Measurement averaging parameters for RxLev & RxQual */
 	struct gsm_power_ctrl_meas_params rxqual_meas;
 	struct gsm_power_ctrl_meas_params rxlev_meas;
+	/* Measurement averaging parameters for C/I: */
+	struct gsm_power_ctrl_meas_params ci_fr_meas;
+	struct gsm_power_ctrl_meas_params ci_hr_meas;
+	struct gsm_power_ctrl_meas_params ci_amr_fr_meas;
+	struct gsm_power_ctrl_meas_params ci_amr_hr_meas;
+	struct gsm_power_ctrl_meas_params ci_sdcch_meas;
+	struct gsm_power_ctrl_meas_params ci_gprs_meas;
 };
 
 extern const struct gsm_power_ctrl_params power_ctrl_params_def;
+void power_ctrl_params_def_reset(struct gsm_power_ctrl_params *params,
+				 enum gsm_power_ctrl_dir dir);
+
+/* Interference Measurement Parameters */
+struct gsm_interf_meas_params {
+	/* Intave: Interference Averaging period (see 3GPP TS 45.008, table A.1) */
+	uint8_t avg_period; /* number of SACCH multiframes, 1 .. 31 */
+	/* Interference level Boundaries (see 3GPP TS 52.021, section 9.4.25) */
+	uint8_t bounds_dbm[6]; /* -x dBm values for boundaries 0 .. X5 */
+};
+
+extern const struct gsm_interf_meas_params interf_meas_params_def;
+
+enum rsl_cmod_spd chan_mode_to_rsl_cmod_spd(enum gsm48_chan_mode chan_mode);
 
 #endif /* _GSM_DATA_H */
