@@ -594,11 +594,20 @@ static int parse_attr_resp_info_attr(struct gsm_bts *bts, const struct gsm_bts_t
 
 		memcpy(bts->_features_data, TLVP_VAL(tp, NM_ATT_MANUF_ID), len);
 
-		for (i = 0; i < _NUM_BTS_FEAT; i++) {
-			if (osmo_bts_has_feature(&bts->features, i) != osmo_bts_has_feature(&bts->model->features, i)) {
-				LOGPMO(&bts->mo, DNM, LOGL_NOTICE, "Feature '%s' reported via OML does not match statically "
-				       "set feature: %u != %u. Please fix.\n", get_value_string(osmo_bts_features_descs, i),
-				       osmo_bts_has_feature(&bts->features, i), osmo_bts_has_feature(&bts->model->features, i));
+		/* Check each BTS feature in the reported vector */
+		for (i = 0; i < len * 8; i++) {
+			bool Frep = osmo_bts_has_feature(&bts->features, i);
+			if (i >= _NUM_BTS_FEAT && Frep) {
+				LOGPMO(&bts->mo, DNM, LOGL_NOTICE, "Get Attributes Response: "
+				       "unknown feature 0x%02x. Consider upgrading osmo-bsc.\n", i);
+				continue;
+			}
+
+			bool Fexp = osmo_bts_has_feature(&bts->model->features, i);
+			if (!Frep && Fexp) {
+				LOGPMO(&bts->mo, DNM, LOGL_NOTICE, "Get Attributes Response: "
+				       "reported feature '%s' is not supported, while we thought it is.\n",
+				       osmo_bts_features_name(i));
 			}
 		}
 	}
@@ -884,7 +893,7 @@ bool all_trx_rsl_connected_unlocked(const struct gsm_bts *bts)
 	}
 
 	llist_for_each_entry(trx, &bts->trx_list, list) {
-		if (!trx->rsl_link)
+		if (!trx->rsl_link_primary)
 			return false;
 
 		if (!trx_is_usable(trx))
@@ -1883,7 +1892,7 @@ static int verify_chan_comb(struct gsm_bts_trx_ts *ts, uint8_t chan_comb,
 		switch (chan_comb) {
 		case NM_CHANC_TCHHalf:
 		case NM_CHANC_TCHHalf2:
-		case NM_CHANC_OSMO_TCHFull_TCHHalf_PDCH:
+		case NM_CHANC_OSMO_DYN:
 			/* not supported */
 			*reason = "TCH/H is not supported.";
 			return -EINVAL;
@@ -1981,7 +1990,7 @@ static int verify_chan_comb(struct gsm_bts_trx_ts *ts, uint8_t chan_comb,
 				case NM_CHANC_TCHHalf:
 				case NM_CHANC_IPAC_TCHFull_TCHHalf:
 				case NM_CHANC_IPAC_TCHFull_PDCH:
-				case NM_CHANC_OSMO_TCHFull_TCHHalf_PDCH:
+				case NM_CHANC_OSMO_DYN:
 					return 0;
 				default:
 					*reason = "TS1 must carry a CBCH, SDCCH or TCH.";
@@ -2013,7 +2022,7 @@ static int verify_chan_comb(struct gsm_bts_trx_ts *ts, uint8_t chan_comb,
 				return 0;
 			case NM_CHANC_IPAC_PDCH:
 			case NM_CHANC_IPAC_TCHFull_PDCH:
-			case NM_CHANC_OSMO_TCHFull_TCHHalf_PDCH:
+			case NM_CHANC_OSMO_DYN:
 				if (ts->trx->nr == 0)
 					return 0;
 				else {
@@ -3005,32 +3014,29 @@ static void rsl_connect_timeout(void *data)
 int abis_nm_ipaccess_rsl_connect(struct gsm_bts_trx *trx,
 				 uint32_t ip, uint16_t port, uint8_t stream)
 {
+	struct msgb *attr;
 	struct in_addr ia;
-	uint8_t attr[] = { NM_ATT_IPACC_STREAM_ID, 0,
-			    NM_ATT_IPACC_DST_IP_PORT, 0, 0,
-			    NM_ATT_IPACC_DST_IP, 0, 0, 0, 0 };
-
-	int attr_len = sizeof(attr);
 	int error;
 
 	osmo_timer_setup(&trx->rsl_connect_timeout, rsl_connect_timeout, trx);
 
-	ia.s_addr = htonl(ip);
-	attr[1] = stream;
-	attr[3] = port >> 8;
-	attr[4] = port & 0xff;
-	memcpy(attr + 6, &ia.s_addr, sizeof(uint32_t));
+	attr = msgb_alloc(32, "RSL-connect-attr");
+	msgb_tv_put(attr, NM_ATT_IPACC_STREAM_ID, stream);
+	msgb_tv16_put(attr, NM_ATT_IPACC_DST_IP_PORT, port);
 
 	/* if ip == 0, we use the default IP */
-	if (ip == 0)
-		attr_len -= 5;
+	if (ip != 0) {
+		ia.s_addr = htonl(ip);
+		msgb_tv_fixed_put(attr, NM_ATT_IPACC_DST_IP, 4, (void*)&ia.s_addr);
+	}
 
 	LOG_TRX(trx, DNM, LOGL_INFO, "IPA RSL CONNECT IP=%s PORT=%u STREAM=0x%02x\n",
 		inet_ntoa(ia), port, stream);
 
 	error = abis_nm_ipaccess_msg(trx->bts, NM_MT_IPACC_RSL_CONNECT,
 				     NM_OC_BASEB_TRANSC, trx->bts->bts_nr,
-				     trx->nr, 0xff, attr, attr_len);
+				     trx->nr, 0xff, attr->data, attr->len);
+	msgb_free(attr);
 	if (error == 0)
 		osmo_timer_schedule(&trx->rsl_connect_timeout, 60, 0);
 
