@@ -22,6 +22,10 @@
 #include <time.h>
 
 #include <osmocom/ctrl/control_cmd.h>
+
+#include <osmocom/vty/command.h>
+#include <osmocom/vty/misc.h>
+
 #include <osmocom/gsm/gsm48.h>
 #include <osmocom/bsc/ipaccess.h>
 #include <osmocom/bsc/gsm_data.h>
@@ -31,6 +35,79 @@
 #include <osmocom/bsc/osmo_bsc_rf.h>
 #include <osmocom/bsc/bsc_msc_data.h>
 #include <osmocom/bsc/bts.h>
+#include <osmocom/bsc/neighbor_ident.h>
+
+static int verify_net_apply_config_file(struct ctrl_cmd *cmd, const char *value, void *_data)
+{
+	FILE *cfile;
+
+	if (!cmd->value || cmd->value[0] == '\0')
+		return -1;
+
+	cfile = fopen(cmd->value, "r");
+	if (!cfile)
+		return -1;
+
+	fclose(cfile);
+
+	return 0;
+}
+static int set_net_apply_config_file(struct ctrl_cmd *cmd, void *_data)
+{
+	int rc;
+	FILE *cfile;
+	unsigned cmd_ret = CTRL_CMD_ERROR;
+
+	LOGP(DCTRL, LOGL_NOTICE, "Applying VTY snippet from %s...\n", cmd->value);
+	cfile = fopen(cmd->value, "r");
+	if (!cfile) {
+		LOGP(DCTRL, LOGL_NOTICE, "Applying VTY snippet from %s: fopen() failed: %d\n",
+		     cmd->value, errno);
+		cmd->reply = "NoFile";
+		return cmd_ret;
+	}
+
+	rc = vty_read_config_filep(cfile, NULL);
+	LOGP(DCTRL, LOGL_NOTICE, "Applying VTY snippet from %s returned %d\n", cmd->value, rc);
+	if (rc) {
+		cmd->reply = talloc_asprintf(cmd, "ParseError=%d", rc);
+		if (!cmd->reply)
+			cmd->reply = "OOM";
+		goto close_ret;
+	}
+
+	cmd->reply = "OK";
+	cmd_ret = CTRL_CMD_REPLY;
+close_ret:
+	fclose(cfile);
+	return cmd_ret;
+}
+CTRL_CMD_DEFINE_WO(net_apply_config_file, "apply-config-file");
+
+static int verify_net_write_config_file(struct ctrl_cmd *cmd, const char *value, void *_data)
+{
+	return 0;
+}
+static int set_net_write_config_file(struct ctrl_cmd *cmd, void *_data)
+{
+	const char *cfile_name;
+	unsigned cmd_ret = CTRL_CMD_ERROR;
+
+	if (strcmp(cmd->value, "overwrite"))
+		host_config_set(cmd->value);
+
+	cfile_name = host_config_file();
+
+	LOGP(DCTRL, LOGL_NOTICE, "Writing VTY config to file %s...\n", cfile_name);
+	if (osmo_vty_write_config_file(cfile_name) < 0)
+		goto ret;
+
+	cmd->reply = "OK";
+	cmd_ret = CTRL_CMD_REPLY;
+ret:
+	return cmd_ret;
+}
+CTRL_CMD_DEFINE_WO(net_write_config_file, "write-config-file");
 
 CTRL_CMD_DEFINE(net_mcc, "mcc");
 static int get_net_mcc(struct ctrl_cmd *cmd, void *_data)
@@ -102,7 +179,7 @@ static int set_net_apply_config(struct ctrl_cmd *cmd, void *data)
 		 * so let's us just reboot it. For the sysmoBTS we can just
 		 * restart the process as all state is gone.
 		 */
-		if (!is_sysmobts_v2(bts) && strcmp(cmd->value, "restart") == 0) {
+		if (!is_osmobts(bts) && strcmp(cmd->value, "restart") == 0) {
 			struct gsm_bts_trx *trx;
 			llist_for_each_entry_reverse(trx, &bts->trx_list, list)
 				abis_nm_ipaccess_restart(trx);
@@ -344,6 +421,44 @@ static int get_bts_rf_state(struct ctrl_cmd *cmd, void *data)
 }
 CTRL_CMD_DEFINE_RO(bts_rf_state, "rf_state");
 
+/* Return a list of the states of each TRX for a given BTS.
+ * <bts_nr>,<trx_nr>,<opstate>,<adminstate>,<rf_policy>,<rsl_status>;<bts_nr>,<trx_nr>,...;...;
+ * For details on the string, see bsc_rf_states_c();
+ */
+static int get_bts_rf_states(struct ctrl_cmd *cmd, void *data)
+{
+	struct gsm_bts *bts = cmd->node;
+
+	if (!bts) {
+		cmd->reply = "bts not found.";
+		return CTRL_CMD_ERROR;
+	}
+
+	cmd->reply = bsc_rf_states_of_bts_c(cmd, bts);
+	if (!cmd->reply) {
+		cmd->reply = "OOM.";
+		return CTRL_CMD_ERROR;
+	}
+
+	return CTRL_CMD_REPLY;
+}
+CTRL_CMD_DEFINE_RO(bts_rf_states, "rf_states");
+
+/* Return a list of the states of each TRX for all BTS:
+ * <bts_nr>,<trx_nr>,<opstate>,<adminstate>,<rf_policy>,<rsl_status>;<bts_nr>,<trx_nr>,...;...;
+ * For details on the string, see bsc_rf_states_c();
+ */
+static int get_net_rf_states(struct ctrl_cmd *cmd, void *data)
+{
+	cmd->reply = bsc_rf_states_c(cmd);
+	if (!cmd->reply) {
+		cmd->reply = "OOM.";
+		return CTRL_CMD_ERROR;
+	}
+	return CTRL_CMD_REPLY;
+}
+CTRL_CMD_DEFINE_RO(net_rf_states, "rf_states");
+
 static int get_net_rf_lock(struct ctrl_cmd *cmd, void *data)
 {
 	struct gsm_network *net = cmd->node;
@@ -424,6 +539,37 @@ static int verify_net_rf_lock(struct ctrl_cmd *cmd, const char *value, void *dat
 }
 CTRL_CMD_DEFINE(net_rf_lock, "rf_locked");
 
+static int get_trx_rf_locked(struct ctrl_cmd *cmd, void *data)
+{
+	struct gsm_bts_trx *trx = cmd->node;
+	/* Return rf_locked = 1 only if it is explicitly locked. If it is in shutdown or null state, do not "trick" the
+	 * caller into thinking that sending "rf_locked 0" is necessary to bring the TRX up. */
+	cmd->reply = (trx->mo.nm_state.administrative == NM_STATE_LOCKED) ? "1" : "0";
+	return CTRL_CMD_REPLY;
+}
+
+static int set_trx_rf_locked(struct ctrl_cmd *cmd, void *data)
+{
+	struct gsm_bts_trx *trx = cmd->node;
+	int locked;
+	if (osmo_str_to_int(&locked, cmd->value, 10, 0, 1)) {
+		cmd->reply = "Invalid value";
+		return CTRL_CMD_ERROR;
+	}
+
+	gsm_trx_lock_rf(trx, locked, "ctrl");
+
+	/* Let's not assume the nm FSM has already switched its state, just return the intended rf_locked value. */
+	cmd->reply = locked ? "1" : "0";
+	return CTRL_CMD_REPLY;
+}
+
+static int verify_trx_rf_locked(struct ctrl_cmd *cmd, const char *value, void *data)
+{
+	return osmo_str_to_int(NULL, value, 10, 0, 1);
+}
+CTRL_CMD_DEFINE(trx_rf_locked, "rf_locked");
+
 static int get_net_bts_num(struct ctrl_cmd *cmd, void *data)
 {
 	struct gsm_network *net = cmd->node;
@@ -474,15 +620,180 @@ static int set_trx_max_power(struct ctrl_cmd *cmd, void *_data)
 }
 CTRL_CMD_DEFINE(trx_max_power, "max-power-reduction");
 
+static int verify_bts_c0_power_red(struct ctrl_cmd *cmd, const char *value, void *_data)
+{
+	const int red = atoi(value);
+
+	if (red < 0 || red > 6) {
+		cmd->reply = "Value is out of range";
+		return 1;
+	} else if (red % 2 != 0) {
+		cmd->reply = "Value must be even";
+		return 1;
+	}
+
+	return 0;
+}
+
+static int get_bts_c0_power_red(struct ctrl_cmd *cmd, void *data)
+{
+	const struct gsm_bts *bts = cmd->node;
+
+	cmd->reply = talloc_asprintf(cmd, "%u", bts->c0_max_power_red_db);
+	if (!cmd->reply) {
+		cmd->reply = "OOM.";
+		return CTRL_CMD_ERROR;
+	}
+
+	return CTRL_CMD_REPLY;
+}
+
+static int set_bts_c0_power_red(struct ctrl_cmd *cmd, void *data)
+{
+	struct gsm_bts *bts = cmd->node;
+	const int red = atoi(cmd->value);
+	int rc;
+
+	rc = gsm_bts_set_c0_power_red(bts, red);
+	if (rc == -ENOTSUP) {
+		cmd->reply = "BCCH carrier power reduction is not supported";
+		return CTRL_CMD_ERROR;
+	} else if (rc != 0) {
+		cmd->reply = "Failed to enable BCCH carrier power reduction";
+		return CTRL_CMD_ERROR;
+	}
+
+	return get_bts_c0_power_red(cmd, data);
+}
+
+CTRL_CMD_DEFINE(bts_c0_power_red, "c0-power-reduction");
+
+static int verify_bts_neighbor_list_add_del(struct ctrl_cmd *cmd, const char *value, void *_data)
+{
+	int arfcn;
+
+	if (osmo_str_to_int(&arfcn, value, 10, 0, 1023) < 0) {
+		cmd->reply = "Invalid ARFCN value";
+		return 1;
+	}
+
+	return 0;
+}
+
+static int set_bts_neighbor_list_add_del(struct ctrl_cmd *cmd, void *data, bool add)
+{
+	struct gsm_bts *bts = cmd->node;
+	struct bitvec *bv = &bts->si_common.neigh_list;
+	int arfcn_int;
+	uint16_t arfcn;
+	enum gsm_band unused;
+
+	if (osmo_str_to_int(&arfcn_int, cmd->value, 10, 0, 1023) < 0) {
+		cmd->reply = "Failed to parse ARFCN value";
+		return CTRL_CMD_ERROR;
+	}
+	arfcn = (uint16_t) arfcn_int;
+
+	if (bts->neigh_list_manual_mode == NL_MODE_AUTOMATIC) {
+		cmd->reply = "Neighbor list not in manual mode";
+		return CTRL_CMD_ERROR;
+	}
+
+	if (gsm_arfcn2band_rc(arfcn, &unused) < 0) {
+		cmd->reply = "Invalid arfcn detected";
+		return CTRL_CMD_ERROR;
+	}
+
+	if (add)
+		bitvec_set_bit_pos(bv, arfcn, 1);
+	else
+		bitvec_set_bit_pos(bv, arfcn, 0);
+
+	cmd->reply = "OK";
+	return CTRL_CMD_REPLY;
+}
+
+static int verify_bts_neighbor_list_add(struct ctrl_cmd *cmd, const char *value, void *_data)
+{
+	return verify_bts_neighbor_list_add_del(cmd, value, _data);
+}
+
+static int set_bts_neighbor_list_add(struct ctrl_cmd *cmd, void *data)
+{
+	return set_bts_neighbor_list_add_del(cmd, data, true);
+}
+
+CTRL_CMD_DEFINE_WO(bts_neighbor_list_add, "neighbor-list add");
+
+static int verify_bts_neighbor_list_del(struct ctrl_cmd *cmd, const char *value, void *_data)
+{
+	return verify_bts_neighbor_list_add_del(cmd, value, _data);
+}
+
+static int set_bts_neighbor_list_del(struct ctrl_cmd *cmd, void *data)
+{
+	return set_bts_neighbor_list_add_del(cmd, data, false);
+}
+
+CTRL_CMD_DEFINE_WO(bts_neighbor_list_del, "neighbor-list del");
+
+static int verify_bts_neighbor_list_mode(struct ctrl_cmd *cmd, const char *value, void *_data)
+{
+	if (!strcmp(value, "automatic"))
+		return 0;
+	if (!strcmp(value, "manual"))
+		return 0;
+	if (!strcmp(value, "manual-si5"))
+		return 0;
+
+	cmd->reply = "Invalid mode";
+	return 1;
+}
+
+static int set_bts_neighbor_list_mode(struct ctrl_cmd *cmd, void *data)
+{
+	struct gsm_bts *bts = cmd->node;
+	int mode = NL_MODE_AUTOMATIC;
+
+	if (!strcmp(cmd->value, "automatic"))
+		mode = NL_MODE_AUTOMATIC;
+	else if (!strcmp(cmd->value, "manual"))
+		mode = NL_MODE_MANUAL;
+	else if (!strcmp(cmd->value, "manual-si5"))
+		mode = NL_MODE_MANUAL_SI5SEP;
+
+	switch (mode) {
+	case NL_MODE_MANUAL_SI5SEP:
+	case NL_MODE_MANUAL:
+		/* make sure we clear the current list when switching to
+		 * manual mode */
+		if (bts->neigh_list_manual_mode == 0)
+			memset(&bts->si_common.data.neigh_list, 0, sizeof(bts->si_common.data.neigh_list));
+		break;
+	default:
+		break;
+	}
+
+	bts->neigh_list_manual_mode = mode;
+
+	cmd->reply = "OK";
+	return CTRL_CMD_REPLY;
+}
+
+CTRL_CMD_DEFINE_WO(bts_neighbor_list_mode, "neighbor-list mode");
+
 int bsc_base_ctrl_cmds_install(void)
 {
 	int rc = 0;
+	rc |= ctrl_cmd_install(CTRL_NODE_ROOT, &cmd_net_apply_config_file);
+	rc |= ctrl_cmd_install(CTRL_NODE_ROOT, &cmd_net_write_config_file);
 	rc |= ctrl_cmd_install(CTRL_NODE_ROOT, &cmd_net_mnc);
 	rc |= ctrl_cmd_install(CTRL_NODE_ROOT, &cmd_net_mcc);
 	rc |= ctrl_cmd_install(CTRL_NODE_ROOT, &cmd_net_apply_config);
 	rc |= ctrl_cmd_install(CTRL_NODE_ROOT, &cmd_net_mcc_mnc_apply);
 	rc |= ctrl_cmd_install(CTRL_NODE_ROOT, &cmd_net_rf_lock);
 	rc |= ctrl_cmd_install(CTRL_NODE_ROOT, &cmd_net_bts_num);
+	rc |= ctrl_cmd_install(CTRL_NODE_ROOT, &cmd_net_rf_states);
 
 	rc |= ctrl_cmd_install(CTRL_NODE_BTS, &cmd_bts_lac);
 	rc |= ctrl_cmd_install(CTRL_NODE_BTS, &cmd_bts_ci);
@@ -493,9 +804,17 @@ int bsc_base_ctrl_cmds_install(void)
 	rc |= ctrl_cmd_install(CTRL_NODE_BTS, &cmd_bts_oml_up);
 	rc |= ctrl_cmd_install(CTRL_NODE_BTS, &cmd_bts_gprs_mode);
 	rc |= ctrl_cmd_install(CTRL_NODE_BTS, &cmd_bts_rf_state);
+	rc |= ctrl_cmd_install(CTRL_NODE_BTS, &cmd_bts_rf_states);
+	rc |= ctrl_cmd_install(CTRL_NODE_BTS, &cmd_bts_c0_power_red);
+	rc |= ctrl_cmd_install(CTRL_NODE_BTS, &cmd_bts_neighbor_list_add);
+	rc |= ctrl_cmd_install(CTRL_NODE_BTS, &cmd_bts_neighbor_list_del);
+	rc |= ctrl_cmd_install(CTRL_NODE_BTS, &cmd_bts_neighbor_list_mode);
+
+	rc |= neighbor_ident_ctrl_init();
 
 	rc |= ctrl_cmd_install(CTRL_NODE_TRX, &cmd_trx_max_power);
 	rc |= ctrl_cmd_install(CTRL_NODE_TRX, &cmd_trx_arfcn);
+	rc |= ctrl_cmd_install(CTRL_NODE_TRX, &cmd_trx_rf_locked);
 
 	return rc;
 }
