@@ -32,11 +32,13 @@
 #include <osmocom/gsm/tlv.h>
 #include <osmocom/core/msgb.h>
 #include <osmocom/core/socket.h>
+#include <osmocom/core/stat_item.h>
 #include <osmocom/core/talloc.h>
 #include <osmocom/core/utils.h>
 #include <osmocom/bsc/gsm_data.h>
 #include <osmocom/bsc/abis_nm.h>
 #include <osmocom/bsc/abis_rsl.h>
+#include <osmocom/bsc/abis_osmo.h>
 #include <osmocom/bsc/debug.h>
 #include <osmocom/abis/subchan_demux.h>
 #include <osmocom/gsm/ipa.h>
@@ -49,6 +51,7 @@
 #include <osmocom/bsc/bts.h>
 #include <osmocom/bsc/bts_sm.h>
 #include <osmocom/bsc/nm_common_fsm.h>
+#include <osmocom/bsc/bsc_stats.h>
 
 static int bts_model_nanobts_start(struct gsm_network *net);
 static void bts_model_nanobts_e1line_bind_ops(struct e1inp_line *line);
@@ -525,13 +528,13 @@ find_bts_by_unitid(struct gsm_network *net, uint16_t site_id, uint16_t bts_id)
 /* These are exported because they are used by the VTY interface. */
 void ipaccess_drop_rsl(struct gsm_bts_trx *trx, const char *reason)
 {
-	if (!trx->rsl_link)
+	if (!trx->rsl_link_primary)
 		return;
 
 	LOG_TRX(trx, DLINP, LOGL_NOTICE, "Dropping RSL link: %s\n", reason);
-	e1inp_sign_link_destroy(trx->rsl_link);
-	trx->rsl_link = NULL;
-	osmo_stat_item_dec(trx->bts->bts_statg->items[BTS_STAT_RSL_CONNECTED], 1);
+	e1inp_sign_link_destroy(trx->rsl_link_primary);
+	trx->rsl_link_primary = NULL;
+	osmo_stat_item_dec(osmo_stat_item_group_get_item(trx->bts->bts_statg, BTS_STAT_RSL_CONNECTED), 1);
 
 	if (trx->bts->c0 == trx)
 		paging_flush_bts(trx->bts, NULL);
@@ -555,7 +558,12 @@ void ipaccess_drop_oml(struct gsm_bts *bts, const char *reason)
 	e1inp_sign_link_destroy(bts->oml_link);
 	bts->oml_link = NULL;
 	bts->uptime = 0;
-	osmo_stat_item_dec(bts->bts_statg->items[BTS_STAT_OML_CONNECTED], 1);
+	osmo_stat_item_dec(osmo_stat_item_group_get_item(bts->bts_statg, BTS_STAT_OML_CONNECTED), 1);
+	gsm_bts_stats_reset(bts);
+
+	/* Also drop the associated OSMO link */
+	e1inp_sign_link_destroy(bts->osmo_link);
+	bts->osmo_link = NULL;
 
 	/* we have issues reconnecting RSL, drop everything. */
 	llist_for_each_entry(trx, &bts->trx_list, list) {
@@ -631,7 +639,7 @@ static void ipaccess_sign_link_reject(const struct ipaccess_unit *dev, const str
 	/* Write to log and increase counter */
 	LOGP(DLINP, LOGL_ERROR, "Unable to find BTS configuration for %u/%u/%u, disconnecting\n", site_id, bts_id,
 		trx_id);
-	rate_ctr_inc(&bsc_gsmnet->bsc_ctrs->ctr[BSC_CTR_UNKNOWN_UNIT_ID]);
+	rate_ctr_inc(rate_ctr_group_get_ctr(bsc_gsmnet->bsc_ctrs, BSC_CTR_UNKNOWN_UNIT_ID));
 
 	/* Get remote IP */
 	if (osmo_sock_get_remote_ip(ts->driver.ipaccess.fd.fd, ip, sizeof(ip)))
@@ -678,10 +686,11 @@ ipaccess_sign_link_up(void *unit_data, struct e1inp_line *line,
 	struct e1inp_sign_link *sign_link = NULL;
 	struct timespec tp;
 	int rc;
+	struct e1inp_ts *sign_ts = e1inp_line_ipa_oml_ts(line);
 
 	bts = find_bts_by_unitid(bsc_gsmnet, dev->site_id, dev->bts_id);
 	if (!bts) {
-		ipaccess_sign_link_reject(dev, &line->ts[E1INP_SIGN_OML - 1]);
+		ipaccess_sign_link_reject(dev, sign_ts);
 		return NULL;
 	}
 	DEBUGP(DLINP, "%s: Identified BTS %u/%u/%u\n", e1inp_signtype_name(type),
@@ -701,17 +710,21 @@ ipaccess_sign_link_up(void *unit_data, struct e1inp_line *line,
 
 		/* create new OML link. */
 		sign_link = bts->oml_link =
-			e1inp_sign_link_create(&line->ts[E1INP_SIGN_OML - 1],
+			e1inp_sign_link_create(sign_ts,
 						E1INP_SIGN_OML, bts->c0,
 						bts->oml_tei, 0);
-		rc = clock_gettime(CLOCK_MONOTONIC, &tp);
+		rc = osmo_clock_gettime(CLOCK_MONOTONIC, &tp);
 		bts->uptime = (rc < 0) ? 0 : tp.tv_sec; /* we don't need sub-second precision for uptime */
 		if (!(sign_link->trx->bts->ip_access.flags & OML_UP)) {
 			e1inp_event(sign_link->ts, S_L_INP_TEI_UP,
 					sign_link->tei, sign_link->sapi);
 			sign_link->trx->bts->ip_access.flags |= OML_UP;
 		}
-		osmo_stat_item_inc(bts->bts_statg->items[BTS_STAT_OML_CONNECTED], 1);
+		osmo_stat_item_inc(osmo_stat_item_group_get_item(bts->bts_statg, BTS_STAT_OML_CONNECTED), 1);
+
+		/* Create link for E1INP_SIGN_OSMO */
+		//SAPI must be 0, no IPAC_PROTO_EXT_PCU, see ipaccess_bts_read_cb
+		bts->osmo_link = e1inp_sign_link_create(sign_ts, E1INP_SIGN_OSMO, bts->c0, IPAC_PROTO_OSMO, 0);
 		break;
 	case E1INP_SIGN_RSL: {
 		struct e1inp_ts *ts;
@@ -726,12 +739,12 @@ ipaccess_sign_link_up(void *unit_data, struct e1inp_line *line,
 
 		/* set new RSL link for this TRX. */
 		line = bts->oml_link->ts->line;
-		ts = &line->ts[E1INP_SIGN_RSL + dev->trx_id - 1];
+		ts = e1inp_line_ipa_rsl_ts(line, dev->trx_id);
 		e1inp_ts_config_sign(ts, line);
-		sign_link = trx->rsl_link =
+		sign_link = trx->rsl_link_primary =
 				e1inp_sign_link_create(ts, E1INP_SIGN_RSL,
-						       trx, trx->rsl_tei, 0);
-		trx->rsl_link->ts->sign.delay = 0;
+						       trx, trx->rsl_tei_primary, 0);
+		trx->rsl_link_primary->ts->sign.delay = 0;
 		if (!(sign_link->trx->bts->ip_access.flags &
 					(RSL_UP << sign_link->trx->nr))) {
 			e1inp_event(sign_link->ts, S_L_INP_TEI_UP,
@@ -739,7 +752,7 @@ ipaccess_sign_link_up(void *unit_data, struct e1inp_line *line,
 			sign_link->trx->bts->ip_access.flags |=
 					(RSL_UP << sign_link->trx->nr);
 		}
-		osmo_stat_item_inc(bts->bts_statg->items[BTS_STAT_RSL_CONNECTED], 1);
+		osmo_stat_item_inc(osmo_stat_item_group_get_item(bts->bts_statg, BTS_STAT_RSL_CONNECTED), 1);
 		break;
 	}
 	default:
@@ -751,7 +764,7 @@ ipaccess_sign_link_up(void *unit_data, struct e1inp_line *line,
 static void ipaccess_sign_link_down(struct e1inp_line *line)
 {
 	/* No matter what link went down, we close both signal links. */
-	struct e1inp_ts *ts = &line->ts[E1INP_SIGN_OML-1];
+	struct e1inp_ts *ts = e1inp_line_ipa_oml_ts(line);
 	struct gsm_bts *bts = NULL;
 	struct e1inp_sign_link *link;
 
@@ -784,6 +797,9 @@ static int ipaccess_sign_link(struct msgb *msg)
 	case E1INP_SIGN_OML:
 	        ret = abis_nm_rcvmsg(msg);
 	        break;
+	case E1INP_SIGN_OSMO:
+	        ret = abis_osmo_rcvmsg(msg);
+	        break;
 	default:
 		LOGP(DLINP, LOGL_ERROR, "Unknown signal link type %d\n",
 			link->type);
@@ -904,13 +920,17 @@ static void enc_power_params(struct msgb *msg, const struct gsm_power_ctrl_param
 	thresh_comp->red_step_size = cp->red_step_size_db & 0x0f;
 }
 
+void osmobts_enc_power_params_osmo_ext(struct msgb *msg, const struct gsm_power_ctrl_params *cp);
 static void add_power_params_ie(struct msgb *msg, enum abis_rsl_ie iei,
+				const struct gsm_bts_trx *trx,
 				const struct gsm_power_ctrl_params *cp)
 {
 	uint8_t *ie_len = msgb_tl_put(msg, iei);
 	uint8_t msg_len = msgb_length(msg);
 
 	enc_power_params(msg, cp);
+	if (iei == RSL_IE_MS_POWER_PARAM && is_osmobts(trx->bts))
+		osmobts_enc_power_params_osmo_ext(msg, cp);
 
 	*ie_len = msgb_length(msg) - msg_len;
 }
@@ -942,11 +962,11 @@ static int power_ctrl_send_def_params(const struct gsm_bts_trx *trx)
 
 	/* MS/BS Power Parameters IEs */
 	if (ms_power_ctrl->mode == GSM_PWR_CTRL_MODE_DYN_BTS)
-		add_power_params_ie(msg, RSL_IE_MS_POWER_PARAM, ms_power_ctrl);
+		add_power_params_ie(msg, RSL_IE_MS_POWER_PARAM, trx, ms_power_ctrl);
 	if (bs_power_ctrl->mode == GSM_PWR_CTRL_MODE_DYN_BTS)
-		add_power_params_ie(msg, RSL_IE_BS_POWER_PARAM, bs_power_ctrl);
+		add_power_params_ie(msg, RSL_IE_BS_POWER_PARAM, trx, bs_power_ctrl);
 
-	msg->dst = trx->rsl_link;
+	msg->dst = trx->rsl_link_primary;
 
 	return abis_rsl_sendmsg(msg);
 }
