@@ -434,7 +434,7 @@ static void nm_reconfig_trx(struct gsm_bts_trx *trx)
 		abis_nm_conn_terr_sign(trx, e1l->e1_nr, e1l->e1_ts,
 					e1l->e1_ts_ss);
 		abis_nm_establish_tei(trx->bts, trx->nr, e1l->e1_nr,
-				      e1l->e1_ts, e1l->e1_ts_ss, trx->rsl_tei);
+				      e1l->e1_ts, e1l->e1_ts_ss, trx->rsl_tei_primary);
 
 		/* Set Radio Attributes */
 		if (trx == trx->bts->c0)
@@ -552,6 +552,7 @@ static int shutdown_om(struct gsm_bts *bts)
 	/* Reset BTS Site manager resource */
 	abis_nm_bs11_reset_resource(bts);
 
+	gsm_bts_stats_reset(bts);
 	gsm_bts_all_ts_dispatch(bts, TS_EV_OML_DOWN, NULL);
 
 	return 0;
