@@ -45,6 +45,7 @@
 #include <osmocom/bsc/gsm_08_08.h>
 #include <osmocom/bsc/bts.h>
 #include <osmocom/bsc/lcs_loc_req.h>
+#include <osmocom/bsc/bsc_stats.h>
 
 #define LOG_FMT_BTS "bts %u lac-ci %u-%u arfcn-bsic %d-%d"
 #define LOG_ARGS_BTS(bts) \
@@ -61,7 +62,7 @@
 		lchan ? lchan->ts->nr : 0, \
 		lchan ? gsm_lchant_name(lchan->type) : "?", \
 		lchan ? lchan->nr : 0, \
-		lchan ? gsm48_chan_mode_name(lchan->tch_mode) : "?"
+		lchan ? gsm48_chan_mode_name(lchan->current_ch_mode_rate.chan_mode) : "?"
 
 #define LOG_FMT_TO_LCHAN "%u-%u-%u-%s%s%s-%u"
 #define LOG_ARGS_TO_LCHAN(lchan) \
@@ -88,8 +89,8 @@
 		LOG_HO(conn, LOGL_DEBUG, "(BSC) incrementing rate counter: %s %s\n", \
 		       bsc_ctr_description[counter].name, \
 		       bsc_ctr_description[counter].description); \
-		rate_ctr_inc(&conn->network->bsc_ctrs->ctr[counter]); \
-	} while(0)
+		rate_ctr_inc(rate_ctr_group_get_ctr(conn->network->bsc_ctrs, counter)); \
+	} while (0)
 
 /* Assume presence of local var 'conn' as struct gsm_subscriber_connection.
  * Handles bts == NULL gracefully
@@ -102,15 +103,18 @@
 		       bts_ctr_description[counter].name, \
 		       bts_ctr_description[counter].description); \
 		if (bts) \
-			rate_ctr_inc(&bts->bts_ctrs->ctr[counter]); \
+			rate_ctr_inc(rate_ctr_group_get_ctr(bts->bts_ctrs, counter)); \
 		else \
-			rate_ctr_inc(&conn->network->bts_unknown_ctrs->ctr[counter]); \
-	} while(0)
+			rate_ctr_inc(rate_ctr_group_get_ctr(conn->network->bts_unknown_ctrs, counter)); \
+	} while (0)
 
+/* Count handover result on both bts and bsc level.
+ * Call with 'counter' being the counter name without the "BSC_"/"BTS_" part,
+ * e.g. ho_count(conn_get_bts(conn), CTR_HANDOVER_ATTEMPTED); */
 #define ho_count(bts, counter) do { \
-        ho_count_bsc(BSC_##counter); \
-        ho_count_bts(bts, BTS_##counter); \
-} while(0)
+		ho_count_bsc(BSC_##counter); \
+		ho_count_bts(bts, BTS_##counter); \
+	} while (0)
 
 static uint8_t g_next_ho_ref = 1;
 
@@ -146,7 +150,7 @@ const char *handover_status(struct gsm_subscriber_connection *conn)
 		snprintf(buf, sizeof(buf),
 			 "("LOG_FMT_FROM_LCHAN") --HO-> (%s) " LOG_FMT_HO_SCOPE,
 			 LOG_ARGS_FROM_LCHAN(conn->lchan),
-			 neighbor_ident_key_name(&ho->target_cell),
+			 cell_ab_to_str_c(OTC_SELECT, &ho->target_cell_ab),
 			 LOG_ARGS_HO_SCOPE(conn));
 
 	else if (ho->scope & HO_INTER_BSC_IN) {
@@ -211,15 +215,15 @@ static const struct osmo_tdef_state_timeout ho_fsm_timeouts[32] = {
 		LOG_HO(conn, LOGL_ERROR, "Handover failed in state %s, %s: " fmt "\n", \
 		       osmo_fsm_inst_state_name(conn->fi), handover_result_name(result), ## args); \
 		handover_end(conn, result); \
-	} while(0)
+	} while (0)
 
 #define ho_success() do { \
 		LOG_HO(conn, LOGL_DEBUG, "Handover succeeded\n"); \
 		handover_end(conn, HO_RESULT_OK); \
-	} while(0)
+	} while (0)
 
 /* issue handover to a cell identified by ARFCN and BSIC */
-void handover_request(struct handover_out_req *req)
+int handover_request(struct handover_out_req *req)
 {
 	struct gsm_subscriber_connection *conn;
 	OSMO_ASSERT(req->old_lchan);
@@ -227,12 +231,9 @@ void handover_request(struct handover_out_req *req)
 	conn = req->old_lchan->conn;
 	OSMO_ASSERT(conn && conn->fi);
 
-	/* Make sure the handover target neighbor_ident_key contains the correct source bts nr */
-	req->target_nik.from_bts = req->old_lchan->ts->trx->bts->nr;
-
 	/* To make sure we're allowed to start a handover, go through a gscon event dispatch. If that is accepted, the
 	 * same req is passed to handover_start(). */
-	osmo_fsm_inst_dispatch(conn->fi, GSCON_EV_HANDOVER_START, req);
+	return osmo_fsm_inst_dispatch(conn->fi, GSCON_EV_HANDOVER_START, req);
 }
 
 /* Check that ho has old_lchan and/or new_lchan and conn pointers match.
@@ -275,7 +276,8 @@ static void handover_reset(struct gsm_subscriber_connection *conn)
 	struct osmo_mgcpc_ep_ci *ci;
 	if (conn->ho.new_lchan)
 		/* New lchan was activated but never passed to a conn */
-		lchan_release(conn->ho.new_lchan, false, true, RSL_ERR_EQUIPMENT_FAIL);
+		lchan_release(conn->ho.new_lchan, false, true, RSL_ERR_EQUIPMENT_FAIL,
+			      NULL);
 
 	ci = conn->ho.created_ci_for_msc;
 	if (ci) {
@@ -315,10 +317,10 @@ void handover_start(struct handover_out_req *req)
 
 	OSMO_ASSERT(req && req->old_lchan && req->old_lchan->conn);
 	struct gsm_subscriber_connection *conn = req->old_lchan->conn;
-	const struct neighbor_ident_key *search_for = &req->target_nik;
+	const struct cell_ab *search_for = &req->target_cell_ab;
 	struct handover *ho = &conn->ho;
 	struct gsm_bts *local_target_cell = NULL;
-	const struct gsm0808_cell_id_list2 *remote_target_cell = NULL;
+	struct gsm0808_cell_id_list2 remote_target_cells = {};
 
 	if (conn->ho.fi) {
 		LOG_HO(conn, LOGL_ERROR, "Handover requested while another handover is ongoing; Ignore\n");
@@ -335,9 +337,9 @@ void handover_start(struct handover_out_req *req)
 	ho->from_hodec_id = req->from_hodec_id;
 	ho->new_lchan_type = req->new_lchan_type == GSM_LCHAN_NONE ?
 		req->old_lchan->type : req->new_lchan_type;
-	ho->target_cell = req->target_nik;
+	ho->target_cell_ab = req->target_cell_ab;
 
-	if (find_handover_target_cell(&local_target_cell, &remote_target_cell,
+	if (find_handover_target_cell(&local_target_cell, &remote_target_cells,
 				      conn, search_for, true)) {
 		handover_end(conn, HO_RESULT_ERROR);
 		return;
@@ -349,8 +351,8 @@ void handover_start(struct handover_out_req *req)
 		return;
 	}
 
-	if (remote_target_cell) {
-		handover_start_inter_bsc_out(conn, remote_target_cell);
+	if (remote_target_cells.id_list_len) {
+		handover_start_inter_bsc_out(conn, &remote_target_cells);
 		return;
 	}
 
@@ -375,6 +377,7 @@ static void handover_start_intra_bsc(struct gsm_subscriber_connection *conn)
 	ho->scope = (ho->new_bts == bts) ? HO_INTRA_CELL : HO_INTRA_BSC;
 	ho->ho_ref = g_next_ho_ref++;
 	ho->async = true;
+	gsm_bts_cell_id_list(&ho->target_cell_ids, ho->new_bts);
 
 	ho->new_lchan = lchan_select_by_type(ho->new_bts, ho->new_lchan_type);
 
@@ -384,6 +387,7 @@ static void handover_start_intra_bsc(struct gsm_subscriber_connection *conn)
 	} else {
 		ho_count(bts, CTR_INTRA_BSC_HO_ATTEMPTED);
 		ho_fsm_update_id(fi, "intraBSC");
+		ho_count_bts(ho->new_bts, BTS_CTR_INCOMING_INTRA_BSC_HO_ATTEMPTED);
 	}
 
 	if (!ho->new_lchan) {
@@ -397,16 +401,16 @@ static void handover_start_intra_bsc(struct gsm_subscriber_connection *conn)
 	ho_fsm_state_chg(HO_ST_WAIT_LCHAN_ACTIVE);
 
 	info = (struct lchan_activate_info){
-		.activ_for = FOR_HANDOVER,
+		.activ_for = ACTIVATE_FOR_HANDOVER,
 		.for_conn = conn,
-		.chan_mode = conn->lchan->tch_mode,
+		.ch_mode_rate = conn->lchan->current_ch_mode_rate,
 		.encr = conn->lchan->encr,
 		.requires_voice_stream = conn->lchan->mgw_endpoint_ci_bts ? true : false,
 		.msc_assigned_cic = conn->ho.inter_bsc_in.msc_assigned_cic,
 		.re_use_mgw_endpoint_from_lchan = conn->lchan,
 		.wait_before_switching_rtp = true,
-		.s15_s0 = conn->lchan->activate.info.s15_s0,
 	};
+	info.ch_mode_rate.chan_rate = chan_t_to_chan_rate(ho->new_lchan->type);
 
 	/* For intra-cell handover, we know the accurate Timing Advance from the previous lchan. For inter-cell
 	 * handover, no Timing Advance for the new cell is known, so leave it unset. */
@@ -418,6 +422,32 @@ static void handover_start_intra_bsc(struct gsm_subscriber_connection *conn)
 	lchan_activate(ho->new_lchan, &info);
 }
 
+/* 3GPP TS 48.008 § 3.2.2.58  Old BSS to New BSS Information */
+static void parse_old2new_bss_info(struct gsm_subscriber_connection *conn,
+				   const uint8_t* data, uint16_t len,
+				   struct handover_in_req *req)
+{
+	/* § 3.2.2.58: Information contained here shall take precedence over
+	   duplicate information from Information Elements in the HANDOVER
+	   REQUEST  as long as the coding is understood by the new BSS */
+	/* § 3.2.2.58: <<Reception of an erroneous "Old BSS to New BSS
+	   information" shall not cause a rejection of the HANDOVER REQUEST
+	   message; the "Old BSS to New BSS information" information element
+	   shall be discarded and the handover resource allocation procedure
+	   shall continue>>. See also 3.1.19.7. */
+	struct tlv_parsed tp;
+	if (tlv_parse(&tp, &gsm0808_old_bss_to_new_bss_info_att_tlvdef, data, len, 0, 0) <= 0) {
+		LOG_HO(conn, LOGL_NOTICE, "Failed to parse IE \"Old BSS to New BSS information\"\n");
+		return;
+	}
+
+	if (TLVP_VAL(&tp, GSM0808_FE_IE_LAST_USED_EUTRAN_PLMN_ID)) {
+		req->last_eutran_plmn_valid = true;
+		osmo_plmn_from_bcd(TLVP_VAL(&tp, GSM0808_FE_IE_LAST_USED_EUTRAN_PLMN_ID),
+				   &req->last_eutran_plmn);
+	}
+}
+
 /* 3GPP TS 48.008 § 3.2.1.8 Handover Request */
 static bool parse_ho_request(struct gsm_subscriber_connection *conn, const struct msgb *msg,
 			     struct handover_in_req *req)
@@ -429,6 +459,7 @@ static bool parse_ho_request(struct gsm_subscriber_connection *conn, const struc
 	int payload_length;
 	bool aoip = gscon_is_aoip(conn);
 	bool sccplite = gscon_is_sccplite(conn);
+	bool has_a54 = false;
 
 	if ((aoip && sccplite) || !(aoip || sccplite)) {
 		LOG_HO(conn, LOGL_ERROR, "Received BSSMAP Handover Request, but conn is not"
@@ -460,6 +491,15 @@ static bool parse_ho_request(struct gsm_subscriber_connection *conn, const struc
 		return false;
 	}
 
+	if ((e = TLVP_GET(tp, GSM0808_IE_KC_128))) {
+		if (e->len != 16) {
+			LOG_HO(conn, LOGL_ERROR, "Invalid length in Kc128 IE: %u bytes (expected 16)\n", e->len);
+			return false;
+		}
+		memcpy(req->kc128, e->val, 16);
+		req->kc128_present = true;
+	}
+
 	if ((e = TLVP_GET(tp, GSM0808_IE_CLASSMARK_INFORMATION_TYPE_1))) {
 		if (e->len != sizeof(req->classmark.classmark1)) {
 			LOG_HO(conn, LOGL_ERROR, "Classmark Information 1 has wrong size\n");
@@ -488,9 +528,10 @@ static bool parse_ho_request(struct gsm_subscriber_connection *conn, const struc
 			       req->chosen_encr_alg);
 	}
 
-	LOG_HO(conn, LOGL_DEBUG, "Handover Request encryption info: chosen=A5/%u key=%s\n",
-	       (req->chosen_encr_alg ? : 1) - 1, req->ei.key_len?
-	       osmo_hexdump_nospc(req->ei.key, req->ei.key_len) : "none");
+	LOG_HO(conn, LOGL_DEBUG, "Handover Request encryption info: chosen=A5/%u key=%s kc128=%s\n",
+	       (req->chosen_encr_alg ? : 1) - 1,
+	       req->ei.key_len ? osmo_hexdump_nospc(req->ei.key, req->ei.key_len) : "none",
+	       has_a54 ? osmo_hexdump_nospc(req->kc128, 16) : "none");
 
 	if (TLVP_PRESENT(tp, GSM0808_IE_AOIP_TRASP_ADDR)) {
 		int rc;
@@ -558,6 +599,10 @@ static bool parse_ho_request(struct gsm_subscriber_connection *conn, const struc
 		return false;
 	}
 
+	if ((e = TLVP_GET(tp, GSM0808_IE_OLD_BSS_TO_NEW_BSS_INFORMATION))) {
+		parse_old2new_bss_info(conn, e->val, e->len, req);
+	}
+
 	/* A lot of IEs remain ignored... */
 
 	return true;
@@ -565,7 +610,7 @@ static bool parse_ho_request(struct gsm_subscriber_connection *conn, const struc
 
 static bool chan_mode_is_tch(enum gsm48_chan_mode mode)
 {
-	switch (mode) {
+	switch (gsm48_chan_mode_to_non_vamos(mode)) {
 	case GSM48_CMODE_SPEECH_V1:
 	case GSM48_CMODE_SPEECH_EFR:
 	case GSM48_CMODE_SPEECH_AMR:
@@ -665,10 +710,9 @@ void handover_start_inter_bsc_in(struct gsm_subscriber_connection *conn,
 	ho_fsm_state_chg(HO_ST_WAIT_LCHAN_ACTIVE);
 
 	info = (struct lchan_activate_info){
-		.activ_for = FOR_HANDOVER,
+		.activ_for = ACTIVATE_FOR_HANDOVER,
 		.for_conn = conn,
-		.chan_mode = ch_mode_rate.chan_mode,
-		.s15_s0 = ch_mode_rate.s15_s0,
+		.ch_mode_rate = ch_mode_rate,
 		.requires_voice_stream = chan_mode_is_tch(ch_mode_rate.chan_mode),
 		.msc_assigned_cic = req->msc_assigned_cic,
 	};
@@ -692,9 +736,39 @@ void handover_start_inter_bsc_in(struct gsm_subscriber_connection *conn,
 		info.encr.key_len = req->ei.key_len;
 	}
 
+	if (req->kc128_present) {
+		memcpy(info.encr.kc128, req->kc128, 16);
+		info.encr.kc128_present = true;
+	}
+
+	if (req->last_eutran_plmn_valid) {
+		conn->fast_return.allowed = ho->new_bts->srvcc_fast_return_allowed;
+		conn->fast_return.last_eutran_plmn_valid = true;
+		memcpy(&conn->fast_return.last_eutran_plmn, &req->last_eutran_plmn,
+		       sizeof(conn->fast_return.last_eutran_plmn));
+		ho_count(ho->new_bts, CTR_SRVCC_ATTEMPTED);
+	}
+
 	lchan_activate(ho->new_lchan, &info);
 }
 
+/* Create functions result_counter_{BSC,BTS}_{HANDOVER,...}(), to evaluate the handover result and return
+ * BSC_CTR_HANDOVER_ATTEMPTED,
+ * BSC_CTR_HANDOVER_COMPLETED,
+ * BSC_CTR_HANDOVER_STOPPED,
+ * BSC_CTR_HANDOVER_NO_CHANNEL,
+ * BSC_CTR_HANDOVER_TIMEOUT,
+ * BSC_CTR_HANDOVER_FAILED,
+ * BSC_CTR_HANDOVER_ERROR,
+ * or
+ * BTS_CTR_HANDOVER_ATTEMPTED,
+ * BTS_CTR_HANDOVER_COMPLETED,
+ * BTS_CTR_HANDOVER_STOPPED,
+ * BTS_CTR_HANDOVER_NO_CHANNEL,
+ * BTS_CTR_HANDOVER_TIMEOUT,
+ * BTS_CTR_HANDOVER_FAILED,
+ * BTS_CTR_HANDOVER_ERROR,
+ */
 #define FUNC_RESULT_COUNTER(obj, name) \
 static int result_counter_##obj##_##name(enum handover_result result) \
 { \
@@ -720,6 +794,7 @@ FUNC_RESULT_COUNTER(BSC, INTRA_CELL_HO)
 FUNC_RESULT_COUNTER(BSC, INTRA_BSC_HO)
 FUNC_RESULT_COUNTER(BSC, INTER_BSC_HO_IN)
 
+/* INTRA_BSC_HO_OUT does not have a NO_CHANNEL result, so can't do this with FUNC_RESULT_COUNTER() macro. */
 static int result_counter_BSC_INTER_BSC_HO_OUT(enum handover_result result) {
 	switch (result) {
 	case HO_RESULT_OK:
@@ -755,8 +830,10 @@ static int result_counter_bsc(enum handover_scope scope, enum handover_result re
 FUNC_RESULT_COUNTER(BTS, HANDOVER)
 FUNC_RESULT_COUNTER(BTS, INTRA_CELL_HO)
 FUNC_RESULT_COUNTER(BTS, INTRA_BSC_HO)
+FUNC_RESULT_COUNTER(BTS, INCOMING_INTRA_BSC_HO)
 FUNC_RESULT_COUNTER(BTS, INTER_BSC_HO_IN)
 
+/* INTRA_BSC_HO_OUT does not have a NO_CHANNEL result, so can't do this with FUNC_RESULT_COUNTER() macro. */
 static int result_counter_BTS_INTER_BSC_HO_OUT(enum handover_result result) {
 	switch (result) {
 	case HO_RESULT_OK:
@@ -789,6 +866,9 @@ static int result_counter_bts(enum handover_scope scope, enum handover_result re
 	}
 }
 
+FUNC_RESULT_COUNTER(BSC, SRVCC)
+FUNC_RESULT_COUNTER(BTS, SRVCC)
+
 static void send_handover_performed(struct gsm_subscriber_connection *conn)
 {
 	struct gsm_lchan *lchan = conn->lchan;
@@ -814,7 +894,7 @@ static void send_handover_performed(struct gsm_subscriber_connection *conn)
 	};
 
 	/* Chosen Channel 3.2.2.33 */
-	ho_perf_params.chosen_channel = gsm0808_chosen_channel(lchan->type, lchan->tch_mode);
+	ho_perf_params.chosen_channel = gsm0808_chosen_channel(lchan->type, lchan->current_ch_mode_rate.chan_mode);
 	if (!ho_perf_params.chosen_channel) {
 		LOG_HO(conn, LOGL_ERROR, "Failed to generate Chosen Channel IE, can't send HANDOVER PERFORMED!\n");
 		return;
@@ -827,14 +907,15 @@ static void send_handover_performed(struct gsm_subscriber_connection *conn)
 
 	if (ho->new_lchan->activate.info.requires_voice_stream) {
 		/* Speech Version (chosen) 3.2.2.51 */
-		ho_perf_params.speech_version_chosen = gsm0808_permitted_speech(lchan->type, lchan->tch_mode);
+		ho_perf_params.speech_version_chosen = gsm0808_permitted_speech(lchan->type,
+										lchan->current_ch_mode_rate.chan_mode);
 		ho_perf_params.speech_version_chosen_present = true;
 
 		/* Speech Codec (chosen) 3.2.2.104 */
 		if (gscon_is_aoip(conn)) {
 			/* Extrapolate speech codec from speech mode */
 			gsm0808_speech_codec_from_chan_type(&sc, ho_perf_params.speech_version_chosen);
-			sc.cfg = conn->lchan->ch_mode_rate.s15_s0;
+			sc.cfg = conn->lchan->current_ch_mode_rate.s15_s0;
 			memcpy(&ho_perf_params.speech_codec_chosen, &sc, sizeof(sc));
 			ho_perf_params.speech_codec_chosen_present = true;
 		}
@@ -846,7 +927,7 @@ static void send_handover_performed(struct gsm_subscriber_connection *conn)
 		return;
 	}
 
-	rate_ctr_inc(&conn->sccp.msc->msc_ctrs->ctr[MSC_CTR_BSSMAP_TX_DT1_HANDOVER_PERFORMED]);
+	rate_ctr_inc(rate_ctr_group_get_ctr(conn->sccp.msc->msc_ctrs, MSC_CTR_BSSMAP_TX_DT1_HANDOVER_PERFORMED));
 	rc = gscon_sigtran_send(conn, msg);
 	if (rc < 0) {
 		LOG_HO(conn, LOGL_ERROR, "message sending failed, can't send HANDOVER PERFORMED!\n");
@@ -910,7 +991,8 @@ void handover_end(struct gsm_subscriber_connection *conn, enum handover_result r
 				/* 3GPP TS 48.008 3.1.5.3.3 "Abnormal Conditions": if neither MS reports
 				 * HO Failure nor the MSC sends a Clear Command, we should release the
 				 * dedicated radio resources and send a Clear Request to the MSC. */
-				lchan_release(conn->lchan, true, true, GSM48_RR_CAUSE_ABNORMAL_TIMER);
+				lchan_release(conn->lchan, true, true, GSM48_RR_CAUSE_ABNORMAL_TIMER,
+					      gscon_last_eutran_plmn(conn));
 				/* Once the channel release is through, the BSSMAP Clear will follow. */
 				break;
 			}
@@ -940,6 +1022,14 @@ void handover_end(struct gsm_subscriber_connection *conn, enum handover_result r
 	ho_count_bsc(result_counter_bsc(ho->scope, result));
 	ho_count_bts(bts, result_counter_BTS_HANDOVER(result));
 	ho_count_bts(bts, result_counter_bts(ho->scope, result));
+	/* For inter-cell HO, also increment the "INCOMING" counters on the target BTS. */
+	if (ho->scope & HO_INTRA_BSC)
+		ho_count_bts(ho->new_bts, result_counter_BTS_INCOMING_INTRA_BSC_HO(result));
+	if (ho->scope & HO_INTER_BSC_IN && conn->fast_return.last_eutran_plmn_valid) {
+		/* From outside local BSC and with Last EUTRAN PLMN Id => SRVCC */
+		ho_count_bsc(result_counter_BSC_SRVCC(result));
+		ho_count_bts(bts, result_counter_BTS_SRVCC(result));
+	}
 
 	LOG_HO(conn, LOGL_INFO, "Result: %s\n", handover_result_name(result));
 
@@ -966,7 +1056,7 @@ void handover_end(struct gsm_subscriber_connection *conn, enum handover_result r
 	/* Detach the new_lchan last, so we can still see it in above logging */
 	if (ho->new_lchan) {
 		/* Release new lchan, it didn't work out */
-		lchan_release(ho->new_lchan, false, true, RSL_ERR_EQUIPMENT_FAIL);
+		lchan_release(ho->new_lchan, false, true, RSL_ERR_EQUIPMENT_FAIL, NULL);
 		ho->new_lchan = NULL;
 	}
 
