@@ -63,10 +63,10 @@ struct osmo_tdef_state_timeout lchan_rtp_fsm_timeouts[32] = {
 #define lchan_rtp_fail(fmt, args...) do { \
 		struct gsm_lchan *_lchan = fi->priv; \
 		uint32_t state_was = fi->state; \
-		lchan_set_last_error(_lchan, "lchan-rtp failure in state %s: " fmt, \
+		LCHAN_SET_LAST_ERROR(_lchan, "lchan-rtp failure in state %s: " fmt, \
 				     osmo_fsm_state_name(fi->fsm, state_was), ## args); \
 		osmo_fsm_inst_dispatch(_lchan->fi, LCHAN_EV_RTP_ERROR, 0); \
-	} while(0)
+	} while (0)
 
 /* Called from lchan_fsm_init(), does not need to be visible in lchan_rtp_fsm.h */
 void lchan_rtp_fsm_init()
@@ -268,19 +268,19 @@ static void lchan_rtp_fsm_wait_ipacc_crcx_ack_onenter(struct osmo_fsm_inst *fi,
 		return;
 	}
 
-	val = ipacc_speech_mode(lchan->tch_mode, lchan->type);
+	val = ipacc_speech_mode(lchan->activate.ch_mode_rate.chan_mode, lchan->type);
 	if (val < 0) {
-		lchan_rtp_fail("Cannot determine Abis/IP speech mode for tch_mode=%s type=%s\n",
-			   get_value_string(gsm48_chan_mode_names, lchan->tch_mode),
+		lchan_rtp_fail("Cannot determine Abis/IP speech mode for tch_mode=%s type=%s",
+			   get_value_string(gsm48_chan_mode_names, lchan->activate.ch_mode_rate.chan_mode),
 			   gsm_lchant_name(lchan->type));
 		return;
 	}
 	lchan->abis_ip.speech_mode = val;
 
-	val = ipacc_payload_type(lchan->tch_mode, lchan->type);
+	val = ipacc_payload_type(lchan->activate.ch_mode_rate.chan_mode, lchan->type);
 	if (val < 0) {
-		lchan_rtp_fail("Cannot determine Abis/IP payload type for tch_mode=%s type=%s\n",
-			   get_value_string(gsm48_chan_mode_names, lchan->tch_mode),
+		lchan_rtp_fail("Cannot determine Abis/IP payload type for tch_mode=%s type=%s",
+			   get_value_string(gsm48_chan_mode_names, lchan->activate.ch_mode_rate.chan_mode),
 			   gsm_lchant_name(lchan->type));
 		return;
 	}
@@ -796,7 +796,7 @@ static struct osmo_fsm lchan_rtp_fsm = {
 /* Depending on the channel mode and rate, return the codec type that is signalled towards the MGW. */
 static enum mgcp_codecs chan_mode_to_mgcp_codec(enum gsm48_chan_mode chan_mode, bool full_rate)
 {
-	switch (chan_mode) {
+	switch (gsm48_chan_mode_to_non_vamos(chan_mode)) {
 	case GSM48_CMODE_SPEECH_V1:
 		if (full_rate)
 			return CODEC_GSM_8000_1;
@@ -834,14 +834,14 @@ static int chan_mode_to_mgcp_bss_pt(enum mgcp_codecs codec)
 
 void mgcp_pick_codec(struct mgcp_conn_peer *verb_info, const struct gsm_lchan *lchan, bool bss_side)
 {
-	enum mgcp_codecs codec = chan_mode_to_mgcp_codec(lchan->tch_mode,
+	enum mgcp_codecs codec = chan_mode_to_mgcp_codec(lchan->activate.ch_mode_rate.chan_mode,
 							 lchan->type == GSM_LCHAN_TCH_H? false : true);
 	int custom_pt;
 
 	if (codec < 0) {
 		LOG_LCHAN(lchan, LOGL_ERROR,
 			  "Unable to determine MGCP codec type for %s in chan-mode %s\n",
-			  gsm_lchant_name(lchan->type), gsm48_chan_mode_name(lchan->tch_mode));
+			  gsm_lchant_name(lchan->type), gsm48_chan_mode_name(lchan->activate.ch_mode_rate.chan_mode));
 		verb_info->codecs_len = 0;
 		return;
 	}
