@@ -57,6 +57,7 @@
 #include <osmocom/vty/cpu_sched_vty.h>
 
 #include <osmocom/mgcp_client/mgcp_client_endpoint_fsm.h>
+#include <osmocom/mgcp_client/mgcp_client_pool.h>
 
 #include <osmocom/abis/abis.h>
 #include <osmocom/bsc/abis_om2000.h>
@@ -67,8 +68,10 @@
 #include <osmocom/bsc/codec_pref.h>
 #include <osmocom/bsc/system_information.h>
 #include <osmocom/bsc/bts.h>
+#include <osmocom/bsc/bsc_stats.h>
 
 #include <osmocom/mgcp_client/mgcp_client.h>
+#include <osmocom/mgcp_client/mgcp_client_pool.h>
 
 #include <osmocom/sigtran/xua_msg.h>
 
@@ -84,7 +87,6 @@
 
 #include "../../bscconfig.h"
 
-struct gsm_network *bsc_gsmnet = 0;
 static const char *config_file = "osmo-bsc.cfg";
 static const char *rf_ctrl = NULL;
 static int daemonize = 0;
@@ -104,10 +106,8 @@ static void print_help()
 	printf("  -T --timestamp		Print a timestamp in the debug output.\n");
 	printf("  -V --version               	Print the version of OsmoBSC.\n");
 	printf("  -c --config-file filename	The config file to use.\n");
-	printf("  -l --local IP			The local address of the MGCP.\n");
 	printf("  -e --log-level number		Set a global loglevel.\n");
 	printf("  -r --rf-ctl NAME		A unix domain socket to listen for cmds.\n");
-	printf("  -t --testmode			A special mode to provoke failures at the MSC.\n");
 
 	printf("\nVTY reference generation:\n");
 	printf("     --vty-ref-mode MODE	VTY reference generation mode (e.g. 'expert').\n");
@@ -152,16 +152,14 @@ static void handle_options(int argc, char **argv)
 			{"disable-color", 0, 0, 's'},
 			{"timestamp", 0, 0, 'T'},
 			{"version", 0, 0, 'V' },
-			{"local", 1, 0, 'l'},
 			{"log-level", 1, 0, 'e'},
 			{"rf-ctl", 1, 0, 'r'},
-			{"testmode", 0, 0, 't'},
 			{"vty-ref-mode", 1, &long_option, 1},
 			{"vty-ref-xml", 0, &long_option, 2},
 			{0, 0, 0, 0}
 		};
 
-		c = getopt_long(argc, argv, "hd:DsTVc:e:r:t",
+		c = getopt_long(argc, argv, "hd:DsTVc:e:r:",
 				long_options, &option_index);
 		if (c == -1)
 			break;
@@ -256,14 +254,14 @@ static int nm_sig_cb(unsigned int subsys, unsigned int signal,
 }
 
 /* Produce a MA as specified in 10.5.2.21 */
-static int generate_ma_for_ts(struct gsm_bts_trx_ts *ts)
+static void generate_ma_for_ts(struct gsm_bts_trx_ts *ts)
 {
 	/* we have three bitvecs: the per-timeslot ARFCNs, the cell chan ARFCNs
 	 * and the MA */
+	const size_t num_cell_arfcns = ts->trx->bts->si_common.cell_chan_num;
 	const struct bitvec *cell_chan = &ts->trx->bts->si_common.cell_alloc;
 	const struct bitvec *ts_arfcn = &ts->hopping.arfcns;
 	struct bitvec *ma = &ts->hopping.ma;
-	unsigned int num_cell_arfcns;
 	int i;
 
 	/* re-set the MA to all-zero */
@@ -271,14 +269,7 @@ static int generate_ma_for_ts(struct gsm_bts_trx_ts *ts)
 	memset(ma->data, 0, ma->data_len);
 
 	if (!ts->hopping.enabled)
-		return 0;
-
-	/* count the number of ARFCNs in the cell channel allocation */
-	num_cell_arfcns = 0;
-	for (i = 0; i < 1024; i++) {
-		if (bitvec_get_bit_pos(cell_chan, i))
-			num_cell_arfcns++;
-	}
+		return;
 
 	/* pad it to octet-aligned number of bits */
 	ts->hopping.ma_len = OSMO_BYTES_FOR_BITS(num_cell_arfcns);
@@ -303,8 +294,20 @@ static int generate_ma_for_ts(struct gsm_bts_trx_ts *ts)
 		else
 			bitvec_set_bit_pos(ma, ma->cur_bit, 0);
 	}
+}
 
-	return 0;
+static void generate_ma_for_bts(struct gsm_bts *bts)
+{
+	struct gsm_bts_trx *trx;
+	unsigned int tn;
+
+	OSMO_ASSERT(bts->si_common.cell_chan_num > 0);
+	OSMO_ASSERT(bts->si_common.cell_chan_num <= 64);
+
+	llist_for_each_entry(trx, &bts->trx_list, list) {
+		for (tn = 0; tn < ARRAY_SIZE(trx->ts); tn++)
+			generate_ma_for_ts(&trx->ts[tn]);
+	}
 }
 
 static void bootstrap_rsl(struct gsm_bts_trx *trx)
@@ -352,7 +355,6 @@ static void bootstrap_rsl(struct gsm_bts_trx *trx)
 
 	for (i = 0; i < ARRAY_SIZE(trx->ts); i++) {
 		struct gsm_bts_trx_ts *ts = &trx->ts[i];
-		generate_ma_for_ts(ts);
 		OSMO_ASSERT(ts->fi);
 		osmo_fsm_inst_dispatch(ts->fi, TS_EV_RSL_READY, NULL);
 	}
@@ -375,110 +377,54 @@ static void all_ts_dispatch_event(struct gsm_bts_trx *trx, uint32_t event)
 	}
 }
 
-/* Callback function to be called every time we receive a signal from INPUT */
-static int inp_sig_cb(unsigned int subsys, unsigned int signal,
-		      void *handler_data, void *signal_data)
+struct osmo_timer_list update_connection_stats_timer;
+
+/* Periodically call bsc_update_connection_stats() to keep stat items updated.
+ * It would be nicer to trigger this only when OML or RSL state is seen to flip. I tried hard to find all code paths
+ * that should call this and failed to get accurate results; this trivial timer covers all of them. */
+static void update_connection_stats_cb(void *data)
 {
-	struct input_signal_data *isd = signal_data;
-	struct gsm_bts_trx *trx = isd->trx;
-
-	if (subsys != SS_L_INPUT)
-		return -EINVAL;
-
-	LOGP(DLMI, LOGL_DEBUG, "%s(): Input signal '%s' received\n", __func__,
-		get_value_string(e1inp_signal_names, signal));
-	switch (signal) {
-	case S_L_INP_TEI_UP:
-		if (isd->link_type == E1INP_SIGN_OML) {
-			/* TODO: this is required for the Nokia BTS, hopping is configured
-			   during OML, other MA is not set.  */
-			struct gsm_bts_trx *cur_trx;
-			uint8_t ca[20];
-			/* has to be called before generate_ma_for_ts to
-			  set bts->si_common.cell_alloc */
-			generate_cell_chan_list(ca, trx->bts);
-
-			llist_for_each_entry(cur_trx, &trx->bts->trx_list, list) {
-				int i;
-				for (i = 0; i < ARRAY_SIZE(cur_trx->ts); i++)
-					generate_ma_for_ts(&cur_trx->ts[i]);
-			}
-		}
-		if (isd->link_type == E1INP_SIGN_RSL)
-			bootstrap_rsl(trx);
-		break;
-	case S_L_INP_TEI_DN:
-		LOG_TRX(trx, DLMI, LOGL_ERROR, "Lost E1 %s link\n", e1inp_signtype_name(isd->link_type));
-
-		if (isd->link_type == E1INP_SIGN_OML) {
-			rate_ctr_inc(&trx->bts->bts_ctrs->ctr[BTS_CTR_BTS_OML_FAIL]);
-			all_ts_dispatch_event(trx, TS_EV_OML_DOWN);
-		} else if (isd->link_type == E1INP_SIGN_RSL) {
-			rate_ctr_inc(&trx->bts->bts_ctrs->ctr[BTS_CTR_BTS_RSL_FAIL]);
-			acc_ramp_abort(&trx->bts->acc_ramp);
-			all_ts_dispatch_event(trx, TS_EV_RSL_DOWN);
-			if (trx->nr == 0)
-				osmo_timer_del(&trx->bts->cbch_timer);
-		}
-
-		gsm_bts_sm_mo_reset(trx->bts->site_mgr);
-
-		abis_nm_clear_queue(trx->bts);
-		break;
-	default:
-		break;
-	}
-
-	return 0;
+	bsc_update_connection_stats(bsc_gsmnet);
+	osmo_timer_setup(&update_connection_stats_timer, update_connection_stats_cb, NULL);
+	osmo_timer_schedule(&update_connection_stats_timer, 1, 0);
 }
 
-static int bootstrap_bts(struct gsm_bts *bts)
+static int check_bts(struct gsm_bts *bts)
 {
 	struct gsm_bts_trx *trx;
-	unsigned int n = 0;
 
 	if (!bts->model)
 		return -EFAULT;
 
-	if (bts->model->start && !bts->model->started) {
-		int ret = bts->model->start(bts->network);
-		if (ret < 0)
-			return ret;
-
-		bts->model->started = true;
-	}
-
-	/* FIXME: What about secondary TRX of a BTS?  What about a BTS that has TRX
-	 * in different bands? Why is 'band' a parameter of the BTS and not of the TRX? */
 	switch (bts->band) {
 	case GSM_BAND_1800:
 		if (bts->c0->arfcn < 512 || bts->c0->arfcn > 885) {
-			LOGP(DNM, LOGL_ERROR, "GSM1800 channel must be between 512-885.\n");
+			LOGP(DNM, LOGL_ERROR, "(bts=%u) GSM1800 channel (%u) must be between 512-885.\n",
+			     bts->nr, bts->c0->arfcn);
 			return -EINVAL;
 		}
 		break;
 	case GSM_BAND_1900:
 		if (bts->c0->arfcn < 512 || bts->c0->arfcn > 810) {
-			LOGP(DNM, LOGL_ERROR, "GSM1900 channel must be between 512-810.\n");
-			return -EINVAL;
+			LOGP(DNM, LOGL_ERROR, "(bts=%u) GSM1900 channel (%u) must be between 512-810.\n",
+			     bts->nr, bts->c0->arfcn);
 		}
 		break;
 	case GSM_BAND_900:
 		if ((bts->c0->arfcn > 124 && bts->c0->arfcn < 955) ||
 		    bts->c0->arfcn > 1023)  {
-			LOGP(DNM, LOGL_ERROR, "GSM900 channel must be between 0-124, 955-1023.\n");
-			return -EINVAL;
+			LOGP(DNM, LOGL_ERROR, "(bts=%u) GSM900 channel (%u) must be between 0-124, 955-1023.\n",
+			     bts->nr, bts->c0->arfcn);
 		}
 		break;
 	case GSM_BAND_850:
 		if (bts->c0->arfcn < 128 || bts->c0->arfcn > 251) {
-			LOGP(DNM, LOGL_ERROR, "GSM850 channel must be between 128-251.\n");
-			return -EINVAL;
+			LOGP(DNM, LOGL_ERROR, "(bts=%u) GSM850 channel (%u) must be between 128-251.\n",
+			     bts->nr, bts->c0->arfcn);
 		}
 		break;
 	default:
-		LOGP(DNM, LOGL_ERROR, "Unsupported frequency band.\n");
-		return -EINVAL;
+		LOGP(DNM, LOGL_ERROR, "(bts=%u) Unsupported frequency band.\n", bts->nr);
 	}
 
 	/* Verify the physical channel mapping */
@@ -490,10 +436,14 @@ static int bootstrap_bts(struct gsm_bts *bts)
 		}
 	}
 
-	/* Control Channel Description is set from vty/config */
+	return 0;
+}
 
-	/* Indicate R99 MSC in SI3 */
-	bts->si_common.chan_desc.mscr = 1;
+static void bootstrap_bts(struct gsm_bts *bts)
+{
+	unsigned int n = 0;
+
+	/* Control Channel Description is set from vty/config */
 
 	/* Determine the value of CCCH_CONF. Is TS0/C0 combined? */
 	if (bts->c0->ts[0].pchan_from_config != GSM_PCHAN_CCCH) {
@@ -515,19 +465,73 @@ static int bootstrap_bts(struct gsm_bts *bts)
 		bts->si_common.chan_desc.ccch_conf = (n << 1);
 	}
 
-	bts->si_common.cell_options.pwrc = 0; /* PWRC not set */
-
-	bts->si_common.cell_sel_par.acs = 0;
-
-	bts->si_common.ncc_permitted = 0xff;
-
-	bts->chan_load_samples_idx = 0;
-
 	/* ACC ramping is initialized from vty/config */
 
 	/* Initialize the BTS state */
 	gsm_bts_sm_mo_reset(bts->site_mgr);
 
+	/* Generate Mobile Allocation bit-masks for all timeslots.
+	 * This needs to be done here, because it's used for TS configuration. */
+	generate_ma_for_bts(bts);
+}
+
+/* Callback function to be called every time we receive a signal from INPUT */
+static int inp_sig_cb(unsigned int subsys, unsigned int signal,
+		      void *handler_data, void *signal_data)
+{
+	struct input_signal_data *isd = signal_data;
+	struct gsm_bts_trx *trx = isd->trx;
+	int rc;
+
+	if (subsys != SS_L_INPUT)
+		return -EINVAL;
+
+	LOGP(DLMI, LOGL_DEBUG, "%s(): Input signal '%s' received\n", __func__,
+		get_value_string(e1inp_signal_names, signal));
+	switch (signal) {
+	case S_L_INP_TEI_UP:
+		if (isd->link_type == E1INP_SIGN_OML) {
+			/* Check parameters and apply vty config dependent parameters */
+			rc = check_bts(trx->bts);
+			if (rc < 0) {
+				LOGP(DNM, LOGL_ERROR, "(bts=%u) Error in BTS configuration -- cannot bootstrap BTS\n",
+				     trx->bts->nr);
+				return rc;
+			}
+			bootstrap_bts(trx->bts);
+		}
+		if (isd->link_type == E1INP_SIGN_RSL) {
+			rc = check_bts(trx->bts);
+			if (rc < 0) {
+				LOGP(DNM, LOGL_ERROR, "(bts=%u) Error in BTS configuration -- cannot bootstrap RSL\n",
+				     trx->bts->nr);
+				return rc;
+			}
+			bootstrap_rsl(trx);
+		}
+		break;
+	case S_L_INP_TEI_DN:
+		LOG_TRX(trx, DLMI, LOGL_ERROR, "Lost E1 %s link\n", e1inp_signtype_name(isd->link_type));
+
+		if (isd->link_type == E1INP_SIGN_OML) {
+			rate_ctr_inc(rate_ctr_group_get_ctr(trx->bts->bts_ctrs, BTS_CTR_BTS_OML_FAIL));
+			all_ts_dispatch_event(trx, TS_EV_OML_DOWN);
+		} else if (isd->link_type == E1INP_SIGN_RSL) {
+			rate_ctr_inc(rate_ctr_group_get_ctr(trx->bts->bts_ctrs, BTS_CTR_BTS_RSL_FAIL));
+			acc_ramp_abort(&trx->bts->acc_ramp);
+			all_ts_dispatch_event(trx, TS_EV_RSL_DOWN);
+			if (trx->nr == 0)
+				osmo_timer_del(&trx->bts->cbch_timer);
+		}
+
+		gsm_bts_sm_mo_reset(trx->bts->site_mgr);
+
+		abis_nm_clear_queue(trx->bts);
+		break;
+	default:
+		break;
+	}
+
 	return 0;
 }
 
@@ -552,11 +556,12 @@ static int bsc_network_configure(const char *config_file)
 	osmo_signal_register_handler(SS_L_INPUT, inp_sig_cb, NULL);
 
 	llist_for_each_entry(bts, &bsc_gsmnet->bts_list, list) {
-		rc = bootstrap_bts(bts);
+		rc = check_bts(bts);
 		if (rc < 0) {
-			LOGP(DNM, LOGL_FATAL, "Error bootstrapping BTS\n");
+			LOGP(DNM, LOGL_FATAL, "(bts=%u) cannot bootstrap BTS, invalid BTS configuration\n", bts->nr);
 			return rc;
 		}
+		bootstrap_bts(bts);
 		rc = e1_reconfig_bts(bts);
 		if (rc < 0) {
 			LOGP(DNM, LOGL_FATAL, "Error enabling E1 input driver\n");
@@ -613,7 +618,6 @@ static int bsc_vty_go_parent(struct vty *vty)
 			/* set vty->index correctly ! */
 			struct gsm_bts_trx_ts *ts = vty->index;
 			vty->index = ts->trx;
-			vty->index_sub = &ts->trx->description;
 		}
 		break;
 	case OML_NODE:
@@ -704,10 +708,12 @@ static void signal_handler(int signum)
 	switch (signum) {
 	case SIGINT:
 	case SIGTERM:
+		/* If SIGTERM was already sent before, just terminate immediately. */
+		if (osmo_select_shutdown_requested())
+			exit(-1);
 		bsc_shutdown_net(bsc_gsmnet);
 		osmo_signal_dispatch(SS_L_GLOBAL, S_L_GLOBAL_SHUTDOWN, NULL);
-		sleep(3);
-		exit(0);
+		osmo_select_shutdown_request();
 		break;
 	case SIGABRT:
 		/* in case of abort, we want to obtain a talloc report and
@@ -847,6 +853,12 @@ static const struct log_info_cat osmo_bsc_categories[] = {
 		.description = "RESET/ACK on A and Lb interfaces",
 		.enabled = 1, .loglevel = LOGL_NOTICE,
 	},
+	[DLOOP] = {
+		.name = "DLOOP",
+		.description = "Control loops",
+		.color = "\033[0;34m",
+		.enabled = 1, .loglevel = LOGL_NOTICE,
+	},
 };
 
 static int filter_fn(const struct log_context *ctx, struct log_target *tar)
@@ -873,6 +885,41 @@ extern void *tall_fle_ctx;
 extern void *tall_tqe_ctx;
 extern void *tall_ctr_ctx;
 
+static int bsc_mgw_setup(void)
+{
+	struct mgcp_client *mgcp_client_single;
+	unsigned int pool_members_initalized;
+
+	/* Initalize MGW pool. This initalizes and connects all MGCP clients that are currently configured in
+	 * the pool. Adding additional MGCP clients to the pool is possible but the user has to configure and
+	 * (re)connect them manually from the VTY. */
+	pool_members_initalized = mgcp_client_pool_connect(bsc_gsmnet->mgw.mgw_pool);
+	if (pool_members_initalized) {
+		LOGP(DNM, LOGL_NOTICE,
+		     "MGW pool with %u pool members configured, (ignoring MGW configuration in VTY node 'msc').\n",
+		     pool_members_initalized);
+		return 0;
+	}
+
+	/* Initialize and connect a single MGCP client. This MGCP client will appear as the one and only pool
+	 * member if there is no MGW pool configured. */
+	LOGP(DNM, LOGL_NOTICE, "No MGW pool configured, using MGW configuration in VTY node 'msc'\n");
+        mgcp_client_single = mgcp_client_init(bsc_gsmnet, bsc_gsmnet->mgw.conf);
+	if (!mgcp_client_single) {
+		LOGP(DNM, LOGL_ERROR, "MGW (single) client initalization failed\n");
+		return -EINVAL;
+	}
+	if (mgcp_client_connect(mgcp_client_single)) {
+		LOGP(DNM, LOGL_ERROR, "MGW (single) connect failed at (%s:%u)\n",
+		     bsc_gsmnet->mgw.conf->remote_addr,
+		     bsc_gsmnet->mgw.conf->remote_port);
+		return -EINVAL;
+	}
+	mgcp_client_pool_register_single(bsc_gsmnet->mgw.mgw_pool, mgcp_client_single);
+
+	return 0;
+}
+
 int main(int argc, char **argv)
 {
 	struct bsc_msc_data *msc;
@@ -903,6 +950,7 @@ int main(int argc, char **argv)
 	}
 
 	bsc_gsmnet->mgw.conf = talloc_zero(bsc_gsmnet, struct mgcp_client_conf);
+	bsc_gsmnet->mgw.mgw_pool = mgcp_client_pool_alloc(bsc_gsmnet);
 	mgcp_client_conf_init(bsc_gsmnet->mgw.conf);
 
 	bts_init();
@@ -930,7 +978,6 @@ int main(int argc, char **argv)
 	/* seed the PRNG */
 	srand(time(NULL));
 
-	ts_fsm_init();
 	lchan_fsm_init();
 	bsc_subscr_conn_fsm_init();
 	assignment_fsm_init();
@@ -944,6 +991,11 @@ int main(int argc, char **argv)
 		exit(1);
 	}
 
+	if (neighbors_check_cfg()) {
+		fprintf(stderr, "Errors in neighbor configuration, check the DHO log. exiting.\n");
+		exit(1);
+	}
+
 	/* start control interface after reading config for
 	 * ctrl_vty_get_bind_addr() */
 	bsc_gsmnet->ctrl = bsc_controlif_setup(bsc_gsmnet,
@@ -1001,14 +1053,8 @@ int main(int argc, char **argv)
 		}
 	}
 
-	bsc_gsmnet->mgw.client = mgcp_client_init(bsc_gsmnet, bsc_gsmnet->mgw.conf);
-
-	if (mgcp_client_connect(bsc_gsmnet->mgw.client)) {
-		LOGP(DNM, LOGL_ERROR, "MGW connect failed at (%s:%u)\n",
-		     bsc_gsmnet->mgw.conf->remote_addr,
-		     bsc_gsmnet->mgw.conf->remote_port);
+	if (bsc_mgw_setup() != 0)
 		exit(1);
-	}
 
 	if (osmo_bsc_sigtran_init(&bsc_gsmnet->mscs) != 0) {
 		LOGP(DNM, LOGL_ERROR, "Failed to initialize sigtran backhaul.\n");
@@ -1027,6 +1073,8 @@ int main(int argc, char **argv)
 	signal(SIGUSR2, &signal_handler);
 	osmo_init_ignore_signals();
 
+	update_connection_stats_cb(NULL);
+
 	if (daemonize) {
 		rc = osmo_daemonize();
 		if (rc < 0) {
@@ -1035,7 +1083,7 @@ int main(int argc, char **argv)
 		}
 	}
 
-	while (1) {
+	while (!osmo_select_shutdown_done()) {
 		osmo_select_main_ctx(0);
 	}
 
