@@ -33,11 +33,11 @@
 #include <osmocom/gsm/gsm48_ie.h>
 #include <osmocom/gsm/gsm48_rest_octets.h>
 #include <osmocom/gsm/gsm48.h>
+#include <osmocom/gsm/gsm48_arfcn_range_encode.h>
 
 #include <osmocom/bsc/debug.h>
 #include <osmocom/bsc/gsm_data.h>
 #include <osmocom/bsc/abis_rsl.h>
-#include <osmocom/bsc/arfcn_range_encode.h>
 #include <osmocom/bsc/gsm_04_08_rr.h>
 #include <osmocom/bsc/acc.h>
 #include <osmocom/bsc/neighbor_ident.h>
@@ -431,7 +431,7 @@ static inline int enc_freq_lst_var_bitmap(uint8_t *chan_list,
 	return 0;
 }
 
-int range_encode(enum gsm48_range r, int *arfcns, int arfcns_used, int *w,
+int range_encode(enum osmo_gsm48_range r, int *arfcns, int arfcns_used, int *w,
 		 int f0, uint8_t *chan_list)
 {
 	/*
@@ -440,22 +440,22 @@ int range_encode(enum gsm48_range r, int *arfcns, int arfcns_used, int *w,
 	 */
 	int rc, f0_included;
 
-	range_enc_filter_arfcns(arfcns, arfcns_used, f0, &f0_included);
+	osmo_gsm48_range_enc_filter_arfcns(arfcns, arfcns_used, f0, &f0_included);
 
-	rc = range_enc_arfcns(r, arfcns, arfcns_used, w, 0);
+	rc = osmo_gsm48_range_enc_arfcns(r, arfcns, arfcns_used, w, 0);
 	if (rc < 0)
 		return rc;
 
 	/* Select the range and the amount of bits needed */
 	switch (r) {
-	case ARFCN_RANGE_128:
-		return range_enc_range128(chan_list, f0, w);
-	case ARFCN_RANGE_256:
-		return range_enc_range256(chan_list, f0, w);
-	case ARFCN_RANGE_512:
-		return range_enc_range512(chan_list, f0, w);
-	case ARFCN_RANGE_1024:
-		return range_enc_range1024(chan_list, f0, f0_included, w);
+	case OSMO_GSM48_ARFCN_RANGE_128:
+		return osmo_gsm48_range_enc_128(chan_list, f0, w);
+	case OSMO_GSM48_ARFCN_RANGE_256:
+		return osmo_gsm48_range_enc_256(chan_list, f0, w);
+	case OSMO_GSM48_ARFCN_RANGE_512:
+		return osmo_gsm48_range_enc_512(chan_list, f0, w);
+	case OSMO_GSM48_ARFCN_RANGE_1024:
+		return osmo_gsm48_range_enc_1024(chan_list, f0, f0_included, w);
 	default:
 		return -ERANGE;
 	};
@@ -468,8 +468,8 @@ static inline int enc_freq_lst_range(uint8_t *chan_list,
 				const struct bitvec *bv, const struct gsm_bts *bts,
 				bool bis, bool ter, bool pgsm)
 {
-	int arfcns[RANGE_ENC_MAX_ARFCNS];
-	int w[RANGE_ENC_MAX_ARFCNS];
+	int arfcns[OSMO_GSM48_RANGE_ENC_MAX_ARFCNS];
+	int w[OSMO_GSM48_RANGE_ENC_MAX_ARFCNS];
 	int arfcns_used = 0;
 	int i, range, f0;
 
@@ -488,8 +488,8 @@ static inline int enc_freq_lst_range(uint8_t *chan_list,
 	/*
 	 * Check if the given list of ARFCNs can be encoded.
 	 */
-	range = range_enc_determine_range(arfcns, arfcns_used, &f0);
-	if (range == ARFCN_RANGE_INVALID)
+	range = osmo_gsm48_range_enc_determine_range(arfcns, arfcns_used, &f0);
+	if (range == OSMO_GSM48_ARFCN_RANGE_INVALID)
 		return -2;
 
 	memset(w, 0, sizeof(w));
@@ -574,51 +574,63 @@ static int bitvec2freq_list(uint8_t *chan_list, const struct bitvec *bv,
 	return -EINVAL;
 }
 
-/* generate a cell channel list as per Section 10.5.2.1b of 04.08 */
-int generate_cell_chan_list(uint8_t *chan_list, struct gsm_bts *bts)
+/* (Re)generate Cell Allocation (basically a bit-vector of all cell channels) */
+int generate_cell_chan_alloc(struct gsm_bts *bts)
 {
-	struct gsm_bts_trx *trx;
-	struct bitvec *bv = &bts->si_common.cell_alloc;
+	const struct gsm_bts_trx *trx;
+	unsigned int chan, chan_num;
+	unsigned int tn;
 
-	/* Zero-initialize the bit-vector */
-	memset(bv->data, 0, bv->data_len);
+	/* Temporary Cell Allocation bit-vector */
+	uint8_t ca[1024 / 8] = { 0 };
+	struct bitvec bv = {
+		.data_len = sizeof(ca),
+		.data = &ca[0],
+	};
 
-	/* first we generate a bitvec of all TRX ARFCN's in our BTS */
+	/* Calculate a bit-mask of all assigned ARFCNs */
 	llist_for_each_entry(trx, &bts->trx_list, list) {
-		unsigned int i, j;
 		/* Always add the TRX's ARFCN */
-		bitvec_set_bit_pos(bv, trx->arfcn, 1);
-		for (i = 0; i < ARRAY_SIZE(trx->ts); i++) {
-			struct gsm_bts_trx_ts *ts = &trx->ts[i];
+		bitvec_set_bit_pos(&bv, trx->arfcn, 1);
+		for (tn = 0; tn < ARRAY_SIZE(trx->ts); tn++) {
+			const struct gsm_bts_trx_ts *ts = &trx->ts[tn];
 			/* Add any ARFCNs present in hopping channels */
-			for (j = 0; j < 1024; j++) {
-				if (bitvec_get_bit_pos(&ts->hopping.arfcns, j))
-					bitvec_set_bit_pos(bv, j, 1);
+			for (chan = 0; chan < sizeof(ca) * 8; chan++) {
+				if (bitvec_get_bit_pos(&ts->hopping.arfcns, chan) == ONE)
+					bitvec_set_bit_pos(&bv, chan, ONE);
 			}
 		}
 	}
 
-	/* then we generate a GSM 04.08 frequency list from the bitvec */
-	return bitvec2freq_list(chan_list, bv, bts, false, false);
+	/* Calculate the overall number of assigned ARFCNs */
+	for (chan_num = 0, chan = 0; chan < sizeof(ca) * 8; chan++) {
+		if (bitvec_get_bit_pos(&bv, chan) == ONE)
+			chan_num++;
+	}
+
+	/* The Mobile Allocation IE may contain up to 64 bits, so here we
+	 * cannot allow more than 64 channels in Cell Allocation. */
+	if (chan_num > 64) {
+		LOGP(DRR, LOGL_ERROR, "Failed to (re)generate Cell Allocation: "
+		     "number of channels (%u) exceeds the maximum number of "
+		     "ARFCNs in Mobile Allocation (64)\n", chan_num);
+		return -E2BIG;
+	}
+
+	/* Commit the new Channel Allocation */
+	memcpy(&bts->si_common.data.cell_alloc[0], ca, sizeof(ca));
+	bts->si_common.cell_chan_num = chan_num;
+
+	return 0;
 }
 
-struct generate_bcch_chan_list__ni_iter_data {
-	struct gsm_bts *bts;
-	struct bitvec *bv;
-};
-
-static bool generate_bcch_chan_list__ni_iter_cb(const struct neighbor_ident_key *key,
-						const struct gsm0808_cell_id_list2 *val,
-						void *cb_data)
+/* generate a cell channel list as per Section 10.5.2.1b of 04.08 */
+int generate_cell_chan_list(uint8_t *chan_list, struct gsm_bts *bts)
 {
-	struct generate_bcch_chan_list__ni_iter_data *data = cb_data;
+	const struct bitvec *bv = &bts->si_common.cell_alloc;
 
-	if (key->from_bts != NEIGHBOR_IDENT_KEY_ANY_BTS
-	    && key->from_bts != data->bts->nr)
-		return true;
-
-	bitvec_set_bit_pos(data->bv, key->arfcn, 1);
-	return true;
+	/* generate a Frequency List from the Cell Allocation */
+	return bitvec2freq_list(chan_list, bv, bts, false, false);
 }
 
 /*! generate a cell channel list as per Section 10.5.2.22 of 04.08
@@ -646,7 +658,7 @@ static int generate_bcch_chan_list(uint8_t *chan_list, struct gsm_bts *bts,
 		/* Zero-initialize the bit-vector */
 		memset(bv->data, 0, bv->data_len);
 
-		if (llist_empty(&bts->local_neighbors)) {
+		if (llist_empty(&bts->neighbors)) {
 			/* There are no explicit neighbors, assume all BTS are. */
 			llist_for_each_entry(cur_bts, &bts->network->bts_list, list) {
 				if (cur_bts == bts)
@@ -655,21 +667,23 @@ static int generate_bcch_chan_list(uint8_t *chan_list, struct gsm_bts *bts,
 			}
 		} else {
 			/* Only add explicit neighbor cells */
-			struct gsm_bts_ref *neigh;
-			llist_for_each_entry(neigh, &bts->local_neighbors, entry) {
-				bitvec_set_bit_pos(bv, neigh->bts->c0->arfcn, 1);
+			struct neighbor *n;
+			llist_for_each_entry(n, &bts->neighbors, entry) {
+				if (n->type == NEIGHBOR_TYPE_CELL_ID && n->cell_id.ab_present) {
+					bitvec_set_bit_pos(bv, n->cell_id.ab.arfcn, 1);
+				} else {
+					struct gsm_bts *neigh_bts;
+					if (resolve_local_neighbor(&neigh_bts, bts, n)) {
+						LOGP(DHO, LOGL_ERROR,
+						     "Neither local nor remote neighbor: BTS %u -> %s\n",
+						     bts->nr, neighbor_to_str_c(OTC_SELECT, n));
+						continue;
+					}
+					if (neigh_bts->c0)
+						bitvec_set_bit_pos(bv, neigh_bts->c0->arfcn, 1);
+				}
 			}
 		}
-
-		/* Also add neighboring BSS cells' ARFCNs */
-		{
-			struct generate_bcch_chan_list__ni_iter_data data = {
-				.bv = bv,
-				.bts = bts,
-			};
-			neighbor_ident_iter(bts->network->neighbor_bss_cells,
-					    generate_bcch_chan_list__ni_iter_cb, &data);
-		}
 	}
 
 	/* then we generate a GSM 04.08 frequency list from the bitvec */
@@ -1003,8 +1017,11 @@ static int generate_si4(enum osmo_sysinfo_type t, struct gsm_bts *bts)
 		const struct gsm_bts_trx_ts *ts = cbch_lchan->ts;
 		struct gsm48_chan_desc cd;
 
-		/* 10.5.2.5 (TV) CBCH Channel Description IE */
-		gsm48_lchan2chan_desc_as_configured(&cd, cbch_lchan);
+		/* 10.5.2.5 (TV) CBCH Channel Description IE.
+		 * CBCH is never in VAMOS mode, so just pass allow_osmo_cbits == false. */
+		if (gsm48_lchan_and_pchan2chan_desc(&cd, cbch_lchan, cbch_lchan->ts->pchan_from_config,
+						    gsm_ts_tsc(cbch_lchan->ts), false))
+			return -EINVAL;
 		tail = tv_fixed_put(tail, GSM48_IE_CBCH_CHAN_DESC,
 				    sizeof(cd), (uint8_t *) &cd);
 		l2_plen += 1 + sizeof(cd);
