@@ -26,7 +26,6 @@
 
 #include <osmocom/bsc/gsm_data.h>
 #include <osmocom/bsc/debug.h>
-#include <osmocom/bsc/arfcn_range_encode.h>
 #include <osmocom/bsc/system_information.h>
 #include <osmocom/bsc/abis_rsl.h>
 #include <osmocom/bsc/bts.h>
@@ -58,8 +57,6 @@
 			__FILE__, __LINE__, (int) res, # cmp, (int) wanted);	\
 	}
 
-
-
 static inline void gen(struct gsm_bts *bts, const char *s)
 {
 	int r;
@@ -273,338 +270,6 @@ static inline void test_si2q_long(struct gsm_network *net)
 	bts_del(bts);
 }
 
-struct {
-	int range;
-	int arfcns_num;
-	int arfcns[RANGE_ENC_MAX_ARFCNS];
-} arfcn_test_ranges[] = {
-	{ARFCN_RANGE_512, 12,
-		{ 1, 12, 31, 51, 57, 91, 97, 98, 113, 117, 120, 125 }},
-	{ARFCN_RANGE_512, 17,
-		{ 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17 }},
-	{ARFCN_RANGE_512, 18,
-		{ 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18 }},
-	{ARFCN_RANGE_512, 18,
-		{ 1, 17, 31, 45, 58, 79, 81, 97,
-		  113, 127, 213, 277, 287, 311, 331, 391,
-		  417, 511 }},
-	{ARFCN_RANGE_512, 6,
-		{ 1, 17, 31, 45, 58, 79 }},
-	{ARFCN_RANGE_512, 6,
-		{ 10, 17, 31, 45, 58, 79 }},
-	{ARFCN_RANGE_1024, 17,
-		{ 0, 17, 31, 45, 58, 79, 81, 97,
-		  113, 127, 213, 277, 287, 311, 331, 391,
-		  1023 }},
-	{ARFCN_RANGE_1024, 16,
-		{ 17, 31, 45, 58, 79, 81, 97, 113,
-		  127, 213, 277, 287, 311, 331, 391, 1023 }},
-	{-1}
-};
-
-static int test_single_range_encoding(int range, const int *orig_arfcns,
-				       int arfcns_num, int silent)
-{
-	int arfcns[RANGE_ENC_MAX_ARFCNS];
-	int w[RANGE_ENC_MAX_ARFCNS];
-	int f0_included = 0;
-	int rc, f0;
-	uint8_t chan_list[16] = {0};
-	struct gsm_sysinfo_freq dec_freq[1024] = {{0}};
-	int dec_arfcns[RANGE_ENC_MAX_ARFCNS] = {0};
-	int dec_arfcns_count = 0;
-	int arfcns_used = 0;
-	int i;
-
-	arfcns_used = arfcns_num;
-	memmove(arfcns, orig_arfcns, sizeof(arfcns));
-
-	f0 = range == ARFCN_RANGE_1024 ? 0 : arfcns[0];
-	/*
-	 * Manipulate the ARFCN list according to the rules in J4 depending
-	 * on the selected range.
-	 */
-	arfcns_used = range_enc_filter_arfcns(arfcns, arfcns_used,
-					      f0, &f0_included);
-
-	memset(w, 0, sizeof(w));
-	range_enc_arfcns(range, arfcns, arfcns_used, w, 0);
-
-	if (!silent)
-		fprintf(stderr, "range=%d, arfcns_used=%d, f0=%d, f0_included=%d\n",
-			range, arfcns_used, f0, f0_included);
-
-	/* Select the range and the amount of bits needed */
-	switch (range) {
-	case ARFCN_RANGE_128:
-		range_enc_range128(chan_list, f0, w);
-		break;
-	case ARFCN_RANGE_256:
-		range_enc_range256(chan_list, f0, w);
-		break;
-	case ARFCN_RANGE_512:
-		range_enc_range512(chan_list, f0, w);
-		break;
-	case ARFCN_RANGE_1024:
-		range_enc_range1024(chan_list, f0, f0_included, w);
-		break;
-	default:
-		return 1;
-	};
-
-	if (!silent)
-		printf("chan_list = %s\n",
-		       osmo_hexdump(chan_list, sizeof(chan_list)));
-
-	rc = gsm48_decode_freq_list(dec_freq, chan_list, sizeof(chan_list),
-				    0xfe, 1);
-	if (rc != 0) {
-		printf("Cannot decode freq list, rc = %d\n", rc);
-		return 1;
-	}
-
-	for (i = 0; i < ARRAY_SIZE(dec_freq); i++) {
-		if (dec_freq[i].mask &&
-		    dec_arfcns_count < ARRAY_SIZE(dec_arfcns))
-			dec_arfcns[dec_arfcns_count++] = i;
-	}
-
-	if (!silent) {
-		printf("Decoded freqs %d (expected %d)\n",
-		       dec_arfcns_count, arfcns_num);
-		printf("Decoded: ");
-		for (i = 0; i < dec_arfcns_count; i++) {
-			printf("%d ", dec_arfcns[i]);
-			if (dec_arfcns[i] != orig_arfcns[i])
-				printf("(!= %d) ", orig_arfcns[i]);
-		}
-		printf("\n");
-	}
-
-	if (dec_arfcns_count != arfcns_num) {
-		printf("Wrong number of arfcns\n");
-		return 1;
-	}
-
-	if (memcmp(dec_arfcns, orig_arfcns, sizeof(dec_arfcns)) != 0) {
-		printf("Decoding error, got wrong freqs\n");
-		fprintf(stderr, " w = ");
-		for (i = 0; i < ARRAY_SIZE(w); i++)
-			fprintf(stderr, "%d ", w[i]);
-		fprintf(stderr, "\n");
-		return 1;
-	}
-
-	return 0;
-}
-
-static void test_random_range_encoding(int range, int max_arfcn_num)
-{
-	int arfcns_num = 0;
-	int test_idx;
-	int rc, max_count;
-	int num_tests = 1024;
-
-	printf("Random range test: range %d, max num ARFCNs %d\n",
-	       range, max_arfcn_num);
-
-	srandom(1);
-
-	for (max_count = 1; max_count < max_arfcn_num; max_count++) {
-		for (test_idx = 0; test_idx < num_tests; test_idx++) {
-			int count;
-			int i;
-			int min_freq = 0;
-
-			int rnd_arfcns[RANGE_ENC_MAX_ARFCNS] = {0};
-			char rnd_arfcns_set[1024] = {0};
-
-			if (range < ARFCN_RANGE_1024)
-				min_freq = random() % (1023 - range);
-
-			for (count = max_count; count; ) {
-				int arfcn = min_freq + random() % (range + 1);
-				OSMO_ASSERT(arfcn < ARRAY_SIZE(rnd_arfcns_set));
-
-				if (!rnd_arfcns_set[arfcn]) {
-					rnd_arfcns_set[arfcn] = 1;
-					count -= 1;
-				}
-			}
-
-			arfcns_num = 0;
-			for (i = 0; i < ARRAY_SIZE(rnd_arfcns_set); i++)
-				if (rnd_arfcns_set[i])
-					rnd_arfcns[arfcns_num++] = i;
-
-			rc = test_single_range_encoding(range, rnd_arfcns,
-							arfcns_num, 1);
-			if (rc != 0) {
-				printf("Failed on test %d, range %d, num ARFCNs %d\n",
-				       test_idx, range, max_count);
-				test_single_range_encoding(range, rnd_arfcns,
-							   arfcns_num, 0);
-				return;
-			}
-		}
-	}
-}
-
-static void test_range_encoding()
-{
-	int *arfcns;
-	int arfcns_num = 0;
-	int test_idx;
-	int range;
-
-	for (test_idx = 0; arfcn_test_ranges[test_idx].arfcns_num > 0; test_idx++)
-	{
-		arfcns_num = arfcn_test_ranges[test_idx].arfcns_num;
-		arfcns = &arfcn_test_ranges[test_idx].arfcns[0];
-		range = arfcn_test_ranges[test_idx].range;
-
-		printf("Range test %d: range %d, num ARFCNs %d\n",
-		       test_idx, range, arfcns_num);
-
-		test_single_range_encoding(range, arfcns, arfcns_num, 0);
-	}
-
-	test_random_range_encoding(ARFCN_RANGE_128, 29);
-	test_random_range_encoding(ARFCN_RANGE_256, 22);
-	test_random_range_encoding(ARFCN_RANGE_512, 18);
-	test_random_range_encoding(ARFCN_RANGE_1024, 16);
-}
-
-static int freqs1[] = {
-	12, 70, 121, 190, 250, 320, 401, 475, 520, 574, 634, 700, 764, 830, 905, 980
-};
-
-static int freqs2[] = {
-	402, 460, 1, 67, 131, 197, 272, 347,
-};
-
-static int freqs3[] = {
-	68, 128, 198, 279, 353, 398, 452,
-
-};
-
-static int w_out[] = {
-	122, 2, 69, 204, 75, 66, 60, 70, 83, 3, 24, 67, 54, 64, 70, 9,
-};
-
-static int range128[] = {
-	1, 1 + 127,
-};
-
-static int range256[] = {
-	1, 1 + 128,
-};
-
-static int range512[] = {
-	1, 1+ 511,
-};
-
-
-static void test_arfcn_filter()
-{
-	int arfcns[50], i, res, f0_included;
-	for (i = 0; i < ARRAY_SIZE(arfcns); ++i)
-		arfcns[i] = (i + 1) * 2;
-
-	/* check that the arfcn is taken out. f0_included is only set for Range1024 */
-	f0_included = 24;
-	res = range_enc_filter_arfcns(arfcns, ARRAY_SIZE(arfcns),
-			arfcns[0], &f0_included);
-	VERIFY(res, ==, ARRAY_SIZE(arfcns) - 1);
-	VERIFY(f0_included, ==, 1);
-	for (i = 0; i < res; ++i)
-		VERIFY(arfcns[i], ==, ((i+2) * 2) - (2+1));
-
-	/* check with range1024, ARFCN 0 is included */
-	for (i = 0; i < ARRAY_SIZE(arfcns); ++i)
-		arfcns[i] = i * 2;
-	res = range_enc_filter_arfcns(arfcns, ARRAY_SIZE(arfcns),
-			0, &f0_included);
-	VERIFY(res, ==, ARRAY_SIZE(arfcns) - 1);
-	VERIFY(f0_included, ==, 1);
-	for (i = 0; i < res; ++i)
-		VERIFY(arfcns[i], ==, (i + 1) * 2 - 1);
-
-	/* check with range1024, ARFCN 0 not included */
-	for (i = 0; i < ARRAY_SIZE(arfcns); ++i)
-		arfcns[i] = (i + 1) * 2;
-	res = range_enc_filter_arfcns(arfcns, ARRAY_SIZE(arfcns),
-			0, &f0_included);
-	VERIFY(res, ==, ARRAY_SIZE(arfcns));
-	VERIFY(f0_included, ==, 0);
-	for (i = 0; i < res; ++i)
-		VERIFY(arfcns[i], ==, ((i + 1) * 2) - 1);
-}
-
-static void test_print_encoding()
-{
-	int rc;
-	int w[17];
-	uint8_t chan_list[16];
-	memset(chan_list, 0x23, sizeof(chan_list));
-
-	for (rc = 0; rc < ARRAY_SIZE(w); ++rc)
-		switch (rc % 3) {
-		case 0:
-			w[rc] = 0xAAAA;
-			break;
-		case 1:
-			w[rc] = 0x5555;
-			break;
-		case 2:
-			w[rc] = 0x9696;
-			break;
-		}
-
-	range_enc_range512(chan_list, (1 << 9) | 0x96, w);
-
-	printf("Range512: %s\n", osmo_hexdump(chan_list, ARRAY_SIZE(chan_list)));
-}
-
-static void test_si_range_helpers()
-{
-	int ws[(sizeof(freqs1)/sizeof(freqs1[0]))];
-	int i, f0 = 0xFFFFFF;
-
-	memset(&ws[0], 0x23, sizeof(ws));
-
-	i = range_enc_find_index(1023, freqs1, ARRAY_SIZE(freqs1));
-	printf("Element is: %d => freqs[i] = %d\n", i, i >= 0 ? freqs1[i] : -1);
-	VERIFY(i, ==, 2);
-
-	i = range_enc_find_index(511, freqs2, ARRAY_SIZE(freqs2));
-	printf("Element is: %d => freqs[i] = %d\n", i,  i >= 0 ? freqs2[i] : -1);
-	VERIFY(i, ==, 2);
-
-	i = range_enc_find_index(511, freqs3, ARRAY_SIZE(freqs3));
-	printf("Element is: %d => freqs[i] = %d\n", i,  i >= 0 ? freqs3[i] : -1);
-	VERIFY(i, ==, 0);
-
-	range_enc_arfcns(1023, freqs1, ARRAY_SIZE(freqs1), ws, 0);
-
-	for (i = 0; i < sizeof(freqs1)/sizeof(freqs1[0]); ++i) {
-		printf("w[%d]=%d\n", i, ws[i]);
-		VERIFY(ws[i], ==, w_out[i]);
-	}
-
-	i = range_enc_determine_range(range128, ARRAY_SIZE(range128), &f0);
-	VERIFY(i, ==, ARFCN_RANGE_128);
-	VERIFY(f0, ==, 1);
-
-	i = range_enc_determine_range(range256, ARRAY_SIZE(range256), &f0);
-	VERIFY(i, ==, ARFCN_RANGE_256);
-	VERIFY(f0, ==, 1);
-
-	i = range_enc_determine_range(range512, ARRAY_SIZE(range512), &f0);
-	VERIFY(i, ==, ARFCN_RANGE_512);
-	VERIFY(f0, ==, 1);
-}
-
 static void test_si_ba_ind(struct gsm_network *net)
 {
 	struct gsm_bts *bts = bts_init(net);
@@ -779,10 +444,10 @@ static void test_gsm48_ra_id_by_bts()
 
 static void test_gsm48_multirate_config()
 {
-	uint8_t lv[7];
 	struct gsm48_multi_rate_conf *gsm48_ie;
 	struct amr_multirate_conf mr;
 	int rc;
+	struct msgb *msg = msgb_alloc(32, "test_gsm48_multirate_config");
 
 	memset(&mr, 0, sizeof(mr));
 
@@ -806,75 +471,88 @@ static void test_gsm48_multirate_config()
 	mr.ms_mode[1].mode = 4;
 	mr.ms_mode[2].mode = 5;
 	mr.ms_mode[3].mode = 7;
-	rc = gsm48_multirate_config(lv, gsm48_ie, mr.ms_mode, 4);
+	msgb_trim(msg, 0);
+	rc = gsm48_multirate_config(msg, gsm48_ie, mr.ms_mode, 4);
 	OSMO_ASSERT(rc == 0);
 	printf("gsm48_multirate_config(): rc=%i, lv=%s\n", rc,
-	       osmo_hexdump_nospc(lv, lv[0]));
+	       osmo_hexdump_nospc(msg->data, msg->len));
 
 	/* Test #2: 4 active set members, but wrong mode order: */
 	mr.ms_mode[3].mode = 2;
 	mr.ms_mode[2].mode = 4;
 	mr.ms_mode[1].mode = 5;
 	mr.ms_mode[0].mode = 7;
-	rc = gsm48_multirate_config(lv, gsm48_ie, mr.ms_mode, 4);
+	msgb_trim(msg, 0);
+	rc = gsm48_multirate_config(msg, gsm48_ie, mr.ms_mode, 4);
 	OSMO_ASSERT(rc == -EINVAL);
 
 	/* Test #3: Normal configuration with 3 active set members */
 	mr.ms_mode[0].mode = 2;
 	mr.ms_mode[1].mode = 4;
 	mr.ms_mode[2].mode = 5;
-	mr.ms_mode[3].mode = 0;
+	mr.ms_mode[3].mode = 7;
 	gsm48_ie->m12_2 = 0;
 	mr.ms_mode[2].threshold = 0;
 	mr.ms_mode[2].hysteresis = 0;
 
-	rc = gsm48_multirate_config(lv, gsm48_ie, mr.ms_mode, 3);
+	msgb_trim(msg, 0);
+	rc = gsm48_multirate_config(msg, gsm48_ie, mr.ms_mode, 3);
 	OSMO_ASSERT(rc == 0);
 	printf("gsm48_multirate_config(): rc=%i, lv=%s\n", rc,
-	       osmo_hexdump_nospc(lv, lv[0]));
+	       osmo_hexdump_nospc(msg->data, msg->len));
 
 	/* Test #4: 3 active set members, but wrong mode order: */
 	mr.ms_mode[0].mode = 2;
 	mr.ms_mode[2].mode = 4;
 	mr.ms_mode[1].mode = 5;
-	rc = gsm48_multirate_config(lv, gsm48_ie, mr.ms_mode, 3);
+	msgb_trim(msg, 0);
+	rc = gsm48_multirate_config(msg, gsm48_ie, mr.ms_mode, 3);
 	OSMO_ASSERT(rc == -EINVAL);
 
 	/* Test #5: Normal configuration with 2 active set members */
 	mr.ms_mode[0].mode = 2;
 	mr.ms_mode[1].mode = 4;
-	mr.ms_mode[2].mode = 0;
+	mr.ms_mode[2].mode = 5;
+	mr.ms_mode[3].mode = 7;
 	gsm48_ie->m7_95 = 0;
 	mr.ms_mode[1].threshold = 0;
 	mr.ms_mode[1].hysteresis = 0;
 
-	rc = gsm48_multirate_config(lv, gsm48_ie, mr.ms_mode, 2);
+	msgb_trim(msg, 0);
+	rc = gsm48_multirate_config(msg, gsm48_ie, mr.ms_mode, 2);
 	OSMO_ASSERT(rc == 0);
 	printf("gsm48_multirate_config(): rc=%i, lv=%s\n", rc,
-	       osmo_hexdump_nospc(lv, lv[0]));
+	       osmo_hexdump_nospc(msg->data, msg->len));
 
 	/* Test #6: 2 active set members, but wrong mode order: */
 	mr.ms_mode[1].mode = 2;
 	mr.ms_mode[0].mode = 4;
-	rc = gsm48_multirate_config(lv, gsm48_ie, mr.ms_mode, 2);
+	msgb_trim(msg, 0);
+	rc = gsm48_multirate_config(msg, gsm48_ie, mr.ms_mode, 2);
 	OSMO_ASSERT(rc == -EINVAL);
 
 	/* Test #7: Normal configuration with 1 active set member */
 	mr.ms_mode[0].mode = 2;
-	mr.ms_mode[1].mode = 0;
+	mr.ms_mode[1].mode = 4;
+	mr.ms_mode[2].mode = 5;
+	mr.ms_mode[3].mode = 7;
 	gsm48_ie->m7_40 = 0;
 	mr.ms_mode[0].threshold = 0;
 	mr.ms_mode[0].hysteresis = 0;
 
-	rc = gsm48_multirate_config(lv, gsm48_ie, mr.ms_mode, 1);
+	msgb_trim(msg, 0);
+	rc = gsm48_multirate_config(msg, gsm48_ie, mr.ms_mode, 1);
 	OSMO_ASSERT(rc == 0);
 	printf("gsm48_multirate_config(): rc=%i, lv=%s\n", rc,
-	       osmo_hexdump_nospc(lv, lv[0]));
+	       osmo_hexdump_nospc(msg->data, msg->len));
 
 	/* Test #8: 0 active set members: */
 	mr.ms_mode[0].mode = 0;
-	rc = gsm48_multirate_config(lv, gsm48_ie, mr.ms_mode, 1);
+	msgb_trim(msg, 0);
+	rc = gsm48_multirate_config(msg, gsm48_ie, mr.ms_mode, 1);
 	OSMO_ASSERT(rc == -EINVAL);
+
+	msgb_free(msg);
 }
 
 static const struct log_info_cat log_categories[] = {
@@ -899,11 +577,7 @@ int main(int argc, char **argv)
 		printf("Network init failure.\n");
 		return EXIT_FAILURE;
 	}
-
-	test_si_range_helpers();
-	test_arfcn_filter();
-	test_print_encoding();
-	test_range_encoding();
+	bsc_gsmnet = net;
 
 	test_si2q_segfault(net);
 	test_si2q_e(net);
@@ -921,59 +595,3 @@ int main(int argc, char **argv)
 
 	return EXIT_SUCCESS;
 }
-
-struct gsm_subscriber_connection *bsc_subscr_con_allocate(struct gsm_network *net) {
-	OSMO_ASSERT(0);
-}
-
-void gscon_release_lchans(struct gsm_subscriber_connection *conn, bool do_rr_release) {
-	OSMO_ASSERT(0);
-}
-
-bool on_gsm_ts_init(struct gsm_bts_trx_ts *ts)
-{
-	return true;
-}
-
-void ts_fsm_alloc(struct gsm_bts_trx_ts *ts) {}
-
-void bsc_cm_update(struct gsm_subscriber_connection *conn,
-		   const uint8_t *cm2, uint8_t cm2_len,
-		   const uint8_t *cm3, uint8_t cm3_len) {}
-
-int rsl_siemens_mrpci(struct gsm_lchan *lchan, struct rsl_mrpci *mrpci)
-{ return 0; }
-
-int rsl_chan_mode_modify_req(struct gsm_lchan *ts) { return 0; }
-
-int rsl_tx_ipacc_crcx(const struct gsm_lchan *lchan) { return 0; }
-
-void gscon_submit_rsl_dtap(struct gsm_subscriber_connection *conn,
-			   struct msgb *msg, int link_id, int allow_sacch) {}
-
-bool lchan_may_receive_data(struct gsm_lchan *lchan) { return  true; }
-
-int bsc_compl_l3(struct gsm_lchan *lchan, struct msgb *msg, uint16_t chosen_channel)
-{ return 0; }
-
-void bsc_dtap(struct gsm_subscriber_connection *conn, uint8_t link_id,
-	      struct msgb *msg) {}
-
-void bsc_cipher_mode_compl(struct gsm_subscriber_connection *conn,
-			   struct msgb *msg, uint8_t chosen_encr) {}
-
-const char *bsc_subscr_name(struct bsc_subscr *bsub) { return NULL; }
-
-void lchan_release(struct gsm_lchan *lchan, bool do_rr_release,
-		   bool err, enum gsm48_rr_cause cause_rr) {}
-
-int rsl_data_request(struct msgb *msg, uint8_t link_id) { return 0; }
-
-int rsl_encryption_cmd(struct msgb *msg) { return 0; }
-
-int rsl_chan_ms_power_ctrl(struct gsm_lchan *lchan) { return 0; }
-
-void pcu_info_update(struct gsm_bts *bts) {};
-int rsl_sacch_filling(struct gsm_bts_trx *trx, uint8_t type, const uint8_t *data, int len) { return 0; }
-int rsl_bcch_info(const struct gsm_bts_trx *trx, enum osmo_sysinfo_type si_type, const uint8_t *data, int len)
-{ return 0; }
