@@ -35,6 +35,7 @@
 #include <osmocom/bsc/bsc_subscriber.h>
 #include <osmocom/bsc/lchan_select.h>
 #include <osmocom/bsc/lchan_fsm.h>
+#include <osmocom/bsc/assignment_fsm.h>
 #include <osmocom/bsc/handover_decision.h>
 #include <osmocom/bsc/system_information.h>
 #include <osmocom/bsc/handover.h>
@@ -51,13 +52,12 @@
 #include <osmocom/bsc/bts.h>
 #include <osmocom/bsc/paging.h>
 #include <osmocom/bsc/vty.h>
+#include <osmocom/mgcp_client/mgcp_client_pool.h>
 
 #include "../../bscconfig.h"
 
 void *ctx;
 
-struct gsm_network *bsc_gsmnet;
-
 /* override, requires '-Wl,--wrap=osmo_mgcpc_ep_ci_request'.
  * Catch modification of an MGCP connection. */
 void __real_osmo_mgcpc_ep_ci_request(struct osmo_mgcpc_ep_ci *ci,
@@ -82,7 +82,7 @@ void __wrap_osmo_mgcpc_ep_ci_request(struct osmo_mgcpc_ep_ci *ci,
 
 uint8_t meas_rep_ba = 0, meas_rep_valid = 1, meas_valid = 1, meas_multi_rep = 0;
 uint8_t meas_ul_rxlev = 0, meas_ul_rxqual = 0;
-uint8_t meas_tx_power_ms = 0, meas_tx_power_bs = 0;
+uint8_t meas_tx_power_ms = 0;
 uint8_t meas_dtx_ms = 0, meas_dtx_bs = 0, meas_nr = 0;
 char *codec_tch_f = NULL;
 char *codec_tch_h = NULL;
@@ -93,16 +93,52 @@ struct neighbor_meas {
 	uint8_t bcch_f;
 };
 
+const struct timeval fake_time_start_time = { 123, 456 };
+
+void fake_time_passes(time_t secs, suseconds_t usecs)
+{
+	struct timeval diff;
+	/* Add time to osmo_fsm timers, using osmo_gettimeofday() */
+	osmo_gettimeofday_override_add(secs, usecs);
+	/* Add time to penalty timers, using osmo_clock_gettime() */
+	osmo_clock_override_add(CLOCK_MONOTONIC, secs, usecs * 1000);
+
+	timersub(&osmo_gettimeofday_override_time, &fake_time_start_time, &diff);
+	fprintf(stderr, "Total time passed: %d.%06d s\n", (int)diff.tv_sec, (int)diff.tv_usec);
+
+	osmo_timers_prepare();
+	osmo_timers_update();
+}
+
+void fake_time_start()
+{
+	struct timespec *clock_override;
+
+	/* osmo_fsm uses osmo_gettimeofday(). To affect FSM timeouts, we need osmo_gettimeofday_override. */
+	osmo_gettimeofday_override_time = fake_time_start_time;
+	osmo_gettimeofday_override = true;
+
+	/* Penalty timers use osmo_clock_gettime(CLOCK_MONOTONIC). To affect these timeouts, we need
+	 * osmo_gettimeofday_override. */
+	clock_override = osmo_clock_override_gettimespec(CLOCK_MONOTONIC);
+	OSMO_ASSERT(clock_override);
+	clock_override->tv_sec = fake_time_start_time.tv_sec;
+	clock_override->tv_nsec = fake_time_start_time.tv_usec * 1000;
+	osmo_clock_override_enable(CLOCK_MONOTONIC, true);
+	fake_time_passes(0, 0);
+}
+
 static void gen_meas_rep(struct gsm_lchan *lchan,
-			 uint8_t rxlev, uint8_t rxqual, uint8_t ta,
+			 uint8_t bs_power_db, uint8_t rxlev, uint8_t rxqual, uint8_t ta,
 			 int neighbors_count, struct neighbor_meas *neighbors)
 {
 	struct msgb *msg = msgb_alloc_headroom(256, 64, "RSL");
 	struct abis_rsl_dchan_hdr *dh;
-	uint8_t chan_nr = gsm_lchan2chan_nr(lchan);
 	uint8_t ulm[3], l1i[2], *buf;
 	struct gsm48_hdr *gh;
 	struct gsm48_meas_res *mr;
+	int chan_nr = gsm_lchan2chan_nr(lchan, true);
+	OSMO_ASSERT(chan_nr >= 0);
 
 	dh = (struct abis_rsl_dchan_hdr *) msgb_put(msg, sizeof(*dh));
 	dh->c.msg_discr = ABIS_RSL_MDISC_DED_CHAN;
@@ -117,7 +153,7 @@ static void gen_meas_rep(struct gsm_lchan *lchan,
 	ulm[2] = (meas_ul_rxqual << 3) | meas_ul_rxqual;
 	msgb_tlv_put(msg, RSL_IE_UPLINK_MEAS, sizeof(ulm), ulm);
 
-	msgb_tv_put(msg, RSL_IE_BS_POWER, meas_tx_power_bs);
+	msgb_tv_put(msg, RSL_IE_BS_POWER, (bs_power_db / 2) & 0xf);
 
 	l1i[0] = 0;
 	l1i[1] = ta;
@@ -174,7 +210,7 @@ static void gen_meas_rep(struct gsm_lchan *lchan,
 	mr->bcch_f_nc6_hi = neighbors[5].bcch_f >> 2;
 	mr->bcch_f_nc6_lo = neighbors[5].bcch_f & 3;
 
-	msg->dst = lchan->ts->trx->bts->c0->rsl_link;
+	msg->dst = rsl_chan_link(lchan);
 	msg->l2h = (unsigned char *)dh;
 	msg->l3h = (unsigned char *)gh;
 
@@ -185,7 +221,7 @@ enum gsm_phys_chan_config pchan_from_str(const char *str)
 {
 	enum gsm_phys_chan_config pchan;
 	if (!strcmp(str, "dyn"))
-		return GSM_PCHAN_TCH_F_TCH_H_PDCH;
+		return GSM_PCHAN_OSMO_DYN;
 	if (!strcmp(str, "c+s4"))
 		return GSM_PCHAN_CCCH_SDCCH4;
 	if (!strcmp(str, "-"))
@@ -205,6 +241,7 @@ const char * const bts_default_ts[] = {
 static struct gsm_bts *_create_bts(int num_trx, const char * const *ts_args, int ts_args_count)
 {
 	static int arfcn = 870;
+	static int ci = 0;
 	struct gsm_bts *bts;
 	struct e1inp_sign_link *rsl_link;
 	int i;
@@ -220,6 +257,7 @@ static struct gsm_bts *_create_bts(int num_trx, const char * const *ts_args, int
 	}
 
 	bts->location_area_code = 23;
+	bts->cell_identity = ci++;
 	bts->c0->arfcn = arfcn++;
 
 	bts->codec.efr = 1;
@@ -228,7 +266,7 @@ static struct gsm_bts *_create_bts(int num_trx, const char * const *ts_args, int
 
 	rsl_link = talloc_zero(ctx, struct e1inp_sign_link);
 	rsl_link->trx = bts->c0;
-	bts->c0->rsl_link = rsl_link;
+	bts->c0->rsl_link_primary = rsl_link;
 
 	for (trx_i = 0; trx_i < num_trx; trx_i++) {
 		while (!(trx = gsm_bts_trx_num(bts, trx_i)))
@@ -266,9 +304,9 @@ static struct gsm_bts *_create_bts(int num_trx, const char * const *ts_args, int
 
 			/* Unused dyn TS start out as used for PDCH */
 			switch (trx->ts[i].pchan_on_init) {
-			case GSM_PCHAN_TCH_F_TCH_H_PDCH:
+			case GSM_PCHAN_OSMO_DYN:
 			case GSM_PCHAN_TCH_F_PDCH:
-				trx->ts[i].pchan_is = GSM_PCHAN_PDCH;
+				ts_set_pchan_is(&trx->ts[i], GSM_PCHAN_PDCH);
 				break;
 			default:
 				break;
@@ -283,6 +321,34 @@ static struct gsm_bts *_create_bts(int num_trx, const char * const *ts_args, int
 	return bts;
 }
 
+char *lchans_use_str(struct gsm_bts_trx_ts *ts, const char *established_prefix, char established_char)
+{
+	char state_chars[8] = { 0 };
+	struct gsm_lchan *lchan;
+	bool any_lchans_established = false;
+	bool any_lchans_in_use = false;
+	ts_for_n_lchans(lchan, ts, ts->max_primary_lchans) {
+		char state_char;
+		if (lchan_state_is(lchan, LCHAN_ST_UNUSED)) {
+			state_char = '-';
+		} else {
+			any_lchans_in_use = true;
+			if (lchan_state_is(lchan, LCHAN_ST_ESTABLISHED)) {
+				any_lchans_established = true;
+				state_char = established_char;
+			} else {
+				state_char = '!';
+			}
+		}
+		state_chars[lchan->nr] = state_char;
+	}
+	if (!any_lchans_in_use)
+		return "-";
+	if (!any_lchans_established)
+		established_prefix = "";
+	return talloc_asprintf(OTC_SELECT, "%s%s", established_prefix, state_chars);
+}
+
 const char *ts_use_str(struct gsm_bts_trx_ts *ts)
 {
 	switch (ts->pchan_is) {
@@ -293,20 +359,10 @@ const char *ts_use_str(struct gsm_bts_trx_ts *ts)
 		return "-";
 
 	case GSM_PCHAN_TCH_F:
-		if (lchan_state_is(&ts->lchan[0], LCHAN_ST_ESTABLISHED))
-			return "TCH/F";
-		else
-			return "-";
+		return lchans_use_str(ts, "TCH/", 'F');
 
 	case GSM_PCHAN_TCH_H:
-		if (lchan_state_is(&ts->lchan[0], LCHAN_ST_ESTABLISHED)
-		    && lchan_state_is(&ts->lchan[1], LCHAN_ST_ESTABLISHED))
-			return "TCH/HH";
-		if (lchan_state_is(&ts->lchan[0], LCHAN_ST_ESTABLISHED))
-			return "TCH/H-";
-		if (lchan_state_is(&ts->lchan[1], LCHAN_ST_ESTABLISHED))
-			return "TCH/-H";
-		return "-";
+		return lchans_use_str(ts, "TCH/", 'H');
 
 	default:
 		return gsm_pchan_name(ts->pchan_is);
@@ -370,6 +426,10 @@ void create_conn(struct gsm_lchan *lchan)
 	snprintf(imsi, sizeof(imsi), "%06u", next_imsi);
 	lchan->conn->bsub = bsc_subscr_find_or_create_by_imsi(net->bsc_subscribers, imsi, BSUB_USE_CONN);
 
+	/* Set RTP data that the MSC normally would have sent */
+	OSMO_STRLCPY_ARRAY(conn->user_plane.msc_assigned_rtp_addr, "1.2.3.4");
+	conn->user_plane.msc_assigned_rtp_port = 1234;
+
 	/* kick the FSM from INIT through to the ACTIVE state */
 	osmo_fsm_inst_dispatch(conn->fi, GSCON_EV_MO_COMPL_L3, NULL);
 	osmo_fsm_inst_dispatch(conn->fi, GSCON_EV_A_CONN_CFM, NULL);
@@ -384,25 +444,25 @@ struct gsm_lchan *lchan_act(struct gsm_lchan *lchan, int full_rate, const char *
 	/* Fake osmo_mgcpc_ep_ci to indicate that the lchan is used for voice */
 	lchan->mgw_endpoint_ci_bts = (void*)1;
 
-	if (lchan->ts->pchan_on_init == GSM_PCHAN_TCH_F_TCH_H_PDCH)
-		lchan->ts->pchan_is = full_rate ? GSM_PCHAN_TCH_F : GSM_PCHAN_TCH_H;
+	if (lchan->ts->pchan_on_init == GSM_PCHAN_OSMO_DYN)
+		ts_set_pchan_is(lchan->ts, full_rate ? GSM_PCHAN_TCH_F : GSM_PCHAN_TCH_H);
 	if (lchan->ts->pchan_on_init == GSM_PCHAN_TCH_F_PDCH) {
 		OSMO_ASSERT(full_rate);
-		lchan->ts->pchan_is = GSM_PCHAN_TCH_F;
+		ts_set_pchan_is(lchan->ts, GSM_PCHAN_TCH_F);
 	}
 
 	LOG_LCHAN(lchan, LOGL_DEBUG, "activated by handover_test.c\n");
 
 	create_conn(lchan);
 	if (!strcasecmp(codec, "FR") && full_rate)
-		lchan->tch_mode = GSM48_CMODE_SPEECH_V1;
+		lchan->current_ch_mode_rate.chan_mode = GSM48_CMODE_SPEECH_V1;
 	else if (!strcasecmp(codec, "HR") && !full_rate)
-		lchan->tch_mode = GSM48_CMODE_SPEECH_V1;
+		lchan->current_ch_mode_rate.chan_mode = GSM48_CMODE_SPEECH_V1;
 	else if (!strcasecmp(codec, "EFR") && full_rate)
-		lchan->tch_mode = GSM48_CMODE_SPEECH_EFR;
+		lchan->current_ch_mode_rate.chan_mode = GSM48_CMODE_SPEECH_EFR;
 	else if (!strcasecmp(codec, "AMR")) {
-		lchan->tch_mode = GSM48_CMODE_SPEECH_AMR;
-		lchan->activate.info.s15_s0 = 0x0002;
+		lchan->current_ch_mode_rate.chan_mode = GSM48_CMODE_SPEECH_AMR;
+		lchan->current_ch_mode_rate.s15_s0 = 0x0002;
 	} else {
 		fprintf(stderr, "Given codec unknown\n");
 		exit(EXIT_FAILURE);
@@ -437,24 +497,27 @@ struct gsm_lchan *create_lchan(struct gsm_bts *bts, int full_rate, const char *c
 
 static void lchan_release_ack(struct gsm_lchan *lchan)
 {
-	if (lchan->fi && lchan->fi->state == LCHAN_ST_WAIT_BEFORE_RF_RELEASE) {
-		/* don't wait before release */
-		osmo_fsm_inst_state_chg(lchan->fi, LCHAN_ST_WAIT_RF_RELEASE_ACK, 0, 0);
-		/* ack the release */
-		osmo_fsm_inst_dispatch(lchan->fi, LCHAN_EV_RSL_RF_CHAN_REL_ACK, 0);
-	}
+	if (!lchan->fi || lchan->fi->state != LCHAN_ST_WAIT_BEFORE_RF_RELEASE)
+		return;
+	/* don't wait before release */
+	osmo_fsm_inst_state_chg(lchan->fi, LCHAN_ST_WAIT_RF_RELEASE_ACK, 0, 0);
+	if (lchan->fi->state == LCHAN_ST_UNUSED)
+		return;
+	/* ack the release */
+	osmo_fsm_inst_dispatch(lchan->fi, LCHAN_EV_RSL_RF_CHAN_REL_ACK, 0);
 }
 
 static void lchan_clear(struct gsm_lchan *lchan)
 {
-	lchan_release(lchan, true, false, 0);
+	lchan_release(lchan, true, false, 0, NULL);
 	lchan_release_ack(lchan);
 }
 
 static void ts_clear(struct gsm_bts_trx_ts *ts)
 {
 	struct gsm_lchan *lchan;
-	ts_for_each_lchan(lchan, ts) {
+
+	ts_for_n_lchans(lchan, ts, ts->max_lchans_possible) {
 		if (lchan_state_is(lchan, LCHAN_ST_UNUSED))
 			continue;
 		lchan_clear(lchan);
@@ -528,8 +591,11 @@ bool _set_ts_use(struct gsm_bts *bts, struct gsm_bts_trx *trx, const char * cons
 static struct gsm_lchan *new_chan_req = NULL;
 static struct gsm_lchan *last_chan_req = NULL;
 
-static struct gsm_lchan *new_ho_req = NULL;
-static struct gsm_lchan *last_ho_req = NULL;
+static struct gsm_lchan *new_ho_cmd = NULL;
+static struct gsm_lchan *last_ho_cmd = NULL;
+
+static struct gsm_lchan *new_as_cmd = NULL;
+static struct gsm_lchan *last_as_cmd = NULL;
 
 /* send channel activation ack */
 static void send_chan_act_ack(struct gsm_lchan *lchan, int act)
@@ -541,20 +607,58 @@ static void send_chan_act_ack(struct gsm_lchan *lchan, int act)
 	dh->c.msg_discr = ABIS_RSL_MDISC_DED_CHAN;
 	dh->c.msg_type = (act) ? RSL_MT_CHAN_ACTIV_ACK : RSL_MT_RF_CHAN_REL_ACK;
 	dh->ie_chan = RSL_IE_CHAN_NR;
-	dh->chan_nr = gsm_lchan2chan_nr(lchan);
+	dh->chan_nr = gsm_lchan2chan_nr(lchan, true);
 
-	msg->dst = lchan->ts->trx->bts->c0->rsl_link;
+	msg->dst = rsl_chan_link(lchan);
 	msg->l2h = (unsigned char *)dh;
 
 	abis_rsl_rcvmsg(msg);
 }
 
+/* Send RR Assignment Complete for SAPI[0] */
+static void send_assignment_complete(struct gsm_lchan *lchan)
+{
+	struct msgb *msg = msgb_alloc_headroom(256, 64, "RSL");
+	struct abis_rsl_rll_hdr *rh;
+	uint8_t chan_nr = gsm_lchan2chan_nr(lchan, true);
+	uint8_t *buf;
+	struct gsm48_hdr *gh;
+	struct gsm48_ho_cpl *hc;
+
+	fprintf(stderr, "- Send RR Assignment Complete for %s\n", gsm_lchan_name(lchan));
+
+	rh = (struct abis_rsl_rll_hdr *) msgb_put(msg, sizeof(*rh));
+	rh->c.msg_discr = ABIS_RSL_MDISC_RLL;
+	rh->c.msg_type = RSL_MT_DATA_IND;
+	rh->ie_chan = RSL_IE_CHAN_NR;
+	rh->chan_nr = chan_nr;
+	rh->ie_link_id = RSL_IE_LINK_IDENT;
+	rh->link_id = 0x00;
+
+	buf = msgb_put(msg, 3);
+	buf[0] = RSL_IE_L3_INFO;
+	buf[1] = (sizeof(*gh) + sizeof(*hc)) >> 8;
+	buf[2] = (sizeof(*gh) + sizeof(*hc)) & 0xff;
+
+	gh = (struct gsm48_hdr *) msgb_put(msg, sizeof(*gh));
+	hc = (struct gsm48_ho_cpl *) msgb_put(msg, sizeof(*hc));
+
+	gh->proto_discr = GSM48_PDISC_RR;
+	gh->msg_type = GSM48_MT_RR_ASS_COMPL;
+
+	msg->dst = rsl_chan_link(lchan);
+	msg->l2h = (unsigned char *)rh;
+	msg->l3h = (unsigned char *)gh;
+
+	abis_rsl_rcvmsg(msg);
+}
+
 /* Send RLL Est Ind for SAPI[0] */
 static void send_est_ind(struct gsm_lchan *lchan)
 {
 	struct msgb *msg = msgb_alloc_headroom(256, 64, "RSL");
 	struct abis_rsl_rll_hdr *rh;
-	uint8_t chan_nr = gsm_lchan2chan_nr(lchan);
+	uint8_t chan_nr = gsm_lchan2chan_nr(lchan, true);
 
 	fprintf(stderr, "- Send EST IND for %s\n", gsm_lchan_name(lchan));
 
@@ -566,7 +670,7 @@ static void send_est_ind(struct gsm_lchan *lchan)
 	rh->ie_link_id = RSL_IE_LINK_IDENT;
 	rh->link_id = 0x00;
 
-	msg->dst = lchan->ts->trx->bts->c0->rsl_link;
+	msg->dst = rsl_chan_link(lchan);
 	msg->l2h = (unsigned char *)rh;
 
 	abis_rsl_rcvmsg(msg);
@@ -576,7 +680,7 @@ static void send_ho_detect(struct gsm_lchan *lchan)
 {
 	struct msgb *msg = msgb_alloc_headroom(256, 64, "RSL");
 	struct abis_rsl_rll_hdr *rh;
-	uint8_t chan_nr = gsm_lchan2chan_nr(lchan);
+	uint8_t chan_nr = gsm_lchan2chan_nr(lchan, true);
 
 	fprintf(stderr, "- Send HO DETECT for %s\n", gsm_lchan_name(lchan));
 
@@ -588,7 +692,7 @@ static void send_ho_detect(struct gsm_lchan *lchan)
 	rh->ie_link_id = RSL_IE_LINK_IDENT;
 	rh->link_id = 0x00;
 
-	msg->dst = lchan->ts->trx->bts->c0->rsl_link;
+	msg->dst = rsl_chan_link(lchan);
 	msg->l2h = (unsigned char *)rh;
 
 	abis_rsl_rcvmsg(msg);
@@ -602,7 +706,7 @@ static void send_ho_complete(struct gsm_lchan *lchan, bool success)
 {
 	struct msgb *msg = msgb_alloc_headroom(256, 64, "RSL");
 	struct abis_rsl_rll_hdr *rh;
-	uint8_t chan_nr = gsm_lchan2chan_nr(lchan);
+	uint8_t chan_nr = gsm_lchan2chan_nr(lchan, true);
 	uint8_t *buf;
 	struct gsm48_hdr *gh;
 	struct gsm48_ho_cpl *hc;
@@ -632,7 +736,7 @@ static void send_ho_complete(struct gsm_lchan *lchan, bool success)
 	gh->msg_type =
 		success ? GSM48_MT_RR_HANDO_COMPL : GSM48_MT_RR_HANDO_FAIL;
 
-	msg->dst = lchan->ts->trx->bts->c0->rsl_link;
+	msg->dst = rsl_chan_link(lchan);
 	msg->l2h = (unsigned char *)rh;
 	msg->l3h = (unsigned char *)gh;
 
@@ -670,7 +774,7 @@ int __wrap_abis_rsl_sendmsg(struct msgb *msg)
 
 		/* send dyn TS back to PDCH if unused */
 		switch (lchan->ts->pchan_on_init) {
-		case GSM_PCHAN_TCH_F_TCH_H_PDCH:
+		case GSM_PCHAN_OSMO_DYN:
 		case GSM_PCHAN_TCH_F_PDCH:
 			switch (lchan->ts->pchan_is) {
 			case GSM_PCHAN_TCH_H:
@@ -681,7 +785,7 @@ int __wrap_abis_rsl_sendmsg(struct msgb *msg)
 					break;
 				/* else fall thru */
 			case GSM_PCHAN_TCH_F:
-				lchan->ts->pchan_is = GSM_PCHAN_PDCH;
+				ts_set_pchan_is(lchan->ts, GSM_PCHAN_PDCH);
 				break;
 			default:
 				break;
@@ -696,13 +800,20 @@ int __wrap_abis_rsl_sendmsg(struct msgb *msg)
 		gh = (struct gsm48_hdr*)msg->l3h;
 		switch (gh->msg_type) {
 		case GSM48_MT_RR_HANDO_CMD:
-		case GSM48_MT_RR_ASS_CMD:
-			if (new_ho_req) {
-				fprintf(stderr, "Test script is erratic: a handover is requested"
-					" while a previous handover request is still unhandled\n");
+			if (new_ho_cmd || new_as_cmd) {
+				fprintf(stderr, "Test script is erratic: seen a Handover Command"
+					" while a previous Assignment or Handover Command is still unhandled\n");
 				exit(1);
 			}
-			new_ho_req = lchan;
+			new_ho_cmd = lchan;
+			break;
+		case GSM48_MT_RR_ASS_CMD:
+			if (new_ho_cmd || new_as_cmd) {
+				fprintf(stderr, "Test script is erratic: seen an Assignment Command"
+					" while a previous Assignment or Handover Command is still unhandled\n");
+				exit(1);
+			}
+			new_as_cmd = lchan;
 			break;
 		}
 		break;
@@ -825,8 +936,8 @@ DEFUN(create_bts, create_bts_cmd,
       "Create N TRX in the new BTS\n"
       "TRX count\n"
       "Timeslot config\n"
-      "Timeslot types for 8 * trx-count, each being one of CCCH+SDCCH4|SDCCH8|TCH/F|TCH/H|TCH/F_TCH/H_PDCH|...;"
-      " shorthands: cs+4 = CCCH+SDCCH4; dyn = TCH/F_TCH/H_PDCH\n")
+      "Timeslot types for 8 * trx-count, each being one of CCCH+SDCCH4|SDCCH8|TCH/F|TCH/H|TCH/F_TCH/H_SDCCH8_PDCH|...;"
+      " shorthands: cs+4 = CCCH+SDCCH4; dyn = TCH/F_TCH/H_SDCCH8_PDCH\n")
 {
 	int num_trx = atoi(argv[0]);
 	VTY_ECHO();
@@ -861,6 +972,7 @@ DEFUN(create_ms, create_ms_cmd,
 struct meas_rep_data {
 	int argc;
 	const char **argv;
+	uint8_t bs_power_db;
 };
 
 static void _meas_rep_cb(struct gsm_lchan *lc, void *data)
@@ -912,14 +1024,15 @@ static void _meas_rep_cb(struct gsm_lchan *lc, void *data)
 			fprintf(stderr, " * Neighbor cell #%d, actual BTS %d: rxlev=%d\n", i, neighbor_bts_nr,
 				nm[i].rxlev);
 	}
-	gen_meas_rep(lc, rxlev, rxqual, ta, argc, nm);
+	gen_meas_rep(lc, d->bs_power_db, rxlev, rxqual, ta, argc, nm);
 }
 
-static int _meas_rep(struct vty *vty, int argc, const char **argv)
+static int _meas_rep(struct vty *vty, uint8_t bs_power_db, int argc, const char **argv)
 {
 	struct meas_rep_data d = {
 		.argc = argc - 4,
 		.argv = argv + 4,
+		.bs_power_db = bs_power_db,
 	};
 	parse_lchan_wildcard_args(argv, _meas_rep_cb, &d);
 	return CMD_SUCCESS;
@@ -947,7 +1060,7 @@ DEFUN(meas_rep, meas_rep_cmd,
       MEAS_REP_DOC MEAS_REP_ARGS_DOC)
 {
 	VTY_ECHO();
-	return _meas_rep(vty, argc, argv);
+	return _meas_rep(vty, 0, argc, argv);
 }
 
 DEFUN(meas_rep_repeat, meas_rep_repeat_cmd,
@@ -962,7 +1075,115 @@ DEFUN(meas_rep_repeat, meas_rep_repeat_cmd,
 	argc -= 1;
 
 	while (count--)
-		_meas_rep(vty, argc, argv);
+		_meas_rep(vty, 0, argc, argv);
+	return CMD_SUCCESS;
+}
+
+DEFUN(meas_rep_repeat_bspower, meas_rep_repeat_bspower_cmd,
+      "meas-rep repeat <0-999> bspower <0-31> " MEAS_REP_ARGS,
+      MEAS_REP_DOC
+      "Resend the same measurement report N times\nN\n"
+      "Send a nonzero BS Power value in the measurement report (downlink power reduction)\nBS Power reduction in dB\n"
+      MEAS_REP_ARGS_DOC)
+{
+	int count = atoi(argv[0]);
+	uint8_t bs_power_db = atoi(argv[1]);
+	VTY_ECHO();
+	argv += 2;
+	argc -= 2;
+
+	while (count--)
+		_meas_rep(vty, bs_power_db, argc, argv);
+	return CMD_SUCCESS;
+}
+
+DEFUN(res_ind, res_ind_cmd,
+      "res-ind trx <0-255> <0-255> levels .LEVELS",
+      "Send Resource Indication for a specific TRX, indicating interference levels per lchan\n"
+      "Indicate a BTS and TRX\n" "BTS nr\n" "TRX nr\n"
+      "Indicate interference levels: each level is an index to bts->interf_meas_params.bounds_dbm[],"
+      " i.e. <0-5> or '-' to omit a report for this timeslot/lchan."
+      " Separate timeslots by spaces, for individual subslots directly concatenate values."
+      " If a timeslot has more subslots than provided, the last given value is repeated."
+      " For example: 'res-ind trx 0 0 levels - 1 23 -': on BTS 0 TRX 0, omit ratings for the entire first timeslot,"
+      " send level=1 for timeslot 1, and for timeslot 2 send level=2 for subslot 0 and level=3 for subslot 1.\n")
+{
+	int i;
+	uint8_t level;
+	struct gsm_bts *bts = bts_by_num_str(argv[0]);
+	struct gsm_bts_trx *trx = trx_by_num_str(bts, argv[1]);
+	struct msgb *msg = msgb_alloc_headroom(256, 64, "RES-IND");
+	struct abis_rsl_common_hdr *rslh;
+	uint8_t *res_info_len;
+	VTY_ECHO();
+
+	/* In this test suite, always act as if the interf_meas_params_cfg were already sent to the BTS via OML */
+	bts->interf_meas_params_used = bts->interf_meas_params_cfg;
+
+	argv += 2;
+	argc -= 2;
+
+	rslh = (struct abis_rsl_common_hdr*)msgb_put(msg, sizeof(*rslh));
+	rslh->msg_discr = ABIS_RSL_MDISC_TRX;
+	rslh->msg_type = RSL_MT_RF_RES_IND;
+	msgb_put_u8(msg, RSL_IE_RESOURCE_INFO);
+	res_info_len = msg->tail;
+	msgb_put_u8(msg, 0);
+
+	level = 0xff;
+	for (i = 0; i < ARRAY_SIZE(trx->ts); i++) {
+		const char *ts_str = NULL;
+		struct gsm_lchan *lchan;
+		size_t given_subslots = 0;
+		struct gsm_bts_trx_ts *ts = &trx->ts[i];
+
+		if (i < argc) {
+			ts_str = argv[i];
+			given_subslots = strlen(ts_str);
+		}
+
+		ts_for_n_lchans(lchan, ts, ts->max_lchans_possible) {
+			int chan_nr;
+
+			if (lchan->nr < given_subslots && ts_str) {
+				char subslot_val = ts_str[lchan->nr];
+				switch (subslot_val) {
+				case '-':
+					level = INTERF_BAND_UNKNOWN;
+					break;
+				case '0':
+				case '1':
+				case '2':
+				case '3':
+				case '4':
+				case '5':
+				case '6':
+				case '7':
+					level = subslot_val - '0';
+					break;
+				default:
+					OSMO_ASSERT(false);
+				}
+			}
+
+			if (level == INTERF_BAND_UNKNOWN)
+				continue;
+
+			chan_nr = gsm_lchan2chan_nr(lchan, true);
+			if (chan_nr < 0)
+				continue;
+
+			msgb_put_u8(msg, chan_nr);
+			msgb_put_u8(msg, level << 5);
+		}
+	}
+
+	*res_info_len = msg->tail - res_info_len - 1;
+
+	msg->dst = trx->rsl_link_primary;
+	msg->l2h = msg->data;
+	abis_rsl_rcvmsg(msg);
+
 	return CMD_SUCCESS;
 }
 
@@ -1011,27 +1232,45 @@ static void _expect_chan_activ(struct gsm_lchan *lchan)
 	send_chan_act_ack(lchan, 1);
 }
 
-static void _expect_ho_req(struct gsm_lchan *lchan)
+static void _expect_ho_cmd(struct gsm_lchan *lchan)
 {
-	fprintf(stderr, "- Expecting handover/assignment request at %s\n",
+	fprintf(stderr, "- Expecting Handover Command at %s\n",
 		gsm_lchan_name(lchan));
 
-	if (!new_ho_req) {
-		fprintf(stderr, "Test failed, because no handover was requested\n");
+	if (!new_ho_cmd) {
+		fprintf(stderr, "Test failed, no Handover Command\n");
 		exit(1);
 	}
-	fprintf(stderr, " * Got handover/assignment request at %s\n", gsm_lchan_name(new_ho_req));
-	if (new_ho_req != lchan) {
-		fprintf(stderr, "Test failed, because handover/assignment was not commanded on the expected lchan\n");
+	fprintf(stderr, " * Got Handover Command at %s\n", gsm_lchan_name(new_ho_cmd));
+	if (new_ho_cmd != lchan) {
+		fprintf(stderr, "Test failed, Handover Command not on the expected lchan\n");
 		exit(1);
 	}
-	last_ho_req = new_ho_req;
-	new_ho_req = NULL;
+	last_ho_cmd = new_ho_cmd;
+	new_ho_cmd = NULL;
+}
+
+static void _expect_as_cmd(struct gsm_lchan *lchan)
+{
+	fprintf(stderr, "- Expecting Assignment Command at %s\n",
+		gsm_lchan_name(lchan));
+
+	if (!new_as_cmd) {
+		fprintf(stderr, "Test failed, no Assignment Command\n");
+		exit(1);
+	}
+	fprintf(stderr, " * Got Assignment Command at %s\n", gsm_lchan_name(new_as_cmd));
+	if (new_as_cmd != lchan) {
+		fprintf(stderr, "Test failed, Assignment Command not on the expected lchan\n");
+		exit(1);
+	}
+	last_as_cmd = new_as_cmd;
+	new_as_cmd = NULL;
 }
 
 DEFUN(expect_chan, expect_chan_cmd,
       "expect-chan " LCHAN_ARGS,
-      "Expect a channel request from BSC to a cell for a specific lchan\n"
+      "Expect RSL Channel Activation of a specific lchan\n"
       LCHAN_ARGS_DOC)
 {
 	VTY_ECHO();
@@ -1039,13 +1278,23 @@ DEFUN(expect_chan, expect_chan_cmd,
 	return CMD_SUCCESS;
 }
 
-DEFUN(expect_ho_req, expect_ho_req_cmd,
-      "expect-ho-req " LCHAN_ARGS,
-      "Expect a handover of a given lchan\n"
+DEFUN(expect_handover_command, expect_handover_command_cmd,
+      "expect-ho-cmd " LCHAN_ARGS,
+      "Expect an RR Handover Command sent to a specific lchan\n"
       LCHAN_ARGS_DOC)
 {
 	VTY_ECHO();
-	_expect_ho_req(parse_lchan_args(argv));
+	_expect_ho_cmd(parse_lchan_args(argv));
+	return CMD_SUCCESS;
+}
+
+DEFUN(expect_assignment_command, expect_assignment_command_cmd,
+      "expect-as-cmd " LCHAN_ARGS,
+      "Expect Assignment Command for a given lchan\n"
+      LCHAN_ARGS_DOC)
+{
+	VTY_ECHO();
+	_expect_as_cmd(parse_lchan_args(argv));
 	return CMD_SUCCESS;
 }
 
@@ -1071,12 +1320,12 @@ DEFUN(ho_complete, ho_complete_cmd,
 		fprintf(stderr, "Cannot ack handover/assignment, because no chan request\n");
 		exit(1);
 	}
-	if (!last_ho_req) {
+	if (!last_ho_cmd) {
 		fprintf(stderr, "Cannot ack handover/assignment, because no ho request\n");
 		exit(1);
 	}
 	send_ho_complete(last_chan_req, true);
-	lchan_release_ack(last_ho_req);
+	lchan_release_ack(last_ho_cmd);
 	return CMD_SUCCESS;
 }
 
@@ -1085,16 +1334,41 @@ DEFUN(expect_ho, expect_ho_cmd,
       "Expect a handover of a specific lchan to a specific target lchan;"
       " shorthand for expect-chan, ack-chan, expect-ho, ho-complete.\n"
       "lchan to handover from\n" LCHAN_ARGS_DOC
-      "lchan that to handover to\n" LCHAN_ARGS_DOC)
+      "lchan to handover to\n" LCHAN_ARGS_DOC)
 {
 	struct gsm_lchan *from = parse_lchan_args(argv);
 	struct gsm_lchan *to = parse_lchan_args(argv+4);
 	VTY_ECHO();
 
 	_expect_chan_activ(to);
-	_expect_ho_req(from);
+	_expect_ho_cmd(from);
 	send_ho_detect(to);
 	send_ho_complete(to, true);
+
+	lchan_release_ack(from);
+	return CMD_SUCCESS;
+}
+
+DEFUN(expect_as, expect_as_cmd,
+      "expect-as from " LCHAN_ARGS " to " LCHAN_ARGS,
+      "Expect an intra-cell re-assignment of a specific lchan to a specific target lchan;"
+      " shorthand for expect-chan, ack-chan, expect-as, TODO.\n"
+      "lchan to be re-assigned elsewhere\n" LCHAN_ARGS_DOC
+      "new lchan to re-assign to\n" LCHAN_ARGS_DOC)
+{
+	struct gsm_lchan *from = parse_lchan_args(argv);
+	struct gsm_lchan *to = parse_lchan_args(argv+4);
+	VTY_ECHO();
+
+	_expect_chan_activ(to);
+	if (from->ts->trx->bts != to->ts->trx->bts) {
+		vty_out(vty, "%% Error: re-assignment only works within the same BTS%s", VTY_NEWLINE);
+		return CMD_WARNING;
+	}
+	_expect_as_cmd(from);
+	send_assignment_complete(to);
+	send_est_ind(to);
+
 	lchan_release_ack(from);
 	return CMD_SUCCESS;
 }
@@ -1108,11 +1382,11 @@ DEFUN(ho_failed, ho_failed_cmd,
 		fprintf(stderr, "Cannot fail handover, because no chan request\n");
 		exit(1);
 	}
-	if (!last_ho_req) {
+	if (!last_ho_cmd) {
 		fprintf(stderr, "Cannot fail handover, because no handover request\n");
 		exit(1);
 	}
-	send_ho_complete(last_ho_req, false);
+	send_ho_complete(last_ho_cmd, false);
 	lchan_release_ack(last_chan_req);
 	return CMD_SUCCESS;
 }
@@ -1173,6 +1447,22 @@ DEFUN(set_ts_use, set_ts_use_cmd,
 	return CMD_SUCCESS;
 }
 
+DEFUN(wait, wait_cmd,
+	"wait <0-999999> [<0-999>]",
+	"Let some fake time pass. The test continues instantaneously, but this overrides osmo_gettimeofday() to let"
+	" given amount of time pass virtually.\n"
+	"Seconds to fake-wait\n"
+	"Microseconds to fake-wait, in addition to the seconds waited\n")
+{
+	time_t seconds = atoi(argv[0]);
+	suseconds_t useconds = 0;
+	VTY_ECHO();
+	if (argc > 1)
+		useconds = atoi(argv[1]) * 1000;
+	fake_time_passes(seconds, useconds);
+	return CMD_SUCCESS;
+}
+
 static void ho_test_vty_init()
 {
 	install_element(CONFIG_NODE, &create_n_bts_cmd);
@@ -1180,18 +1470,23 @@ static void ho_test_vty_init()
 	install_element(CONFIG_NODE, &create_ms_cmd);
 	install_element(CONFIG_NODE, &meas_rep_cmd);
 	install_element(CONFIG_NODE, &meas_rep_repeat_cmd);
+	install_element(CONFIG_NODE, &meas_rep_repeat_bspower_cmd);
+	install_element(CONFIG_NODE, &res_ind_cmd);
 	install_element(CONFIG_NODE, &congestion_check_cmd);
 	install_element(CONFIG_NODE, &expect_no_chan_cmd);
 	install_element(CONFIG_NODE, &expect_chan_cmd);
-	install_element(CONFIG_NODE, &expect_ho_req_cmd);
+	install_element(CONFIG_NODE, &expect_handover_command_cmd);
+	install_element(CONFIG_NODE, &expect_assignment_command_cmd);
 	install_element(CONFIG_NODE, &ho_detection_cmd);
 	install_element(CONFIG_NODE, &ho_complete_cmd);
 	install_element(CONFIG_NODE, &expect_ho_cmd);
+	install_element(CONFIG_NODE, &expect_as_cmd);
 	install_element(CONFIG_NODE, &ho_failed_cmd);
 	install_element(CONFIG_NODE, &expect_ts_use_cmd);
 	install_element(CONFIG_NODE, &codec_f_cmd);
 	install_element(CONFIG_NODE, &codec_h_cmd);
 	install_element(CONFIG_NODE, &set_ts_use_cmd);
+	install_element(CONFIG_NODE, &wait_cmd);
 }
 
 static const struct log_info_cat log_categories[] = {
@@ -1305,25 +1600,36 @@ int main(int argc, char **argv)
 
 	osmo_init_logging2(ctx, &log_info);
 
-	log_set_print_filename2(osmo_stderr_target, LOG_FILENAME_NONE);
+	log_set_print_filename2(osmo_stderr_target, LOG_FILENAME_BASENAME);
+	log_set_print_filename_pos(osmo_stderr_target, LOG_FILENAME_POS_LINE_END);
 	log_set_print_category(osmo_stderr_target, 1);
 	log_set_print_category_hex(osmo_stderr_target, 0);
 	log_set_print_level(osmo_stderr_target, 1);
 	log_set_print_timestamp(osmo_stderr_target, 0);
 	osmo_fsm_log_addr(false);
 
+	/* the 'wait' command above, intended to test penalty timers, adds seconds to the monotonic clock in "fake
+	 * time". */
+	fake_time_start();
+
 	bsc_network_alloc();
 	if (!bsc_gsmnet)
 		exit(1);
 
+	/* The MGCP client which is handling the pool (mgcp_client_pool_vty_init) is used from the bsc_vty_init, so
+	 * we must allocate an empty mgw pool even though we do not need it for this test. */
+	bsc_gsmnet->mgw.mgw_pool = mgcp_client_pool_alloc(bsc_gsmnet);
+	if (!bsc_gsmnet->mgw.mgw_pool)
+		exit(1);
+
 	vty_init(&vty_info);
 	bsc_vty_init(bsc_gsmnet);
 	ho_test_vty_init();
 
-	ts_fsm_init();
 	lchan_fsm_init();
 	bsc_subscr_conn_fsm_init();
 	handover_fsm_init();
+	assignment_fsm_init();
 
 	ho_set_algorithm(bsc_gsmnet->ho, 2);
 	ho_set_ho_active(bsc_gsmnet->ho, true);
@@ -1376,33 +1682,17 @@ void trau_mux_unmap() {}
 void trau_mux_map_lchan() {}
 void trau_recv_lchan() {}
 void trau_send_frame() {}
-int osmo_bsc_sigtran_send(struct gsm_subscriber_connection *conn, struct msgb *msg) { return 0; }
+/* Stub */
 int osmo_bsc_sigtran_open_conn(struct gsm_subscriber_connection *conn, struct msgb *msg) { return 0; }
 void bsc_sapi_n_reject(struct gsm_subscriber_connection *conn, uint8_t dlci, enum gsm0808_cause cause) {}
 void bsc_cipher_mode_compl(struct gsm_subscriber_connection *conn, struct msgb *msg, uint8_t chosen_encr) {}
 int bsc_compl_l3(struct gsm_lchan *lchan, struct msgb *msg, uint16_t chosen_channel)
 { return 0; }
-int bsc_paging_start(struct bsc_paging_params *params)
-{ return 0; }
 void bsc_dtap(struct gsm_subscriber_connection *conn, uint8_t link_id, struct msgb *msg) {}
 void bsc_assign_compl(struct gsm_subscriber_connection *conn, uint8_t rr_cause) {}
 void bsc_cm_update(struct gsm_subscriber_connection *conn,
 		   const uint8_t *cm2, uint8_t cm2_len,
 		   const uint8_t *cm3, uint8_t cm3_len) {}
-struct gsm0808_handover_required;
-int bsc_tx_bssmap_ho_required(struct gsm_lchan *lchan, const struct gsm0808_cell_id_list2 *target_cells)
-{ return 0; }
-int bsc_tx_bssmap_ho_request_ack(struct gsm_subscriber_connection *conn, struct msgb *rr_ho_command)
-{ return 0; }
-int bsc_tx_bssmap_ho_detect(struct gsm_subscriber_connection *conn) { return 0; }
-enum handover_result bsc_tx_bssmap_ho_complete(struct gsm_subscriber_connection *conn,
-					       struct gsm_lchan *lchan) { return HO_RESULT_OK; }
-void bsc_tx_bssmap_ho_failure(struct gsm_subscriber_connection *conn) {}
-void osmo_bsc_sigtran_tx_reset(void) {}
-void osmo_bsc_sigtran_tx_reset_ack(void) {}
-void osmo_bsc_sigtran_reset(void) {}
-void bssmap_reset_alloc(void) {}
-void bssmap_reset_is_conn_ready(void) {}
 const char *osmo_mgcpc_ep_name(const struct osmo_mgcpc_ep *ep)
 {
 	return "fake-ep";
@@ -1411,3 +1701,12 @@ const char *osmo_mgcpc_ep_ci_name(const struct osmo_mgcpc_ep_ci *ci)
 {
 	return "fake-ci";
 }
+const struct mgcp_conn_peer *osmo_mgcpc_ep_ci_get_rtp_info(const struct osmo_mgcpc_ep_ci *ci)
+{
+	static struct mgcp_conn_peer ret = {
+		.addr = "1.2.3.4",
+		.port = 1234,
+		.endpoint = "fake-endpoint",
+	};
+	return &ret;
+}
